/*
 * Copyright (c) 2016, Alliance for Open Media. All rights reserved
 *
 * This source code is subject to the terms of the BSD 2 Clause License and
 * the Alliance for Open Media Patent License 1.0. If the BSD 2 Clause License
 * was not distributed with this source code in the LICENSE file, you can
 * obtain it at www.aomedia.org/license/software. If the Alliance for Open
 * Media Patent License 1.0 was not distributed with this source code in the
 * PATENTS file, you can obtain it at www.aomedia.org/license/patent.
 */

#include "./aom_config.h"
#include "aom/aom_integer.h"
#include "aom_mem/aom_mem.h"
#include "av1/common/blockd.h"
#include "av1/common/entropy.h"
#include "av1/common/entropymode.h"
#include "av1/common/onyxc_int.h"
#include "av1/common/scan.h"
#if CONFIG_LV_MAP
#include "av1/common/txb_common.h"
#endif

// Unconstrained Node Tree
/* clang-format off */
const aom_tree_index av1_coef_con_tree[TREE_SIZE(ENTROPY_TOKENS)] = {
  2, 6,                                // 0 = LOW_VAL
  -TWO_TOKEN, 4,                       // 1 = TWO
  -THREE_TOKEN, -FOUR_TOKEN,           // 2 = THREE
  8, 10,                               // 3 = HIGH_LOW
  -CATEGORY1_TOKEN, -CATEGORY2_TOKEN,  // 4 = CAT_ONE
  12, 14,                              // 5 = CAT_THREEFOUR
  -CATEGORY3_TOKEN, -CATEGORY4_TOKEN,  // 6 = CAT_THREE
  -CATEGORY5_TOKEN, -CATEGORY6_TOKEN   // 7 = CAT_FIVE
};
/* clang-format on */

#if CONFIG_NEW_MULTISYMBOL
/* Extra bits coded from LSB to MSB */
const aom_cdf_prob av1_cat1_cdf0[CDF_SIZE(2)] = { 20352, 32768, 0 };
const aom_cdf_prob *av1_cat1_cdf[] = { av1_cat1_cdf0 };

const aom_cdf_prob av1_cat2_cdf0[CDF_SIZE(4)] = { 11963, 21121, 27719, 32768,
                                                  0 };
const aom_cdf_prob *av1_cat2_cdf[] = { av1_cat2_cdf0 };
const aom_cdf_prob av1_cat3_cdf0[CDF_SIZE(8)] = {
  7001, 12802, 17911, 22144, 25503, 28286, 30737, 32768, 0
};
const aom_cdf_prob *av1_cat3_cdf[] = { av1_cat3_cdf0 };

const aom_cdf_prob av1_cat4_cdf0[CDF_SIZE(16)] = {
  3934,  7460,  10719, 13640, 16203, 18500, 20624, 22528, 24316,
  25919, 27401, 28729, 29894, 30938, 31903, 32768, 0
};
const aom_cdf_prob *av1_cat4_cdf[] = { av1_cat4_cdf0 };

const aom_cdf_prob av1_cat5_cdf0[CDF_SIZE(16)] = {
  2942,  5794,  8473,  11069, 13469, 15795, 17980, 20097, 21952,
  23750, 25439, 27076, 28589, 30056, 31434, 32768, 0
};
const aom_cdf_prob av1_cat5_cdf1[CDF_SIZE(2)] = { 23040, 32768, 0 };
const aom_cdf_prob *av1_cat5_cdf[] = { av1_cat5_cdf0, av1_cat5_cdf1 };

const aom_cdf_prob av1_cat6_cdf0[CDF_SIZE(16)] = { 2382,  4727,  7036,  9309,
                                                   11512, 13681, 15816, 17918,
                                                   19892, 21835, 23748, 25632,
                                                   27458, 29255, 31024, 32768 };
const aom_cdf_prob av1_cat6_cdf1[CDF_SIZE(16)] = { 9314,  15584, 19741, 22540,
                                                   25391, 27310, 28583, 29440,
                                                   30493, 31202, 31672, 31988,
                                                   32310, 32527, 32671, 32768 };
const aom_cdf_prob av1_cat6_cdf2[CDF_SIZE(16)] = { 29548, 31129, 31960, 32004,
                                                   32473, 32498, 32511, 32512,
                                                   32745, 32757, 32764, 32764,
                                                   32768, 32768, 32768, 32768 };
const aom_cdf_prob av1_cat6_cdf3[CDF_SIZE(16)] = { 32006, 32258, 32510, 32512,
                                                   32638, 32639, 32640, 32640,
                                                   32766, 32767, 32768, 32768,
                                                   32768, 32768, 32768, 32768 };
const aom_cdf_prob av1_cat6_cdf4[CDF_SIZE(4)] = { 32513, 32641, 32769, 32768 };
const aom_cdf_prob *av1_cat6_cdf[] = {
  av1_cat6_cdf0, av1_cat6_cdf1, av1_cat6_cdf2, av1_cat6_cdf3, av1_cat6_cdf4
};
#endif
/* Extra bits coded from MSB to LSB */
const aom_prob av1_cat1_prob[] = { 159 };
const aom_prob av1_cat2_prob[] = { 165, 145 };
const aom_prob av1_cat3_prob[] = { 173, 148, 140 };
const aom_prob av1_cat4_prob[] = { 176, 155, 140, 135 };
const aom_prob av1_cat5_prob[] = { 180, 157, 141, 134, 130 };
const aom_prob av1_cat6_prob[] = {
  255, 255, 255, 255, 254, 254, 254, 252, 249,
  243, 230, 196, 177, 153, 140, 133, 130, 129
};

const uint16_t band_count_table[TX_SIZES_ALL][8] = {
#if CONFIG_CB4X4
  { 1, 2, 2, 3, 0, 0, 0 },
#endif
  { 1, 2, 3, 4, 3, 16 - 13, 0 },    { 1, 2, 3, 4, 11, 64 - 21, 0 },
  { 1, 2, 3, 4, 11, 256 - 21, 0 },  { 1, 2, 3, 4, 11, 1024 - 21, 0 },
#if CONFIG_TX64X64
  { 1, 2, 3, 4, 11, 4096 - 21, 0 },
#endif  // CONFIG_TX64X64
  { 1, 2, 3, 4, 8, 32 - 18, 0 },    { 1, 2, 3, 4, 8, 32 - 18, 0 },
  { 1, 2, 3, 4, 11, 128 - 21, 0 },  { 1, 2, 3, 4, 11, 128 - 21, 0 },
  { 1, 2, 3, 4, 11, 512 - 21, 0 },  { 1, 2, 3, 4, 11, 512 - 21, 0 },
  { 1, 2, 3, 4, 11, 64 - 21, 0 },   { 1, 2, 3, 4, 11, 64 - 21, 0 },
  { 1, 2, 3, 4, 11, 256 - 21, 0 },  { 1, 2, 3, 4, 11, 256 - 21, 0 },
};

const uint16_t band_cum_count_table[TX_SIZES_ALL][8] = {
#if CONFIG_CB4X4
  { 0, 1, 3, 6, 10, 13, 16, 0 },
#endif
  { 0, 1, 3, 6, 10, 13, 16, 0 },   { 0, 1, 3, 6, 10, 21, 64, 0 },
  { 0, 1, 3, 6, 10, 21, 256, 0 },  { 0, 1, 3, 6, 10, 21, 1024, 0 },
#if CONFIG_TX64X64
  { 0, 1, 3, 6, 10, 21, 4096, 0 },
#endif  // CONFIG_TX64X64
  { 0, 1, 3, 6, 10, 18, 32, 0 },   { 0, 1, 3, 6, 10, 18, 32, 0 },
  { 0, 1, 3, 6, 10, 21, 128, 0 },  { 0, 1, 3, 6, 10, 21, 128, 0 },
  { 0, 1, 3, 6, 10, 21, 512, 0 },  { 0, 1, 3, 6, 10, 21, 512, 0 },
  { 0, 1, 3, 6, 10, 21, 64, 0 },   { 0, 1, 3, 6, 10, 21, 64, 0 },
  { 0, 1, 3, 6, 10, 21, 256, 0 },  { 0, 1, 3, 6, 10, 21, 256, 0 },
};

const uint8_t av1_coefband_trans_8x8plus[MAX_TX_SQUARE] = {
  0, 1, 1, 2, 2, 2, 3, 3, 3, 3, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 5,
  // beyond MAXBAND_INDEX+1 all values are filled as 5
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
#if CONFIG_TX64X64
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5
#endif  // CONFIG_TX64X64
};

const uint8_t av1_coefband_trans_4x8_8x4[32] = {
  0, 1, 1, 2, 2, 2, 3, 3, 3, 3, 4, 4, 4, 4, 4, 4,
  4, 4, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
};

const uint8_t av1_coefband_trans_4x4[16] = {
  0, 1, 1, 2, 2, 2, 3, 3, 3, 3, 4, 4, 4, 5, 5, 5,
};

const uint8_t av1_pt_energy_class[ENTROPY_TOKENS] = { 0, 1, 2, 3, 3, 4,
                                                      4, 5, 5, 5, 5, 5 };

// Model obtained from a 2-sided zero-centered distribution derived
// from a Pareto distribution. The cdf of the distribution is:
// cdf(x) = 0.5 + 0.5 * sgn(x) * [1 - {alpha/(alpha + |x|)} ^ beta]
//
// For a given beta and a given probablity of the 1-node, the alpha
// is first solved, and then the {alpha, beta} pair is used to generate
// the probabilities for the rest of the nodes.

// beta = 8

// Every odd line in this table can be generated from the even lines
// by averaging :
// av1_pareto8_full[l][node] = (av1_pareto8_full[l-1][node] +
//                              av1_pareto8_full[l+1][node] ) >> 1;
// Values for tokens ONE_TOKEN through CATEGORY6_TOKEN included here.
const aom_prob av1_pareto8_full[COEFF_PROB_MODELS][MODEL_NODES] = {
  { 3, 86, 128, 6, 86, 23, 88, 29 },
  { 6, 86, 128, 11, 87, 42, 91, 52 },
  { 9, 86, 129, 17, 88, 61, 94, 76 },
  { 12, 86, 129, 22, 88, 77, 97, 93 },
  { 15, 87, 129, 28, 89, 93, 100, 110 },
  { 17, 87, 129, 33, 90, 105, 103, 123 },
  { 20, 88, 130, 38, 91, 118, 106, 136 },
  { 23, 88, 130, 43, 91, 128, 108, 146 },
  { 26, 89, 131, 48, 92, 139, 111, 156 },
  { 28, 89, 131, 53, 93, 147, 114, 163 },
  { 31, 90, 131, 58, 94, 156, 117, 171 },
  { 34, 90, 131, 62, 94, 163, 119, 177 },
  { 37, 90, 132, 66, 95, 171, 122, 184 },
  { 39, 90, 132, 70, 96, 177, 124, 189 },
  { 42, 91, 132, 75, 97, 183, 127, 194 },
  { 44, 91, 132, 79, 97, 188, 129, 198 },
  { 47, 92, 133, 83, 98, 193, 132, 202 },
  { 49, 92, 133, 86, 99, 197, 134, 205 },
  { 52, 93, 133, 90, 100, 201, 137, 208 },
  { 54, 93, 133, 94, 100, 204, 139, 211 },
  { 57, 94, 134, 98, 101, 208, 142, 214 },
  { 59, 94, 134, 101, 102, 211, 144, 216 },
  { 62, 94, 135, 105, 103, 214, 146, 218 },
  { 64, 94, 135, 108, 103, 216, 148, 220 },
  { 66, 95, 135, 111, 104, 219, 151, 222 },
  { 68, 95, 135, 114, 105, 221, 153, 223 },
  { 71, 96, 136, 117, 106, 224, 155, 225 },
  { 73, 96, 136, 120, 106, 225, 157, 226 },
  { 76, 97, 136, 123, 107, 227, 159, 228 },
  { 78, 97, 136, 126, 108, 229, 160, 229 },
  { 80, 98, 137, 129, 109, 231, 162, 231 },
  { 82, 98, 137, 131, 109, 232, 164, 232 },
  { 84, 98, 138, 134, 110, 234, 166, 233 },
  { 86, 98, 138, 137, 111, 235, 168, 234 },
  { 89, 99, 138, 140, 112, 236, 170, 235 },
  { 91, 99, 138, 142, 112, 237, 171, 235 },
  { 93, 100, 139, 145, 113, 238, 173, 236 },
  { 95, 100, 139, 147, 114, 239, 174, 237 },
  { 97, 101, 140, 149, 115, 240, 176, 238 },
  { 99, 101, 140, 151, 115, 241, 177, 238 },
  { 101, 102, 140, 154, 116, 242, 179, 239 },
  { 103, 102, 140, 156, 117, 242, 180, 239 },
  { 105, 103, 141, 158, 118, 243, 182, 240 },
  { 107, 103, 141, 160, 118, 243, 183, 240 },
  { 109, 104, 141, 162, 119, 244, 185, 241 },
  { 111, 104, 141, 164, 119, 244, 186, 241 },
  { 113, 104, 142, 166, 120, 245, 187, 242 },
  { 114, 104, 142, 168, 121, 245, 188, 242 },
  { 116, 105, 143, 170, 122, 246, 190, 243 },
  { 118, 105, 143, 171, 122, 246, 191, 243 },
  { 120, 106, 143, 173, 123, 247, 192, 244 },
  { 121, 106, 143, 175, 124, 247, 193, 244 },
  { 123, 107, 144, 177, 125, 248, 195, 244 },
  { 125, 107, 144, 178, 125, 248, 196, 244 },
  { 127, 108, 145, 180, 126, 249, 197, 245 },
  { 128, 108, 145, 181, 127, 249, 198, 245 },
  { 130, 109, 145, 183, 128, 249, 199, 245 },
  { 132, 109, 145, 184, 128, 249, 200, 245 },
  { 134, 110, 146, 186, 129, 250, 201, 246 },
  { 135, 110, 146, 187, 130, 250, 202, 246 },
  { 137, 111, 147, 189, 131, 251, 203, 246 },
  { 138, 111, 147, 190, 131, 251, 204, 246 },
  { 140, 112, 147, 192, 132, 251, 205, 247 },
  { 141, 112, 147, 193, 132, 251, 206, 247 },
  { 143, 113, 148, 194, 133, 251, 207, 247 },
  { 144, 113, 148, 195, 134, 251, 207, 247 },
  { 146, 114, 149, 197, 135, 252, 208, 248 },
  { 147, 114, 149, 198, 135, 252, 209, 248 },
  { 149, 115, 149, 199, 136, 252, 210, 248 },
  { 150, 115, 149, 200, 137, 252, 210, 248 },
  { 152, 115, 150, 201, 138, 252, 211, 248 },
  { 153, 115, 150, 202, 138, 252, 212, 248 },
  { 155, 116, 151, 204, 139, 253, 213, 249 },
  { 156, 116, 151, 205, 139, 253, 213, 249 },
  { 158, 117, 151, 206, 140, 253, 214, 249 },
  { 159, 117, 151, 207, 141, 253, 215, 249 },
  { 161, 118, 152, 208, 142, 253, 216, 249 },
  { 162, 118, 152, 209, 142, 253, 216, 249 },
  { 163, 119, 153, 210, 143, 253, 217, 249 },
  { 164, 119, 153, 211, 143, 253, 217, 249 },
  { 166, 120, 153, 212, 144, 254, 218, 250 },
  { 167, 120, 153, 212, 145, 254, 219, 250 },
  { 168, 121, 154, 213, 146, 254, 220, 250 },
  { 169, 121, 154, 214, 146, 254, 220, 250 },
  { 171, 122, 155, 215, 147, 254, 221, 250 },
  { 172, 122, 155, 216, 147, 254, 221, 250 },
  { 173, 123, 155, 217, 148, 254, 222, 250 },
  { 174, 123, 155, 217, 149, 254, 222, 250 },
  { 176, 124, 156, 218, 150, 254, 223, 250 },
  { 177, 124, 156, 219, 150, 254, 223, 250 },
  { 178, 125, 157, 220, 151, 254, 224, 251 },
  { 179, 125, 157, 220, 151, 254, 224, 251 },
  { 180, 126, 157, 221, 152, 254, 225, 251 },
  { 181, 126, 157, 221, 152, 254, 225, 251 },
  { 183, 127, 158, 222, 153, 254, 226, 251 },
  { 184, 127, 158, 223, 154, 254, 226, 251 },
  { 185, 128, 159, 224, 155, 255, 227, 251 },
  { 186, 128, 159, 224, 155, 255, 227, 251 },
  { 187, 129, 160, 225, 156, 255, 228, 251 },
  { 188, 130, 160, 225, 156, 255, 228, 251 },
  { 189, 131, 160, 226, 157, 255, 228, 251 },
  { 190, 131, 160, 226, 158, 255, 228, 251 },
  { 191, 132, 161, 227, 159, 255, 229, 251 },
  { 192, 132, 161, 227, 159, 255, 229, 251 },
  { 193, 133, 162, 228, 160, 255, 230, 252 },
  { 194, 133, 162, 229, 160, 255, 230, 252 },
  { 195, 134, 163, 230, 161, 255, 231, 252 },
  { 196, 134, 163, 230, 161, 255, 231, 252 },
  { 197, 135, 163, 231, 162, 255, 231, 252 },
  { 198, 135, 163, 231, 162, 255, 231, 252 },
  { 199, 136, 164, 232, 163, 255, 232, 252 },
  { 200, 136, 164, 232, 164, 255, 232, 252 },
  { 201, 137, 165, 233, 165, 255, 233, 252 },
  { 201, 137, 165, 233, 165, 255, 233, 252 },
  { 202, 138, 166, 233, 166, 255, 233, 252 },
  { 203, 138, 166, 233, 166, 255, 233, 252 },
  { 204, 139, 166, 234, 167, 255, 234, 252 },
  { 205, 139, 166, 234, 167, 255, 234, 252 },
  { 206, 140, 167, 235, 168, 255, 235, 252 },
  { 206, 140, 167, 235, 168, 255, 235, 252 },
  { 207, 141, 168, 236, 169, 255, 235, 252 },
  { 208, 141, 168, 236, 170, 255, 235, 252 },
  { 209, 142, 169, 237, 171, 255, 236, 252 },
  { 209, 143, 169, 237, 171, 255, 236, 252 },
  { 210, 144, 169, 237, 172, 255, 236, 252 },
  { 211, 144, 169, 237, 172, 255, 236, 252 },
  { 212, 145, 170, 238, 173, 255, 237, 252 },
  { 213, 145, 170, 238, 173, 255, 237, 252 },
  { 214, 146, 171, 239, 174, 255, 237, 253 },
  { 214, 146, 171, 239, 174, 255, 237, 253 },
  { 215, 147, 172, 240, 175, 255, 238, 253 },
  { 215, 147, 172, 240, 175, 255, 238, 253 },
  { 216, 148, 173, 240, 176, 255, 238, 253 },
  { 217, 148, 173, 240, 176, 255, 238, 253 },
  { 218, 149, 173, 241, 177, 255, 239, 253 },
  { 218, 149, 173, 241, 178, 255, 239, 253 },
  { 219, 150, 174, 241, 179, 255, 239, 253 },
  { 219, 151, 174, 241, 179, 255, 239, 253 },
  { 220, 152, 175, 242, 180, 255, 240, 253 },
  { 221, 152, 175, 242, 180, 255, 240, 253 },
  { 222, 153, 176, 242, 181, 255, 240, 253 },
  { 222, 153, 176, 242, 181, 255, 240, 253 },
  { 223, 154, 177, 243, 182, 255, 240, 253 },
  { 223, 154, 177, 243, 182, 255, 240, 253 },
  { 224, 155, 178, 244, 183, 255, 241, 253 },
  { 224, 155, 178, 244, 183, 255, 241, 253 },
  { 225, 156, 178, 244, 184, 255, 241, 253 },
  { 225, 157, 178, 244, 184, 255, 241, 253 },
  { 226, 158, 179, 244, 185, 255, 242, 253 },
  { 227, 158, 179, 244, 185, 255, 242, 253 },
  { 228, 159, 180, 245, 186, 255, 242, 253 },
  { 228, 159, 180, 245, 186, 255, 242, 253 },
  { 229, 160, 181, 245, 187, 255, 242, 253 },
  { 229, 160, 181, 245, 187, 255, 242, 253 },
  { 230, 161, 182, 246, 188, 255, 243, 253 },
  { 230, 162, 182, 246, 188, 255, 243, 253 },
  { 231, 163, 183, 246, 189, 255, 243, 253 },
  { 231, 163, 183, 246, 189, 255, 243, 253 },
  { 232, 164, 184, 247, 190, 255, 243, 253 },
  { 232, 164, 184, 247, 190, 255, 243, 253 },
  { 233, 165, 185, 247, 191, 255, 244, 253 },
  { 233, 165, 185, 247, 191, 255, 244, 253 },
  { 234, 166, 185, 247, 192, 255, 244, 253 },
  { 234, 167, 185, 247, 192, 255, 244, 253 },
  { 235, 168, 186, 248, 193, 255, 244, 253 },
  { 235, 168, 186, 248, 193, 255, 244, 253 },
  { 236, 169, 187, 248, 194, 255, 244, 253 },
  { 236, 169, 187, 248, 194, 255, 244, 253 },
  { 236, 170, 188, 248, 195, 255, 245, 253 },
  { 236, 170, 188, 248, 195, 255, 245, 253 },
  { 237, 171, 189, 249, 196, 255, 245, 254 },
  { 237, 172, 189, 249, 196, 255, 245, 254 },
  { 238, 173, 190, 249, 197, 255, 245, 254 },
  { 238, 173, 190, 249, 197, 255, 245, 254 },
  { 239, 174, 191, 249, 198, 255, 245, 254 },
  { 239, 174, 191, 249, 198, 255, 245, 254 },
  { 240, 175, 192, 249, 199, 255, 246, 254 },
  { 240, 176, 192, 249, 199, 255, 246, 254 },
  { 240, 177, 193, 250, 200, 255, 246, 254 },
  { 240, 177, 193, 250, 200, 255, 246, 254 },
  { 241, 178, 194, 250, 201, 255, 246, 254 },
  { 241, 178, 194, 250, 201, 255, 246, 254 },
  { 242, 179, 195, 250, 202, 255, 246, 254 },
  { 242, 180, 195, 250, 202, 255, 246, 254 },
  { 242, 181, 196, 250, 203, 255, 247, 254 },
  { 242, 181, 196, 250, 203, 255, 247, 254 },
  { 243, 182, 197, 251, 204, 255, 247, 254 },
  { 243, 183, 197, 251, 204, 255, 247, 254 },
  { 244, 184, 198, 251, 205, 255, 247, 254 },
  { 244, 184, 198, 251, 205, 255, 247, 254 },
  { 244, 185, 199, 251, 206, 255, 247, 254 },
  { 244, 185, 199, 251, 206, 255, 247, 254 },
  { 245, 186, 200, 251, 207, 255, 247, 254 },
  { 245, 187, 200, 251, 207, 255, 247, 254 },
  { 246, 188, 201, 252, 207, 255, 248, 254 },
  { 246, 188, 201, 252, 207, 255, 248, 254 },
  { 246, 189, 202, 252, 208, 255, 248, 254 },
  { 246, 190, 202, 252, 208, 255, 248, 254 },
  { 247, 191, 203, 252, 209, 255, 248, 254 },
  { 247, 191, 203, 252, 209, 255, 248, 254 },
  { 247, 192, 204, 252, 210, 255, 248, 254 },
  { 247, 193, 204, 252, 210, 255, 248, 254 },
  { 248, 194, 205, 252, 211, 255, 248, 254 },
  { 248, 194, 205, 252, 211, 255, 248, 254 },
  { 248, 195, 206, 252, 212, 255, 249, 254 },
  { 248, 196, 206, 252, 212, 255, 249, 254 },
  { 249, 197, 207, 253, 213, 255, 249, 254 },
  { 249, 197, 207, 253, 213, 255, 249, 254 },
  { 249, 198, 208, 253, 214, 255, 249, 254 },
  { 249, 199, 209, 253, 214, 255, 249, 254 },
  { 250, 200, 210, 253, 215, 255, 249, 254 },
  { 250, 200, 210, 253, 215, 255, 249, 254 },
  { 250, 201, 211, 253, 215, 255, 249, 254 },
  { 250, 202, 211, 253, 215, 255, 249, 254 },
  { 250, 203, 212, 253, 216, 255, 249, 254 },
  { 250, 203, 212, 253, 216, 255, 249, 254 },
  { 251, 204, 213, 253, 217, 255, 250, 254 },
  { 251, 205, 213, 253, 217, 255, 250, 254 },
  { 251, 206, 214, 254, 218, 255, 250, 254 },
  { 251, 206, 215, 254, 218, 255, 250, 254 },
  { 252, 207, 216, 254, 219, 255, 250, 254 },
  { 252, 208, 216, 254, 219, 255, 250, 254 },
  { 252, 209, 217, 254, 220, 255, 250, 254 },
  { 252, 210, 217, 254, 220, 255, 250, 254 },
  { 252, 211, 218, 254, 221, 255, 250, 254 },
  { 252, 212, 218, 254, 221, 255, 250, 254 },
  { 253, 213, 219, 254, 222, 255, 250, 254 },
  { 253, 213, 220, 254, 222, 255, 250, 254 },
  { 253, 214, 221, 254, 223, 255, 250, 254 },
  { 253, 215, 221, 254, 223, 255, 250, 254 },
  { 253, 216, 222, 254, 224, 255, 251, 254 },
  { 253, 217, 223, 254, 224, 255, 251, 254 },
  { 253, 218, 224, 254, 225, 255, 251, 254 },
  { 253, 219, 224, 254, 225, 255, 251, 254 },
  { 254, 220, 225, 254, 225, 255, 251, 254 },
  { 254, 221, 226, 254, 225, 255, 251, 254 },
  { 254, 222, 227, 255, 226, 255, 251, 254 },
  { 254, 223, 227, 255, 226, 255, 251, 254 },
  { 254, 224, 228, 255, 227, 255, 251, 254 },
  { 254, 225, 229, 255, 227, 255, 251, 254 },
  { 254, 226, 230, 255, 228, 255, 251, 254 },
  { 254, 227, 230, 255, 229, 255, 251, 254 },
  { 255, 228, 231, 255, 230, 255, 251, 254 },
  { 255, 229, 232, 255, 230, 255, 251, 254 },
  { 255, 230, 233, 255, 231, 255, 252, 254 },
  { 255, 231, 234, 255, 231, 255, 252, 254 },
  { 255, 232, 235, 255, 232, 255, 252, 254 },
  { 255, 233, 236, 255, 232, 255, 252, 254 },
  { 255, 235, 237, 255, 233, 255, 252, 254 },
  { 255, 236, 238, 255, 234, 255, 252, 254 },
  { 255, 238, 240, 255, 235, 255, 252, 255 },
  { 255, 239, 241, 255, 235, 255, 252, 254 },
  { 255, 241, 243, 255, 236, 255, 252, 254 },
  { 255, 243, 245, 255, 237, 255, 252, 254 },
  { 255, 246, 247, 255, 239, 255, 253, 255 },
};

// Model obtained from a 2-sided zero-centered distribution derived
// from a Pareto distribution. The cdf of the distribution is:
// cdf(x) = 0.5 + 0.5 * sgn(x) * [1 - {alpha/(alpha + |x|)} ^ beta]
//
// For a given beta and a given probability of the 1-node, the alpha
// is first solved, and then the {alpha, beta} pair is used to generate
// the probabilities for the rest of the nodes.
//
// The full source code of the generating program is available in:
// tools/gen_constrained_tokenset.py
//
#if CONFIG_NEW_TOKENSET
// Values for tokens TWO_TOKEN through CATEGORY6_TOKEN included
// in the table here : the ONE_TOKEN probability is
// removed and the probabilities rescaled.
//
// ZERO_TOKEN and ONE_TOKEN are coded as one CDF,
// and EOB_TOKEN is coded as flags outside this coder.
const aom_cdf_prob av1_pareto8_tail_probs[COEFF_PROB_MODELS][TAIL_NODES] = {
  { 128, 127, 127, 252, 497, 969, 1839, 3318, 25511 },
  { 256, 254, 251, 496, 966, 1834, 3308, 5408, 19995 },
  { 383, 378, 373, 732, 1408, 2605, 4470, 6646, 15773 },
  { 511, 502, 493, 961, 1824, 3289, 5373, 7298, 12517 },
  { 638, 625, 611, 1182, 2215, 3894, 6064, 7548, 9991 },
  { 766, 746, 726, 1396, 2582, 4428, 6578, 7529, 8017 },
  { 893, 866, 839, 1603, 2927, 4896, 6945, 7332, 6467 },
  { 1020, 984, 950, 1803, 3250, 5305, 7191, 7022, 5243 },
  { 1147, 1102, 1059, 1996, 3552, 5659, 7338, 6646, 4269 },
  { 1274, 1218, 1166, 2183, 3835, 5963, 7403, 6234, 3492 },
  { 1400, 1334, 1270, 2363, 4099, 6223, 7401, 5809, 2869 },
  { 1527, 1447, 1372, 2537, 4345, 6442, 7346, 5386, 2366 },
  { 1654, 1560, 1473, 2704, 4574, 6624, 7247, 4973, 1959 },
  { 1780, 1672, 1571, 2866, 4787, 6771, 7114, 4579, 1628 },
  { 1906, 1782, 1667, 3022, 4984, 6889, 6954, 4206, 1358 },
  { 2032, 1891, 1762, 3172, 5167, 6979, 6773, 3856, 1136 },
  { 2158, 2000, 1854, 3316, 5335, 7044, 6577, 3530, 954 },
  { 2284, 2106, 1944, 3455, 5490, 7087, 6370, 3229, 803 },
  { 2410, 2212, 2032, 3588, 5632, 7109, 6155, 2951, 679 },
  { 2535, 2317, 2119, 3717, 5761, 7113, 5936, 2695, 575 },
  { 2661, 2420, 2203, 3840, 5880, 7101, 5714, 2461, 488 },
  { 2786, 2522, 2286, 3958, 5987, 7074, 5493, 2246, 416 },
  { 2911, 2624, 2367, 4072, 6083, 7033, 5273, 2050, 355 },
  { 3037, 2724, 2446, 4180, 6170, 6981, 5055, 1871, 304 },
  { 3162, 2822, 2523, 4284, 6247, 6919, 4842, 1708, 261 },
  { 3286, 2920, 2599, 4384, 6315, 6848, 4633, 1559, 224 },
  { 3411, 3017, 2672, 4478, 6374, 6768, 4430, 1424, 194 },
  { 3536, 3112, 2745, 4569, 6426, 6681, 4232, 1300, 167 },
  { 3660, 3207, 2815, 4656, 6469, 6588, 4040, 1188, 145 },
  { 3785, 3300, 2883, 4738, 6505, 6490, 3855, 1086, 126 },
  { 3909, 3392, 2950, 4817, 6534, 6387, 3677, 993, 109 },
  { 4033, 3483, 3015, 4891, 6557, 6281, 3505, 908, 95 },
  { 4157, 3573, 3079, 4962, 6573, 6170, 3340, 831, 83 },
  { 4281, 3662, 3141, 5029, 6584, 6058, 3181, 760, 72 },
  { 4405, 3750, 3201, 5093, 6588, 5943, 3029, 696, 63 },
  { 4529, 3837, 3260, 5152, 6587, 5826, 2883, 638, 56 },
  { 4652, 3922, 3317, 5209, 6582, 5709, 2744, 584, 49 },
  { 4775, 4007, 3373, 5262, 6572, 5590, 2610, 536, 43 },
  { 4899, 4090, 3427, 5312, 6557, 5470, 2483, 492, 38 },
  { 5022, 4173, 3480, 5359, 6538, 5351, 2361, 451, 33 },
  { 5145, 4254, 3531, 5403, 6515, 5231, 2246, 414, 29 },
  { 5268, 4334, 3581, 5443, 6489, 5112, 2135, 380, 26 },
  { 5391, 4414, 3629, 5481, 6458, 4993, 2029, 350, 23 },
  { 5514, 4492, 3676, 5515, 6425, 4875, 1929, 321, 21 },
  { 5637, 4569, 3721, 5548, 6388, 4758, 1833, 296, 18 },
  { 5759, 4645, 3766, 5577, 6349, 4642, 1742, 272, 16 },
  { 5881, 4720, 3808, 5604, 6307, 4528, 1656, 250, 14 },
  { 6004, 4794, 3849, 5628, 6262, 4414, 1573, 231, 13 },
  { 6126, 4867, 3890, 5649, 6215, 4302, 1495, 213, 11 },
  { 6248, 4939, 3928, 5669, 6166, 4192, 1420, 196, 10 },
  { 6370, 5010, 3966, 5686, 6114, 4083, 1349, 181, 9 },
  { 6492, 5080, 4002, 5700, 6061, 3976, 1282, 167, 8 },
  { 6614, 5149, 4037, 5712, 6006, 3871, 1218, 154, 7 },
  { 6735, 5217, 4070, 5723, 5950, 3767, 1157, 142, 7 },
  { 6857, 5284, 4103, 5731, 5891, 3666, 1099, 131, 6 },
  { 6978, 5351, 4134, 5737, 5832, 3566, 1044, 121, 5 },
  { 7099, 5415, 4164, 5741, 5771, 3469, 992, 112, 5 },
  { 7221, 5479, 4192, 5743, 5709, 3373, 943, 104, 4 },
  { 7342, 5542, 4220, 5743, 5646, 3279, 896, 96, 4 },
  { 7462, 5604, 4246, 5742, 5583, 3187, 851, 89, 4 },
  { 7584, 5665, 4272, 5739, 5518, 3097, 808, 82, 3 },
  { 7704, 5725, 4296, 5734, 5453, 3009, 768, 76, 3 },
  { 7825, 5784, 4318, 5727, 5386, 2924, 730, 71, 3 },
  { 7945, 5843, 4341, 5719, 5320, 2840, 693, 65, 2 },
  { 8066, 5900, 4361, 5709, 5252, 2758, 659, 61, 2 },
  { 8186, 5956, 4381, 5698, 5185, 2678, 626, 56, 2 },
  { 8306, 6011, 4400, 5685, 5117, 2600, 595, 52, 2 },
  { 8426, 6066, 4418, 5671, 5049, 2523, 565, 48, 2 },
  { 8547, 6119, 4434, 5655, 4981, 2449, 537, 45, 1 },
  { 8666, 6171, 4450, 5638, 4912, 2377, 511, 42, 1 },
  { 8786, 6223, 4465, 5620, 4843, 2306, 485, 39, 1 },
  { 8906, 6274, 4478, 5600, 4775, 2237, 461, 36, 1 },
  { 9025, 6323, 4491, 5580, 4706, 2170, 438, 34, 1 },
  { 9144, 6372, 4503, 5558, 4637, 2105, 417, 31, 1 },
  { 9264, 6420, 4514, 5535, 4568, 2041, 396, 29, 1 },
  { 9383, 6467, 4524, 5511, 4500, 1979, 376, 27, 1 },
  { 9502, 6513, 4532, 5486, 4432, 1919, 358, 25, 1 },
  { 9621, 6558, 4541, 5460, 4364, 1860, 340, 23, 1 },
  { 9740, 6602, 4548, 5433, 4296, 1803, 323, 22, 1 },
  { 9859, 6645, 4554, 5405, 4229, 1748, 307, 20, 1 },
  { 9978, 6688, 4559, 5376, 4161, 1694, 292, 19, 1 },
  { 10096, 6729, 4564, 5347, 4094, 1641, 278, 18, 1 },
  { 10215, 6770, 4568, 5316, 4028, 1590, 264, 16, 1 },
  { 10333, 6809, 4571, 5285, 3962, 1541, 251, 15, 1 },
  { 10452, 6848, 4573, 5253, 3896, 1492, 239, 14, 1 },
  { 10570, 6886, 4574, 5220, 3831, 1446, 227, 13, 1 },
  { 10688, 6923, 4575, 5186, 3767, 1400, 216, 12, 1 },
  { 10806, 6959, 4575, 5152, 3702, 1356, 205, 12, 1 },
  { 10924, 6994, 4574, 5117, 3639, 1313, 195, 11, 1 },
  { 11041, 7029, 4572, 5082, 3576, 1271, 186, 10, 1 },
  { 11159, 7062, 4570, 5046, 3513, 1231, 177, 9, 1 },
  { 11277, 7095, 4566, 5009, 3451, 1192, 168, 9, 1 },
  { 11394, 7127, 4563, 4972, 3390, 1153, 160, 8, 1 },
  { 11512, 7158, 4558, 4934, 3329, 1116, 152, 8, 1 },
  { 11629, 7188, 4553, 4896, 3269, 1080, 145, 7, 1 },
  { 11746, 7217, 4547, 4857, 3210, 1045, 138, 7, 1 },
  { 11864, 7245, 4540, 4818, 3151, 1012, 131, 6, 1 },
  { 11980, 7273, 4533, 4779, 3093, 979, 124, 6, 1 },
  { 12097, 7300, 4525, 4739, 3035, 947, 118, 6, 1 },
  { 12215, 7326, 4516, 4698, 2978, 916, 113, 5, 1 },
  { 12331, 7351, 4507, 4658, 2922, 886, 107, 5, 1 },
  { 12448, 7375, 4497, 4617, 2866, 857, 102, 5, 1 },
  { 12564, 7398, 4487, 4576, 2812, 829, 97, 4, 1 },
  { 12681, 7421, 4476, 4534, 2757, 802, 92, 4, 1 },
  { 12797, 7443, 4464, 4492, 2704, 775, 88, 4, 1 },
  { 12914, 7464, 4452, 4450, 2651, 749, 84, 3, 1 },
  { 13030, 7484, 4439, 4408, 2599, 725, 79, 3, 1 },
  { 13147, 7503, 4426, 4365, 2547, 700, 76, 3, 1 },
  { 13262, 7522, 4412, 4322, 2497, 677, 72, 3, 1 },
  { 13378, 7539, 4398, 4280, 2447, 654, 68, 3, 1 },
  { 13494, 7556, 4383, 4237, 2397, 632, 65, 3, 1 },
  { 13610, 7573, 4368, 4193, 2348, 611, 62, 2, 1 },
  { 13726, 7588, 4352, 4150, 2300, 590, 59, 2, 1 },
  { 13841, 7602, 4335, 4107, 2253, 571, 56, 2, 1 },
  { 13957, 7616, 4318, 4063, 2207, 551, 53, 2, 1 },
  { 14072, 7629, 4301, 4019, 2161, 532, 51, 2, 1 },
  { 14188, 7641, 4283, 3976, 2115, 514, 48, 2, 1 },
  { 14302, 7652, 4265, 3932, 2071, 497, 46, 2, 1 },
  { 14418, 7663, 4246, 3888, 2027, 480, 44, 1, 1 },
  { 14533, 7673, 4227, 3844, 1984, 463, 42, 1, 1 },
  { 14649, 7682, 4207, 3800, 1941, 447, 40, 1, 1 },
  { 14763, 7690, 4187, 3757, 1899, 432, 38, 1, 1 },
  { 14878, 7698, 4166, 3713, 1858, 417, 36, 1, 1 },
  { 14993, 7705, 4146, 3669, 1817, 402, 34, 1, 1 },
  { 15109, 7711, 4124, 3625, 1777, 388, 32, 1, 1 },
  { 15223, 7715, 4103, 3581, 1738, 375, 31, 1, 1 },
  { 15337, 7720, 4081, 3538, 1699, 362, 29, 1, 1 },
  { 15452, 7724, 4058, 3494, 1661, 349, 28, 1, 1 },
  { 15567, 7727, 4035, 3450, 1624, 337, 26, 1, 1 },
  { 15681, 7729, 4012, 3407, 1587, 325, 25, 1, 1 },
  { 15795, 7730, 3989, 3364, 1551, 313, 24, 1, 1 },
  { 15909, 7731, 3965, 3320, 1516, 302, 23, 1, 1 },
  { 16024, 7731, 3940, 3277, 1481, 291, 22, 1, 1 },
  { 16138, 7730, 3916, 3234, 1446, 281, 21, 1, 1 },
  { 16252, 7728, 3891, 3191, 1413, 271, 20, 1, 1 },
  { 16366, 7726, 3866, 3148, 1380, 261, 19, 1, 1 },
  { 16480, 7723, 3840, 3106, 1347, 252, 18, 1, 1 },
  { 16594, 7720, 3814, 3063, 1315, 243, 17, 1, 1 },
  { 16708, 7715, 3788, 3021, 1284, 234, 16, 1, 1 },
  { 16822, 7710, 3762, 2979, 1253, 225, 15, 1, 1 },
  { 16936, 7704, 3735, 2937, 1223, 217, 14, 1, 1 },
  { 17050, 7697, 3708, 2895, 1193, 209, 14, 1, 1 },
  { 17162, 7690, 3681, 2854, 1164, 202, 13, 1, 1 },
  { 17276, 7682, 3654, 2812, 1136, 194, 12, 1, 1 },
  { 17389, 7673, 3626, 2771, 1108, 187, 12, 1, 1 },
  { 17504, 7663, 3598, 2730, 1080, 180, 11, 1, 1 },
  { 17617, 7653, 3570, 2689, 1053, 173, 11, 1, 1 },
  { 17730, 7642, 3541, 2649, 1027, 167, 10, 1, 1 },
  { 17843, 7630, 3513, 2608, 1001, 161, 10, 1, 1 },
  { 17957, 7618, 3484, 2569, 975, 154, 9, 1, 1 },
  { 18069, 7605, 3455, 2529, 950, 149, 9, 1, 1 },
  { 18183, 7591, 3426, 2489, 926, 143, 8, 1, 1 },
  { 18296, 7576, 3396, 2450, 902, 138, 8, 1, 1 },
  { 18410, 7562, 3366, 2411, 878, 132, 7, 1, 1 },
  { 18523, 7545, 3337, 2372, 855, 127, 7, 1, 1 },
  { 18636, 7529, 3306, 2333, 833, 122, 7, 1, 1 },
  { 18749, 7511, 3276, 2295, 811, 118, 6, 1, 1 },
  { 18862, 7493, 3246, 2257, 789, 113, 6, 1, 1 },
  { 18975, 7474, 3215, 2219, 768, 109, 6, 1, 1 },
  { 19088, 7455, 3185, 2182, 747, 104, 5, 1, 1 },
  { 19201, 7435, 3154, 2144, 727, 100, 5, 1, 1 },
  { 19314, 7414, 3123, 2107, 707, 96, 5, 1, 1 },
  { 19427, 7392, 3092, 2071, 687, 92, 5, 1, 1 },
  { 19541, 7370, 3060, 2034, 668, 89, 4, 1, 1 },
  { 19654, 7347, 3029, 1998, 649, 85, 4, 1, 1 },
  { 19766, 7323, 2997, 1963, 631, 82, 4, 1, 1 },
  { 19878, 7299, 2966, 1927, 613, 79, 4, 1, 1 },
  { 19991, 7274, 2934, 1892, 596, 75, 4, 1, 1 },
  { 20105, 7248, 2902, 1857, 579, 72, 3, 1, 1 },
  { 20218, 7222, 2870, 1822, 562, 69, 3, 1, 1 },
  { 20331, 7195, 2838, 1788, 545, 66, 3, 1, 1 },
  { 20443, 7167, 2806, 1754, 529, 64, 3, 1, 1 },
  { 20556, 7138, 2774, 1720, 514, 61, 3, 1, 1 },
  { 20670, 7109, 2741, 1687, 498, 58, 3, 1, 1 },
  { 20783, 7079, 2709, 1654, 483, 56, 2, 1, 1 },
  { 20895, 7049, 2676, 1621, 469, 54, 2, 1, 1 },
  { 21008, 7017, 2644, 1589, 455, 51, 2, 1, 1 },
  { 21121, 6985, 2611, 1557, 441, 49, 2, 1, 1 },
  { 21234, 6953, 2578, 1525, 427, 47, 2, 1, 1 },
  { 21347, 6919, 2545, 1494, 414, 45, 2, 1, 1 },
  { 21460, 6885, 2513, 1462, 401, 43, 2, 1, 1 },
  { 21573, 6850, 2480, 1432, 388, 41, 2, 1, 1 },
  { 21687, 6815, 2447, 1401, 375, 39, 2, 1, 1 },
  { 21801, 6778, 2414, 1371, 363, 38, 1, 1, 1 },
  { 21914, 6741, 2381, 1341, 352, 36, 1, 1, 1 },
  { 22028, 6704, 2348, 1311, 340, 34, 1, 1, 1 },
  { 22141, 6665, 2315, 1282, 329, 33, 1, 1, 1 },
  { 22255, 6626, 2282, 1253, 318, 31, 1, 1, 1 },
  { 22368, 6586, 2249, 1225, 307, 30, 1, 1, 1 },
  { 22482, 6546, 2216, 1196, 297, 28, 1, 1, 1 },
  { 22595, 6505, 2183, 1169, 286, 27, 1, 1, 1 },
  { 22709, 6463, 2149, 1141, 277, 26, 1, 1, 1 },
  { 22823, 6420, 2116, 1114, 267, 25, 1, 1, 1 },
  { 22938, 6377, 2083, 1087, 257, 23, 1, 1, 1 },
  { 23053, 6332, 2050, 1060, 248, 22, 1, 1, 1 },
  { 23167, 6287, 2017, 1034, 239, 21, 1, 1, 1 },
  { 23280, 6242, 1984, 1008, 231, 20, 1, 1, 1 },
  { 23396, 6195, 1951, 982, 222, 19, 1, 1, 1 },
  { 23510, 6148, 1918, 957, 214, 18, 1, 1, 1 },
  { 23625, 6100, 1885, 932, 206, 17, 1, 1, 1 },
  { 23741, 6051, 1852, 907, 198, 16, 1, 1, 1 },
  { 23855, 6002, 1819, 883, 190, 16, 1, 1, 1 },
  { 23971, 5951, 1786, 859, 183, 15, 1, 1, 1 },
  { 24087, 5900, 1753, 835, 176, 14, 1, 1, 1 },
  { 24203, 5848, 1720, 812, 169, 13, 1, 1, 1 },
  { 24318, 5796, 1687, 789, 162, 13, 1, 1, 1 },
  { 24435, 5742, 1655, 766, 155, 12, 1, 1, 1 },
  { 24552, 5688, 1622, 743, 149, 11, 1, 1, 1 },
  { 24669, 5632, 1589, 721, 143, 11, 1, 1, 1 },
  { 24786, 5576, 1557, 699, 137, 10, 1, 1, 1 },
  { 24903, 5519, 1524, 678, 131, 10, 1, 1, 1 },
  { 25021, 5462, 1491, 657, 125, 9, 1, 1, 1 },
  { 25139, 5403, 1459, 636, 120, 8, 1, 1, 1 },
  { 25258, 5343, 1427, 615, 114, 8, 1, 1, 1 },
  { 25376, 5283, 1394, 595, 109, 8, 1, 1, 1 },
  { 25496, 5221, 1362, 575, 104, 7, 1, 1, 1 },
  { 25614, 5159, 1330, 556, 99, 7, 1, 1, 1 },
  { 25735, 5096, 1298, 536, 94, 6, 1, 1, 1 },
  { 25856, 5031, 1265, 517, 90, 6, 1, 1, 1 },
  { 25977, 4966, 1233, 499, 85, 5, 1, 1, 1 },
  { 26098, 4899, 1202, 480, 81, 5, 1, 1, 1 },
  { 26220, 4831, 1170, 462, 77, 5, 1, 1, 1 },
  { 26343, 4763, 1138, 444, 73, 4, 1, 1, 1 },
  { 26466, 4693, 1106, 427, 69, 4, 1, 1, 1 },
  { 26589, 4622, 1075, 410, 65, 4, 1, 1, 1 },
  { 26713, 4550, 1043, 393, 62, 4, 1, 1, 1 },
  { 26840, 4476, 1012, 376, 58, 3, 1, 1, 1 },
  { 26966, 4401, 980, 360, 55, 3, 1, 1, 1 },
  { 27092, 4325, 949, 344, 52, 3, 1, 1, 1 },
  { 27220, 4248, 918, 328, 48, 3, 1, 1, 1 },
  { 27350, 4169, 886, 313, 45, 2, 1, 1, 1 },
  { 27480, 4088, 855, 298, 42, 2, 1, 1, 1 },
  { 27610, 4006, 824, 283, 40, 2, 1, 1, 1 },
  { 27743, 3922, 793, 268, 37, 2, 1, 1, 1 },
  { 27876, 3837, 762, 254, 34, 2, 1, 1, 1 },
  { 28011, 3749, 731, 240, 32, 2, 1, 1, 1 },
  { 28147, 3659, 701, 227, 30, 1, 1, 1, 1 },
  { 28286, 3568, 670, 213, 27, 1, 1, 1, 1 },
  { 28426, 3474, 639, 200, 25, 1, 1, 1, 1 },
  { 28569, 3377, 608, 187, 23, 1, 1, 1, 1 },
  { 28714, 3278, 577, 174, 21, 1, 1, 1, 1 },
  { 28860, 3176, 547, 162, 19, 1, 1, 1, 1 },
  { 29010, 3071, 516, 150, 17, 1, 1, 1, 1 },
  { 29163, 2962, 485, 138, 16, 1, 1, 1, 1 },
  { 29320, 2849, 454, 127, 14, 1, 1, 1, 1 },
  { 29483, 2731, 423, 115, 12, 1, 1, 1, 1 },
  { 29650, 2608, 391, 104, 11, 1, 1, 1, 1 },
  { 29823, 2479, 360, 93, 9, 1, 1, 1, 1 },
  { 30002, 2343, 328, 83, 8, 1, 1, 1, 1 },
  { 30192, 2198, 295, 72, 7, 1, 1, 1, 1 },
  { 30393, 2041, 262, 62, 6, 1, 1, 1, 1 },
  { 30612, 1869, 227, 52, 4, 1, 1, 1, 1 },
  { 30853, 1676, 191, 41, 3, 1, 1, 1, 1 },
  { 31131, 1448, 152, 31, 2, 1, 1, 1, 1 },
  { 31486, 1150, 107, 20, 1, 1, 1, 1, 1 },
};
#elif CONFIG_EC_MULTISYMBOL
// Values for tokens ONE_TOKEN through CATEGORY6_TOKEN included here.
// ZERO_TOKEN and EOB_TOKEN are coded as flags outside this coder.
const aom_cdf_prob
    av1_pareto8_token_probs[COEFF_PROB_MODELS][ENTROPY_TOKENS - 2] = {
      { 128, 127, 127, 126, 251, 495, 965, 1832, 3305, 25412 },
      { 256, 254, 252, 249, 492, 959, 1820, 3283, 5365, 19838 },
      { 384, 379, 374, 369, 724, 1392, 2574, 4417, 6568, 15587 },
      { 512, 503, 494, 486, 946, 1795, 3238, 5289, 7184, 12321 },
      { 640, 626, 612, 599, 1159, 2172, 3818, 5946, 7401, 9795 },
      { 768, 748, 728, 709, 1363, 2522, 4324, 6424, 7352, 7830 },
      { 896, 869, 842, 816, 1559, 2847, 4762, 6755, 7131, 6291 },
      { 1024, 988, 954, 921, 1747, 3148, 5139, 6966, 6803, 5078 },
      { 1152, 1107, 1063, 1022, 1926, 3427, 5460, 7080, 6412, 4119 },
      { 1280, 1224, 1171, 1120, 2098, 3685, 5730, 7113, 5991, 3356 },
      { 1408, 1340, 1276, 1216, 2261, 3923, 5955, 7083, 5560, 2746 },
      { 1536, 1455, 1380, 1308, 2418, 4142, 6140, 7001, 5133, 2255 },
      { 1664, 1569, 1481, 1398, 2567, 4342, 6287, 6879, 4721, 1860 },
      { 1792, 1683, 1580, 1485, 2709, 4525, 6401, 6725, 4329, 1539 },
      { 1920, 1794, 1678, 1570, 2845, 4692, 6486, 6546, 3959, 1278 },
      { 2048, 1905, 1773, 1651, 2974, 4844, 6543, 6350, 3615, 1065 },
      { 2176, 2015, 1867, 1731, 3096, 4980, 6576, 6140, 3296, 891 },
      { 2304, 2123, 1958, 1807, 3212, 5104, 6589, 5922, 3002, 747 },
      { 2432, 2231, 2048, 1882, 3322, 5214, 6581, 5698, 2732, 628 },
      { 2560, 2337, 2136, 1953, 3427, 5311, 6557, 5472, 2485, 530 },
      { 2688, 2442, 2222, 2023, 3525, 5397, 6518, 5246, 2259, 448 },
      { 2816, 2547, 2306, 2090, 3618, 5472, 6465, 5021, 2053, 380 },
      { 2944, 2650, 2388, 2154, 3706, 5537, 6401, 4799, 1866, 323 },
      { 3072, 2752, 2468, 2217, 3788, 5591, 6327, 4581, 1696, 276 },
      { 3200, 2853, 2547, 2277, 3866, 5637, 6243, 4369, 1541, 235 },
      { 3328, 2952, 2624, 2335, 3938, 5673, 6152, 4163, 1401, 202 },
      { 3456, 3051, 2699, 2391, 4006, 5702, 6054, 3962, 1274, 173 },
      { 3584, 3149, 2772, 2444, 4070, 5723, 5950, 3769, 1158, 149 },
      { 3712, 3246, 2843, 2496, 4128, 5736, 5842, 3583, 1054, 128 },
      { 3840, 3341, 2913, 2545, 4183, 5743, 5729, 3404, 959, 111 },
      { 3968, 3436, 2981, 2593, 4233, 5743, 5614, 3232, 872, 96 },
      { 4096, 3529, 3048, 2638, 4280, 5737, 5496, 3067, 794, 83 },
      { 4224, 3621, 3113, 2682, 4322, 5726, 5375, 2909, 724, 72 },
      { 4352, 3712, 3176, 2724, 4361, 5709, 5253, 2759, 659, 63 },
      { 4480, 3803, 3237, 2764, 4396, 5687, 5130, 2615, 601, 55 },
      { 4608, 3892, 3297, 2801, 4428, 5661, 5007, 2478, 548, 48 },
      { 4736, 3980, 3355, 2838, 4456, 5631, 4883, 2347, 500, 42 },
      { 4864, 4067, 3412, 2872, 4481, 5596, 4760, 2223, 456, 37 },
      { 4992, 4152, 3467, 2905, 4503, 5558, 4637, 2105, 417, 32 },
      { 5120, 4237, 3521, 2936, 4521, 5516, 4515, 1993, 381, 28 },
      { 5248, 4321, 3573, 2966, 4537, 5471, 4393, 1886, 348, 25 },
      { 5376, 4404, 3623, 2993, 4550, 5424, 4273, 1785, 318, 22 },
      { 5504, 4486, 3672, 3020, 4560, 5373, 4155, 1688, 291, 19 },
      { 5632, 4566, 3720, 3044, 4568, 5321, 4037, 1597, 266, 17 },
      { 5760, 4646, 3766, 3067, 4572, 5265, 3922, 1511, 244, 15 },
      { 5888, 4724, 3811, 3089, 4575, 5208, 3808, 1429, 223, 13 },
      { 6016, 4802, 3854, 3109, 4575, 5148, 3696, 1352, 204, 12 },
      { 6144, 4878, 3895, 3128, 4573, 5088, 3587, 1278, 187, 10 },
      { 6272, 4953, 3936, 3145, 4568, 5025, 3479, 1209, 172, 9 },
      { 6400, 5028, 3975, 3161, 4561, 4961, 3373, 1143, 158, 8 },
      { 6528, 5101, 4012, 3175, 4553, 4896, 3270, 1081, 145, 7 },
      { 6656, 5173, 4048, 3189, 4542, 4830, 3168, 1022, 133, 7 },
      { 6784, 5244, 4083, 3201, 4530, 4763, 3069, 966, 122, 6 },
      { 6912, 5314, 4117, 3212, 4516, 4694, 2973, 913, 112, 5 },
      { 7040, 5383, 4149, 3221, 4500, 4626, 2878, 863, 103, 5 },
      { 7168, 5452, 4180, 3229, 4482, 4556, 2786, 816, 95, 4 },
      { 7296, 5519, 4210, 3236, 4463, 4486, 2696, 771, 87, 4 },
      { 7424, 5585, 4238, 3242, 4442, 4416, 2609, 729, 80, 3 },
      { 7552, 5650, 4265, 3247, 4420, 4345, 2523, 689, 74, 3 },
      { 7680, 5714, 4291, 3251, 4396, 4274, 2440, 651, 68, 3 },
      { 7808, 5777, 4315, 3254, 4371, 4203, 2359, 616, 63, 2 },
      { 7936, 5838, 4339, 3255, 4345, 4132, 2281, 582, 58, 2 },
      { 8064, 5899, 4361, 3256, 4318, 4061, 2204, 550, 53, 2 },
      { 8192, 5959, 4382, 3255, 4289, 3990, 2130, 520, 49, 2 },
      { 8320, 6018, 4402, 3254, 4259, 3919, 2057, 492, 45, 2 },
      { 8448, 6075, 4421, 3252, 4229, 3848, 1987, 465, 42, 1 },
      { 8576, 6133, 4438, 3248, 4197, 3778, 1919, 439, 39, 1 },
      { 8704, 6188, 4455, 3244, 4164, 3708, 1853, 415, 36, 1 },
      { 8832, 6243, 4470, 3239, 4131, 3638, 1789, 392, 33, 1 },
      { 8960, 6297, 4484, 3233, 4096, 3569, 1727, 371, 30, 1 },
      { 9088, 6349, 4497, 3226, 4061, 3500, 1667, 351, 28, 1 },
      { 9216, 6401, 4509, 3219, 4025, 3432, 1608, 331, 26, 1 },
      { 9344, 6452, 4520, 3210, 3989, 3364, 1551, 313, 24, 1 },
      { 9472, 6501, 4530, 3201, 3952, 3297, 1496, 296, 22, 1 },
      { 9600, 6550, 4539, 3191, 3914, 3230, 1443, 280, 20, 1 },
      { 9728, 6597, 4547, 3180, 3875, 3164, 1392, 265, 19, 1 },
      { 9856, 6644, 4554, 3169, 3836, 3098, 1342, 250, 18, 1 },
      { 9984, 6690, 4560, 3157, 3796, 3034, 1293, 237, 16, 1 },
      { 10112, 6734, 4565, 3144, 3756, 2970, 1247, 224, 15, 1 },
      { 10240, 6778, 4568, 3131, 3716, 2907, 1202, 211, 14, 1 },
      { 10368, 6821, 4571, 3117, 3675, 2844, 1158, 200, 13, 1 },
      { 10496, 6862, 4573, 3102, 3634, 2783, 1116, 189, 12, 1 },
      { 10624, 6903, 4574, 3087, 3592, 2722, 1075, 179, 11, 1 },
      { 10752, 6942, 4575, 3071, 3551, 2662, 1035, 169, 10, 1 },
      { 10880, 6981, 4574, 3054, 3508, 2603, 997, 160, 10, 1 },
      { 11008, 7019, 4572, 3038, 3466, 2544, 960, 151, 9, 1 },
      { 11136, 7055, 4570, 3020, 3424, 2487, 924, 143, 8, 1 },
      { 11264, 7091, 4566, 3002, 3381, 2430, 890, 135, 8, 1 },
      { 11392, 7126, 4563, 2984, 3338, 2374, 856, 127, 7, 1 },
      { 11520, 7159, 4557, 2965, 3295, 2319, 824, 121, 7, 1 },
      { 11648, 7193, 4552, 2945, 3252, 2264, 793, 114, 6, 1 },
      { 11776, 7224, 4545, 2925, 3209, 2211, 763, 108, 6, 1 },
      { 11904, 7255, 4538, 2905, 3165, 2159, 734, 102, 5, 1 },
      { 12032, 7285, 4530, 2884, 3122, 2107, 706, 96, 5, 1 },
      { 12160, 7314, 4520, 2863, 3079, 2056, 679, 91, 5, 1 },
      { 12288, 7341, 4511, 2842, 3036, 2006, 653, 86, 4, 1 },
      { 12416, 7368, 4500, 2820, 2993, 1957, 628, 81, 4, 1 },
      { 12544, 7394, 4489, 2797, 2949, 1909, 604, 77, 4, 1 },
      { 12672, 7419, 4477, 2775, 2906, 1861, 581, 73, 3, 1 },
      { 12800, 7443, 4464, 2752, 2863, 1815, 558, 69, 3, 1 },
      { 12928, 7466, 4451, 2729, 2820, 1769, 536, 65, 3, 1 },
      { 13056, 7488, 4437, 2705, 2777, 1724, 516, 61, 3, 1 },
      { 13184, 7509, 4422, 2682, 2734, 1680, 495, 58, 3, 1 },
      { 13312, 7529, 4406, 2658, 2692, 1637, 476, 55, 2, 1 },
      { 13440, 7548, 4390, 2633, 2650, 1595, 457, 52, 2, 1 },
      { 13568, 7567, 4373, 2609, 2607, 1553, 439, 49, 2, 1 },
      { 13696, 7583, 4356, 2584, 2565, 1513, 422, 46, 2, 1 },
      { 13824, 7600, 4337, 2559, 2523, 1473, 405, 44, 2, 1 },
      { 13952, 7615, 4319, 2533, 2482, 1434, 389, 41, 2, 1 },
      { 14080, 7629, 4300, 2508, 2441, 1395, 373, 39, 2, 1 },
      { 14208, 7643, 4280, 2482, 2400, 1358, 358, 37, 1, 1 },
      { 14336, 7655, 4259, 2457, 2359, 1321, 344, 35, 1, 1 },
      { 14464, 7667, 4238, 2431, 2318, 1285, 330, 33, 1, 1 },
      { 14592, 7677, 4217, 2405, 2278, 1250, 316, 31, 1, 1 },
      { 14720, 7687, 4195, 2378, 2238, 1215, 304, 29, 1, 1 },
      { 14848, 7696, 4172, 2352, 2198, 1181, 291, 28, 1, 1 },
      { 14976, 7703, 4149, 2326, 2159, 1148, 279, 26, 1, 1 },
      { 15104, 7710, 4125, 2299, 2119, 1116, 268, 25, 1, 1 },
      { 15232, 7715, 4101, 2272, 2081, 1085, 257, 23, 1, 1 },
      { 15360, 7721, 4076, 2245, 2042, 1054, 246, 22, 1, 1 },
      { 15488, 7724, 4051, 2219, 2004, 1023, 236, 21, 1, 1 },
      { 15616, 7727, 4025, 2192, 1966, 994, 226, 20, 1, 1 },
      { 15744, 7729, 3999, 2164, 1929, 965, 217, 19, 1, 1 },
      { 15872, 7731, 3972, 2137, 1892, 937, 207, 18, 1, 1 },
      { 16000, 7731, 3945, 2110, 1855, 909, 199, 17, 1, 1 },
      { 16128, 7730, 3918, 2083, 1819, 882, 190, 16, 1, 1 },
      { 16256, 7728, 3890, 2056, 1783, 856, 182, 15, 1, 1 },
      { 16384, 7725, 3862, 2029, 1747, 831, 174, 14, 1, 1 },
      { 16512, 7721, 3833, 2002, 1712, 806, 167, 13, 1, 1 },
      { 16640, 7717, 3804, 1975, 1677, 781, 160, 12, 1, 1 },
      { 16768, 7712, 3775, 1947, 1642, 757, 153, 12, 1, 1 },
      { 16896, 7706, 3745, 1920, 1608, 734, 146, 11, 1, 1 },
      { 17024, 7699, 3714, 1893, 1575, 711, 140, 10, 1, 1 },
      { 17152, 7690, 3684, 1866, 1541, 689, 134, 10, 1, 1 },
      { 17280, 7681, 3653, 1839, 1508, 668, 128, 9, 1, 1 },
      { 17408, 7671, 3621, 1812, 1476, 647, 122, 9, 1, 1 },
      { 17536, 7660, 3590, 1785, 1444, 626, 117, 8, 1, 1 },
      { 17664, 7648, 3558, 1758, 1412, 606, 112, 8, 1, 1 },
      { 17792, 7635, 3526, 1731, 1381, 587, 107, 7, 1, 1 },
      { 17920, 7622, 3493, 1704, 1350, 568, 102, 7, 1, 1 },
      { 18048, 7607, 3461, 1678, 1319, 549, 98, 6, 1, 1 },
      { 18176, 7592, 3428, 1651, 1289, 531, 93, 6, 1, 1 },
      { 18304, 7575, 3394, 1625, 1259, 514, 89, 6, 1, 1 },
      { 18432, 7558, 3361, 1598, 1230, 497, 85, 5, 1, 1 },
      { 18560, 7540, 3327, 1572, 1201, 480, 81, 5, 1, 1 },
      { 18688, 7520, 3293, 1546, 1173, 464, 77, 5, 1, 1 },
      { 18816, 7500, 3258, 1520, 1145, 448, 74, 5, 1, 1 },
      { 18944, 7480, 3224, 1494, 1117, 433, 70, 4, 1, 1 },
      { 19072, 7458, 3189, 1468, 1090, 418, 67, 4, 1, 1 },
      { 19200, 7435, 3154, 1442, 1063, 404, 64, 4, 1, 1 },
      { 19328, 7410, 3119, 1417, 1037, 390, 61, 4, 1, 1 },
      { 19456, 7386, 3084, 1392, 1011, 376, 58, 3, 1, 1 },
      { 19584, 7361, 3048, 1366, 986, 363, 55, 3, 1, 1 },
      { 19712, 7335, 3012, 1341, 960, 350, 53, 3, 1, 1 },
      { 19840, 7307, 2977, 1316, 936, 337, 50, 3, 1, 1 },
      { 19968, 7279, 2941, 1291, 911, 325, 48, 3, 1, 1 },
      { 20096, 7251, 2905, 1267, 887, 313, 45, 2, 1, 1 },
      { 20224, 7220, 2868, 1243, 864, 302, 43, 2, 1, 1 },
      { 20352, 7189, 2832, 1218, 841, 291, 41, 2, 1, 1 },
      { 20480, 7158, 2795, 1194, 818, 280, 39, 2, 1, 1 },
      { 20608, 7124, 2759, 1170, 796, 270, 37, 2, 1, 1 },
      { 20736, 7091, 2722, 1147, 774, 259, 35, 2, 1, 1 },
      { 20864, 7056, 2685, 1123, 752, 250, 34, 2, 1, 1 },
      { 20992, 7021, 2648, 1100, 731, 240, 32, 2, 1, 1 },
      { 21120, 6985, 2612, 1077, 710, 231, 30, 1, 1, 1 },
      { 21248, 6948, 2574, 1054, 690, 222, 29, 1, 1, 1 },
      { 21376, 6911, 2537, 1031, 670, 213, 27, 1, 1, 1 },
      { 21504, 6872, 2500, 1008, 650, 205, 26, 1, 1, 1 },
      { 21632, 6831, 2463, 986, 631, 197, 25, 1, 1, 1 },
      { 21760, 6791, 2426, 964, 612, 189, 23, 1, 1, 1 },
      { 21888, 6749, 2389, 942, 594, 181, 22, 1, 1, 1 },
      { 22016, 6707, 2351, 921, 575, 174, 21, 1, 1, 1 },
      { 22144, 6663, 2314, 899, 558, 167, 20, 1, 1, 1 },
      { 22272, 6619, 2277, 878, 540, 160, 19, 1, 1, 1 },
      { 22400, 6574, 2240, 857, 523, 153, 18, 1, 1, 1 },
      { 22528, 6529, 2202, 836, 507, 146, 17, 1, 1, 1 },
      { 22656, 6482, 2165, 816, 490, 140, 16, 1, 1, 1 },
      { 22784, 6435, 2128, 795, 474, 134, 15, 1, 1, 1 },
      { 22912, 6386, 2091, 775, 459, 128, 14, 1, 1, 1 },
      { 23040, 6336, 2054, 756, 443, 123, 13, 1, 1, 1 },
      { 23168, 6286, 2017, 736, 428, 117, 13, 1, 1, 1 },
      { 23296, 6234, 1980, 717, 414, 112, 12, 1, 1, 1 },
      { 23424, 6183, 1943, 698, 399, 107, 11, 1, 1, 1 },
      { 23552, 6130, 1906, 679, 385, 102, 11, 1, 1, 1 },
      { 23680, 6077, 1869, 660, 372, 97, 10, 1, 1, 1 },
      { 23808, 6022, 1833, 642, 358, 93, 9, 1, 1, 1 },
      { 23936, 5966, 1796, 624, 345, 89, 9, 1, 1, 1 },
      { 24064, 5910, 1760, 606, 333, 84, 8, 1, 1, 1 },
      { 24192, 5853, 1724, 588, 320, 80, 8, 1, 1, 1 },
      { 24320, 5796, 1687, 571, 308, 76, 7, 1, 1, 1 },
      { 24448, 5735, 1651, 554, 297, 73, 7, 1, 1, 1 },
      { 24576, 5677, 1615, 537, 285, 69, 6, 1, 1, 1 },
      { 24704, 5615, 1579, 521, 274, 66, 6, 1, 1, 1 },
      { 24832, 5554, 1544, 504, 263, 62, 6, 1, 1, 1 },
      { 24960, 5492, 1508, 488, 253, 59, 5, 1, 1, 1 },
      { 25088, 5428, 1473, 473, 242, 56, 5, 1, 1, 1 },
      { 25216, 5364, 1438, 457, 232, 53, 5, 1, 1, 1 },
      { 25344, 5300, 1403, 442, 222, 50, 4, 1, 1, 1 },
      { 25472, 5233, 1368, 427, 213, 48, 4, 1, 1, 1 },
      { 25600, 5166, 1334, 412, 204, 45, 4, 1, 1, 1 },
      { 25728, 5098, 1299, 398, 195, 43, 4, 1, 1, 1 },
      { 25856, 5030, 1266, 384, 186, 40, 3, 1, 1, 1 },
      { 25984, 4960, 1232, 370, 178, 38, 3, 1, 1, 1 },
      { 26112, 4890, 1198, 356, 170, 36, 3, 1, 1, 1 },
      { 26240, 4819, 1164, 343, 162, 34, 3, 1, 1, 1 },
      { 26368, 4748, 1132, 329, 154, 32, 2, 1, 1, 1 },
      { 26496, 4675, 1098, 317, 147, 30, 2, 1, 1, 1 },
      { 26624, 4602, 1066, 304, 139, 28, 2, 1, 1, 1 },
      { 26752, 4527, 1034, 292, 132, 26, 2, 1, 1, 1 },
      { 26880, 4451, 1001, 280, 126, 25, 2, 1, 1, 1 },
      { 27008, 4375, 970, 268, 119, 23, 2, 1, 1, 1 },
      { 27136, 4299, 938, 256, 113, 21, 2, 1, 1, 1 },
      { 27264, 4221, 907, 245, 107, 20, 1, 1, 1, 1 },
      { 27392, 4142, 876, 234, 101, 19, 1, 1, 1, 1 },
      { 27520, 4063, 846, 223, 95, 17, 1, 1, 1, 1 },
      { 27648, 3982, 815, 213, 90, 16, 1, 1, 1, 1 },
      { 27776, 3900, 786, 202, 85, 15, 1, 1, 1, 1 },
      { 27904, 3818, 756, 192, 80, 14, 1, 1, 1, 1 },
      { 28032, 3734, 727, 183, 75, 13, 1, 1, 1, 1 },
      { 28160, 3651, 698, 173, 70, 12, 1, 1, 1, 1 },
      { 28288, 3566, 669, 164, 66, 11, 1, 1, 1, 1 },
      { 28416, 3481, 641, 155, 61, 10, 1, 1, 1, 1 },
      { 28544, 3393, 614, 147, 57, 9, 1, 1, 1, 1 },
      { 28672, 3306, 586, 138, 53, 9, 1, 1, 1, 1 },
      { 28800, 3217, 559, 130, 50, 8, 1, 1, 1, 1 },
      { 28928, 3128, 533, 122, 46, 7, 1, 1, 1, 1 },
      { 29056, 3037, 507, 114, 43, 7, 1, 1, 1, 1 },
      { 29184, 2947, 481, 107, 39, 6, 1, 1, 1, 1 },
      { 29312, 2855, 456, 100, 36, 5, 1, 1, 1, 1 },
      { 29440, 2762, 431, 93, 33, 5, 1, 1, 1, 1 },
      { 29568, 2668, 407, 86, 31, 4, 1, 1, 1, 1 },
      { 29696, 2573, 383, 80, 28, 4, 1, 1, 1, 1 },
      { 29824, 2478, 359, 74, 25, 4, 1, 1, 1, 1 },
      { 29952, 2381, 337, 68, 23, 3, 1, 1, 1, 1 },
      { 30080, 2284, 314, 62, 21, 3, 1, 1, 1, 1 },
      { 30208, 2185, 293, 57, 19, 2, 1, 1, 1, 1 },
      { 30336, 2086, 271, 52, 17, 2, 1, 1, 1, 1 },
      { 30464, 1986, 250, 47, 15, 2, 1, 1, 1, 1 },
      { 30592, 1885, 230, 42, 13, 2, 1, 1, 1, 1 },
      { 30720, 1782, 211, 38, 12, 1, 1, 1, 1, 1 },
      { 30848, 1679, 192, 34, 10, 1, 1, 1, 1, 1 },
      { 30976, 1575, 173, 30, 9, 1, 1, 1, 1, 1 },
      { 31104, 1469, 156, 26, 8, 1, 1, 1, 1, 1 },
      { 31232, 1364, 138, 23, 6, 1, 1, 1, 1, 1 },
      { 31360, 1257, 122, 19, 5, 1, 1, 1, 1, 1 },
      { 31488, 1149, 106, 16, 4, 1, 1, 1, 1, 1 },
      { 31616, 1038, 91, 14, 4, 1, 1, 1, 1, 1 },
      { 31744, 928, 77, 11, 3, 1, 1, 1, 1, 1 },
      { 31872, 816, 64, 9, 2, 1, 1, 1, 1, 1 },
      { 32000, 703, 51, 7, 2, 1, 1, 1, 1, 1 },
      { 32128, 589, 40, 5, 1, 1, 1, 1, 1, 1 },
      { 32256, 473, 29, 4, 1, 1, 1, 1, 1, 1 },
      { 32384, 357, 19, 2, 1, 1, 1, 1, 1, 1 },
      { 32512, 238, 11, 1, 1, 1, 1, 1, 1, 1 },
      { 32640, 117, 4, 1, 1, 1, 1, 1, 1, 1 },
    };
#endif  // CONFIG_NEW_TOKENSET

/* clang-format off */
#if CONFIG_Q_ADAPT_PROBS
const av1_coeff_probs_model
default_qctx_coef_probs[QCTX_BINS][TX_SIZES][PLANE_TYPES] = {
    {  // Q_Index 0
#if CONFIG_CB4X4
        {  // TX_SIZE 0
            {  // Y plane
                {  // Intra
                    {  // band 0
                        {182,  34, 137}, { 79,  39, 103}, { 10,  28,  51},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 45,  88, 147}, { 46,  80, 140}, { 25,  69, 119},
                        { 12,  57,  96}, {  4,  41,  65}, {  1,  20,  31},
                    },
                    {  // band 2
                        { 58, 124, 190}, { 39, 106, 178}, { 16,  86, 147},
                        {  7,  69, 114}, {  3,  50,  80}, {  1,  25,  42},
                    },
                    {  // band 3
                        { 90, 138, 215}, { 54, 116, 198}, { 18,  86, 155},
                        {  5,  62, 112}, {  1,  38,  68}, {  1,  17,  30},
                    },
                    {  // band 4
                        {126, 149, 231}, { 82, 114, 211}, { 21,  80, 157},
                        {  6,  56, 105}, {  1,  36,  64}, {  1,  17,  31},
                    },
                    {  // band 5
                        {171,  56, 236}, {140,  54, 219}, { 57,  45, 167},
                        { 26,  36, 113}, { 11,  29,  72}, {  3,  18,  39},
                    },
                },
                {  // Intra
                    {  // band 0
                        {153, 122, 186}, {106, 109, 171}, { 36,  84, 128},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 27, 151, 201}, { 34, 131, 199}, { 23, 102, 161},
                        { 10,  80, 120}, {  4,  52,  78}, {  1,  24,  37},
                    },
                    {  // band 2
                        { 43, 158, 213}, { 35, 133, 203}, {  8,  92, 151},
                        {  2,  64, 106}, {  1,  36,  60}, {  1,  13,  24},
                    },
                    {  // band 3
                        { 68, 167, 223}, { 36, 135, 211}, {  9,  94, 157},
                        {  2,  67, 112}, {  1,  40,  68}, {  1,  17,  31},
                    },
                    {  // band 4
                        {131, 146, 237}, { 72, 119, 223}, { 17,  82, 164},
                        {  4,  55, 107}, {  1,  34,  63}, {  1,  16,  29},
                    },
                    {  // band 5
                        {184,  68, 244}, {153,  59, 232}, { 68,  51, 179},
                        { 31,  40, 123}, { 13,  29,  77}, {  4,  17,  37},
                    },
                },
            },
            {  // UV plane
                {  // Inter
                    {  // band 0
                        {203,  41, 203}, {127,  56, 174}, { 49,  56, 127},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {110, 121, 217}, {119, 113, 213}, { 64,  95, 185},
                        { 30,  72, 144}, {  8,  42,  76}, {  2,  17,  25},
                    },
                    {  // band 2
                        {127, 159, 229}, {115, 134, 223}, { 36, 100, 189},
                        { 11,  75, 142}, {  3,  48,  83}, {  1,  19,  33},
                    },
                    {  // band 3
                        {150, 172, 241}, { 90, 133, 231}, { 28, 102, 192},
                        {  7,  81, 147}, {  1,  53,  91}, {  1,  25,  42},
                    },
                    {  // band 4
                        {184, 144, 248}, {114, 117, 237}, { 37,  89, 192},
                        { 10,  63, 130}, {  4,  42,  76}, {  1,  19,  38},
                    },
                    {  // band 5
                        {207,  79, 250}, {179,  74, 241}, { 83,  67, 199},
                        { 38,  51, 142}, { 17,  37,  97}, { 10,  14,  55},
                    },
                },
                {  // Inter
                    {  // band 0
                        {220,  82, 232}, {150,  93, 214}, { 66,  95, 177},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {116, 160, 227}, {136, 141, 227}, { 67, 114, 190},
                        { 40,  94, 148}, { 21,  70, 107}, { 10,  43,  63},
                    },
                    {  // band 2
                        {124, 173, 235}, {105, 147, 226}, { 27, 107, 184},
                        { 10,  80, 142}, {  3,  50,  86}, {  1,  16,  32},
                    },
                    {  // band 3
                        {149, 179, 243}, { 89, 147, 234}, { 29, 112, 193},
                        {  9,  94, 157}, {  1,  64, 111}, {  1,  25,  43},
                    },
                    {  // band 4
                        {187, 153, 248}, {127, 130, 241}, { 52,  99, 202},
                        { 20,  79, 152}, {  4,  50,  93}, {  1,  19,  32},
                    },
                    {  // band 5
                        {215,  82, 251}, {195,  80, 246}, { 93,  70, 204},
                        { 39,  54, 147}, { 14,  33,  88}, {  6,  14,  39},
                    },
                },
            },
        },
#endif
        {  // TX_SIZE 0
            {  // Y plane
                {  // Intra
                    {  // band 0
                        {182,  34, 137}, { 79,  39, 103}, { 10,  28,  51},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 45,  88, 147}, { 46,  80, 140}, { 25,  69, 119},
                        { 12,  57,  96}, {  4,  41,  65}, {  1,  20,  31},
                    },
                    {  // band 2
                        { 58, 124, 190}, { 39, 106, 178}, { 16,  86, 147},
                        {  7,  69, 114}, {  3,  50,  80}, {  1,  25,  42},
                    },
                    {  // band 3
                        { 90, 138, 215}, { 54, 116, 198}, { 18,  86, 155},
                        {  5,  62, 112}, {  1,  38,  68}, {  1,  17,  30},
                    },
                    {  // band 4
                        {126, 149, 231}, { 82, 114, 211}, { 21,  80, 157},
                        {  6,  56, 105}, {  1,  36,  64}, {  1,  17,  31},
                    },
                    {  // band 5
                        {171,  56, 236}, {140,  54, 219}, { 57,  45, 167},
                        { 26,  36, 113}, { 11,  29,  72}, {  3,  18,  39},
                    },
                },
                {  // Intra
                    {  // band 0
                        {153, 122, 186}, {106, 109, 171}, { 36,  84, 128},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 27, 151, 201}, { 34, 131, 199}, { 23, 102, 161},
                        { 10,  80, 120}, {  4,  52,  78}, {  1,  24,  37},
                    },
                    {  // band 2
                        { 43, 158, 213}, { 35, 133, 203}, {  8,  92, 151},
                        {  2,  64, 106}, {  1,  36,  60}, {  1,  13,  24},
                    },
                    {  // band 3
                        { 68, 167, 223}, { 36, 135, 211}, {  9,  94, 157},
                        {  2,  67, 112}, {  1,  40,  68}, {  1,  17,  31},
                    },
                    {  // band 4
                        {131, 146, 237}, { 72, 119, 223}, { 17,  82, 164},
                        {  4,  55, 107}, {  1,  34,  63}, {  1,  16,  29},
                    },
                    {  // band 5
                        {184,  68, 244}, {153,  59, 232}, { 68,  51, 179},
                        { 31,  40, 123}, { 13,  29,  77}, {  4,  17,  37},
                    },
                },
            },
            {  // UV plane
                {  // Inter
                    {  // band 0
                        {203,  41, 203}, {127,  56, 174}, { 49,  56, 127},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {110, 121, 217}, {119, 113, 213}, { 64,  95, 185},
                        { 30,  72, 144}, {  8,  42,  76}, {  2,  17,  25},
                    },
                    {  // band 2
                        {127, 159, 229}, {115, 134, 223}, { 36, 100, 189},
                        { 11,  75, 142}, {  3,  48,  83}, {  1,  19,  33},
                    },
                    {  // band 3
                        {150, 172, 241}, { 90, 133, 231}, { 28, 102, 192},
                        {  7,  81, 147}, {  1,  53,  91}, {  1,  25,  42},
                    },
                    {  // band 4
                        {184, 144, 248}, {114, 117, 237}, { 37,  89, 192},
                        { 10,  63, 130}, {  4,  42,  76}, {  1,  19,  38},
                    },
                    {  // band 5
                        {207,  79, 250}, {179,  74, 241}, { 83,  67, 199},
                        { 38,  51, 142}, { 17,  37,  97}, { 10,  14,  55},
                    },
                },
                {  // Inter
                    {  // band 0
                        {220,  82, 232}, {150,  93, 214}, { 66,  95, 177},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {116, 160, 227}, {136, 141, 227}, { 67, 114, 190},
                        { 40,  94, 148}, { 21,  70, 107}, { 10,  43,  63},
                    },
                    {  // band 2
                        {124, 173, 235}, {105, 147, 226}, { 27, 107, 184},
                        { 10,  80, 142}, {  3,  50,  86}, {  1,  16,  32},
                    },
                    {  // band 3
                        {149, 179, 243}, { 89, 147, 234}, { 29, 112, 193},
                        {  9,  94, 157}, {  1,  64, 111}, {  1,  25,  43},
                    },
                    {  // band 4
                        {187, 153, 248}, {127, 130, 241}, { 52,  99, 202},
                        { 20,  79, 152}, {  4,  50,  93}, {  1,  19,  32},
                    },
                    {  // band 5
                        {215,  82, 251}, {195,  80, 246}, { 93,  70, 204},
                        { 39,  54, 147}, { 14,  33,  88}, {  6,  14,  39},
                    },
                },
            },
        },
        {  // TX_SIZE 1
            {  // Y plane
                {  // Intra
                    {  // band 0
                        {116,  43, 131}, { 39,  41,  94}, {  4,  28,  47},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 28, 101, 141}, { 27,  95, 140}, { 18,  80, 121},
                        { 10,  61,  95}, {  4,  39,  60}, {  1,  19,  26},
                    },
                    {  // band 2
                        { 29, 150, 183}, { 19, 127, 175}, {  8,  98, 147},
                        {  3,  76, 115}, {  1,  55,  84}, {  1,  29,  43},
                    },
                    {  // band 3
                        { 26, 168, 202}, { 12, 138, 188}, {  2,  98, 149},
                        {  1,  69, 110}, {  1,  40,  65}, {  1,  17,  25},
                    },
                    {  // band 4
                        { 33, 188, 225}, { 12, 155, 207}, {  2, 101, 155},
                        {  1,  65, 106}, {  1,  36,  60}, {  1,  18,  26},
                    },
                    {  // band 5
                        { 79, 205, 242}, { 30, 168, 224}, {  5, 106, 164},
                        {  1,  68, 110}, {  1,  39,  65}, {  1,  18,  28},
                    },
                },
                {  // Intra
                    {  // band 0
                        { 96,  80, 201}, { 51,  88, 168}, { 14,  78, 116},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {  6, 167, 216}, { 32, 152, 211}, { 24, 121, 182},
                        { 13,  98, 149}, { 12,  76, 108}, {  8,  48,  62},
                    },
                    {  // band 2
                        { 17, 176, 225}, { 13, 147, 209}, {  3,  96, 155},
                        {  1,  65, 108}, {  2,  43,  63}, {  2,  23,  25},
                    },
                    {  // band 3
                        { 18, 183, 232}, { 10, 153, 214}, {  1,  96, 154},
                        {  1,  63, 105}, {  1,  39,  59}, {  1,  21,  24},
                    },
                    {  // band 4
                        { 23, 191, 239}, {  8, 159, 221}, {  1,  97, 158},
                        {  1,  61, 105}, {  1,  37,  60}, {  1,  20,  26},
                    },
                    {  // band 5
                        { 70, 201, 243}, { 29, 163, 228}, {  4, 102, 169},
                        {  1,  67, 114}, {  1,  39,  66}, {  1,  17,  29},
                    },
                },
            },
            {  // UV plane
                {  // Inter
                    {  // band 0
                        {181,  38, 192}, { 95,  47, 151}, { 29,  49, 102},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 72, 131, 202}, { 93, 120, 205}, { 50, 103, 179},
                        { 24,  79, 143}, { 11,  47,  78}, {  7,  19,  25},
                    },
                    {  // band 2
                        { 84, 176, 221}, { 56, 144, 214}, { 21, 108, 182},
                        {  8,  83, 139}, {  3,  55,  90}, {  2,  27,  41},
                    },
                    {  // band 3
                        { 84, 195, 234}, { 42, 156, 222}, { 10, 109, 180},
                        {  4,  77, 133}, {  1,  48,  80}, {  1,  23,  35},
                    },
                    {  // band 4
                        { 89, 210, 238}, { 35, 165, 221}, {  6, 106, 172},
                        {  2,  70, 123}, {  1,  44,  74}, {  1,  21,  30},
                    },
                    {  // band 5
                        {114, 221, 247}, { 49, 170, 234}, {  7, 113, 184},
                        {  2,  77, 132}, {  1,  48,  79}, {  1,  25,  33},
                    },
                },
                {  // Inter
                    {  // band 0
                        {192,  66, 237}, {113,  84, 211}, { 35,  84, 154},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 81, 180, 234}, {127, 165, 229}, { 58, 137, 204},
                        { 41, 114, 174}, { 44,  94, 136}, { 29,  66,  86},
                    },
                    {  // band 2
                        { 82, 193, 240}, { 39, 162, 223}, {  8, 113, 179},
                        {  3,  83, 136}, {  6,  62,  84}, {  5,  45,  45},
                    },
                    {  // band 3
                        { 78, 203, 242}, { 31, 170, 227}, {  4, 115, 181},
                        {  1,  82, 135}, {  2,  59,  82}, {  1,  45,  47},
                    },
                    {  // band 4
                        { 76, 210, 239}, { 25, 170, 213}, {  2,  99, 152},
                        {  1,  69, 115}, {  1,  49,  80}, {  1,  47,  57},
                    },
                    {  // band 5
                        {103, 217, 250}, { 42, 180, 237}, {  3, 124, 191},
                        {  1,  90, 150}, {  1,  69, 116}, {  1,  52,  46},
                    },
                },
            },
        },
        {  // TX_SIZE 2
            {  // Y plane
                {  // Intra
                    {  // band 0
                        { 58,  38,  99}, {  9,  26,  51}, {  1,  14,  22},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 14,  78, 109}, { 16,  73, 105}, { 11,  62,  92},
                        {  6,  47,  72}, {  2,  29,  45}, {  1,  12,  18},
                    },
                    {  // band 2
                        { 17, 131, 148}, { 11, 112, 140}, {  5,  87, 118},
                        {  2,  63,  90}, {  1,  42,  63}, {  1,  19,  31},
                    },
                    {  // band 3
                        { 12, 151, 168}, {  6, 116, 152}, {  1,  76, 115},
                        {  1,  50,  81}, {  1,  32,  52}, {  1,  14,  23},
                    },
                    {  // band 4
                        { 10, 174, 191}, {  3, 130, 172}, {  1,  80, 126},
                        {  1,  53,  88}, {  1,  32,  55}, {  1,  14,  24},
                    },
                    {  // band 5
                        { 19, 219, 237}, {  3, 168, 211}, {  1,  90, 142},
                        {  1,  53,  91}, {  1,  29,  51}, {  1,  12,  21},
                    },
                },
                {  // Intra
                    {  // band 0
                        { 21,  46, 184}, { 10,  53, 130}, {  2,  49,  78},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {  3, 169, 198}, { 37, 165, 196}, { 26, 134, 176},
                        { 11, 108, 149}, {  5,  81, 112}, {  3,  47,  64},
                    },
                    {  // band 2
                        { 11, 183, 215}, {  8, 142, 192}, {  2,  91, 141},
                        {  1,  62, 100}, {  1,  38,  62}, {  1,  17,  28},
                    },
                    {  // band 3
                        { 12, 190, 223}, {  6, 149, 199}, {  1,  88, 139},
                        {  1,  56,  93}, {  1,  31,  54}, {  1,  13,  21},
                    },
                    {  // band 4
                        { 11, 197, 230}, {  3, 154, 204}, {  1,  83, 134},
                        {  1,  50,  86}, {  1,  28,  49}, {  1,  12,  21},
                    },
                    {  // band 5
                        { 17, 211, 240}, {  2, 167, 217}, {  1,  88, 143},
                        {  1,  53,  91}, {  1,  30,  53}, {  1,  14,  24},
                    },
                },
            },
            {  // UV plane
                {  // Inter
                    {  // band 0
                        {151,  30, 151}, { 50,  36, 105}, {  8,  34,  66},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 39, 111, 160}, { 62, 111, 165}, { 37,  99, 147},
                        { 15,  77, 118}, {  3,  47,  73}, {  1,  17,  27},
                    },
                    {  // band 2
                        { 48, 170, 190}, { 32, 135, 180}, { 11, 100, 149},
                        {  4,  76, 116}, {  1,  51,  80}, {  1,  22,  36},
                    },
                    {  // band 3
                        { 39, 191, 208}, { 18, 141, 191}, {  3,  96, 150},
                        {  1,  66, 110}, {  1,  41,  69}, {  1,  17,  28},
                    },
                    {  // band 4
                        { 32, 209, 219}, {  8, 152, 201}, {  1,  96, 153},
                        {  1,  63, 106}, {  1,  38,  66}, {  1,  17,  29},
                    },
                    {  // band 5
                        { 33, 230, 237}, {  5, 173, 214}, {  1, 100, 155},
                        {  1,  62, 105}, {  1,  38,  66}, {  1,  18,  32},
                    },
                },
                {  // Inter
                    {  // band 0
                        {149,  38, 231}, { 59,  51, 186}, { 12,  54, 117},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 53, 179, 226}, {126, 176, 223}, { 58, 147, 202},
                        { 28, 118, 174}, { 15,  94, 138}, { 14,  63,  87},
                    },
                    {  // band 2
                        { 58, 196, 232}, { 26, 158, 213}, {  5, 106, 166},
                        {  1,  75, 124}, {  1,  46,  79}, {  1,  23,  39},
                    },
                    {  // band 3
                        { 46, 203, 235}, { 17, 162, 213}, {  2, 104, 165},
                        {  1,  72, 120}, {  1,  44,  74}, {  1,  22,  33},
                    },
                    {  // band 4
                        { 37, 213, 238}, {  8, 167, 216}, {  1, 104, 168},
                        {  1,  68, 119}, {  1,  40,  67}, {  1,  17,  29},
                    },
                    {  // band 5
                        { 30, 228, 239}, {  4, 181, 213}, {  1, 103, 153},
                        {  1,  65, 110}, {  1,  43,  79}, {  1,  27,  56},
                    },
                },
            },
        },
        {  // TX_SIZE 3
            {  // Y plane
                {  // Intra
                    {  // band 0
                        { 76,  25,  53}, {  9,  18,  32}, {  1,  12,  18},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 29,  55,  91}, { 19,  58,  95}, { 15,  57,  89},
                        { 12,  49,  77}, {  3,  29,  44}, {  1,   8,  12},
                    },
                    {  // band 2
                        { 32, 160, 148}, { 33, 143, 146}, { 19, 122, 132},
                        {  6,  90, 102}, {  1,  58,  70}, {  1,  17,  24},
                    },
                    {  // band 3
                        { 16, 181, 181}, {  6, 142, 165}, {  1,  90, 120},
                        {  1,  50,  71}, {  1,  25,  38}, {  1,   9,  14},
                    },
                    {  // band 4
                        { 13, 203, 203}, {  3, 154, 176}, {  1,  80, 108},
                        {  1,  41,  61}, {  1,  24,  37}, {  1,  11,  17},
                    },
                    {  // band 5
                        {  6, 234, 240}, {  1, 178, 204}, {  1,  80, 119},
                        {  1,  45,  71}, {  1,  26,  42}, {  1,  12,  19},
                    },
                },
                {  // Intra
                    {  // band 0
                        { 78,  20, 135}, { 25,  18, 101}, {  5,  19,  57},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {  7, 144, 183}, {117, 151, 195}, {109, 151, 187},
                        { 39, 130, 168}, { 11, 100, 125}, {  4,  59,  64},
                    },
                    {  // band 2
                        { 20, 184, 212}, { 12, 148, 191}, {  2,  98, 141},
                        {  1,  65, 100}, {  1,  39,  61}, {  1,  14,  22},
                    },
                    {  // band 3
                        { 15, 194, 222}, {  6, 153, 198}, {  1,  92, 138},
                        {  1,  58,  91}, {  1,  32,  52}, {  1,  12,  18},
                    },
                    {  // band 4
                        { 14, 206, 232}, {  3, 162, 206}, {  1,  89, 134},
                        {  1,  52,  83}, {  1,  28,  46}, {  1,  11,  17},
                    },
                    {  // band 5
                        {  6, 225, 241}, {  1, 175, 210}, {  1,  81, 125},
                        {  1,  48,  78}, {  1,  28,  46}, {  1,  13,  21},
                    },
                },
            },
            {  // UV plane
                {  // Inter
                    {  // band 0
                        {124,  23,  93}, { 31,  24,  63}, {  6,  24,  46},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 23,  86, 126}, { 45,  90, 145}, { 31,  91, 133},
                        { 19,  80, 114}, {  7,  53,  72}, {  1,  20,  27},
                    },
                    {  // band 2
                        { 51, 186, 189}, { 48, 159, 182}, { 33, 128, 156},
                        { 15,  92, 124}, {  2,  62,  83}, {  1,  29,  43},
                    },
                    {  // band 3
                        { 36, 198, 211}, { 15, 156, 187}, {  3,  97, 137},
                        {  1,  61,  93}, {  1,  35,  57}, {  1,  15,  23},
                    },
                    {  // band 4
                        { 34, 219, 223}, {  9, 162, 193}, {  1,  91, 136},
                        {  1,  58,  92}, {  1,  35,  54}, {  1,  14,  23},
                    },
                    {  // band 5
                        { 19, 243, 243}, {  3, 191, 208}, {  1,  91, 137},
                        {  1,  56,  90}, {  1,  34,  55}, {  1,  16,  24},
                    },
                },
                {  // Inter
                    {  // band 0
                        {119,  20, 197}, { 19,  29, 156}, {  3,  30, 107},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 24, 192, 226}, {161, 193, 227}, { 97, 185, 222},
                        { 31, 158, 204}, { 16, 122, 165}, { 17,  84, 112},
                    },
                    {  // band 2
                        { 26, 202, 229}, { 11, 165, 210}, {  2, 103, 152},
                        {  1,  68, 104}, {  1,  42,  70}, {  1,  16,  36},
                    },
                    {  // band 3
                        { 24, 209, 237}, {  6, 169, 214}, {  1, 102, 154},
                        {  1,  65, 107}, {  1,  45,  68}, {  1,  17,  24},
                    },
                    {  // band 4
                        { 19, 219, 243}, {  4, 183, 226}, {  1, 115, 172},
                        {  1,  73, 119}, {  1,  43,  77}, {  1,  15,  37},
                    },
                    {  // band 5
                        { 11, 237, 241}, {  2, 190, 216}, {  1, 108, 146},
                        {  1,  59,  94}, {  1,  40,  67}, {  1,  30,  53},
                    },
                },
            },
        },
#if CONFIG_TX64X64
        {  // TX_SIZE 4
            {  // Y plane
                {  // Intra
                    {  // band 0
                        { 76,  25,  53}, {  9,  18,  32}, {  1,  12,  18},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 29,  55,  91}, { 19,  58,  95}, { 15,  57,  89},
                        { 12,  49,  77}, {  3,  29,  44}, {  1,   8,  12},
                    },
                    {  // band 2
                        { 32, 160, 148}, { 33, 143, 146}, { 19, 122, 132},
                        {  6,  90, 102}, {  1,  58,  70}, {  1,  17,  24},
                    },
                    {  // band 3
                        { 16, 181, 181}, {  6, 142, 165}, {  1,  90, 120},
                        {  1,  50,  71}, {  1,  25,  38}, {  1,   9,  14},
                    },
                    {  // band 4
                        { 13, 203, 203}, {  3, 154, 176}, {  1,  80, 108},
                        {  1,  41,  61}, {  1,  24,  37}, {  1,  11,  17},
                    },
                    {  // band 5
                        {  6, 234, 240}, {  1, 178, 204}, {  1,  80, 119},
                        {  1,  45,  71}, {  1,  26,  42}, {  1,  12,  19},
                    },
                },
                {  // Intra
                    {  // band 0
                        { 78,  20, 135}, { 25,  18, 101}, {  5,  19,  57},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {  7, 144, 183}, {117, 151, 195}, {109, 151, 187},
                        { 39, 130, 168}, { 11, 100, 125}, {  4,  59,  64},
                    },
                    {  // band 2
                        { 20, 184, 212}, { 12, 148, 191}, {  2,  98, 141},
                        {  1,  65, 100}, {  1,  39,  61}, {  1,  14,  22},
                    },
                    {  // band 3
                        { 15, 194, 222}, {  6, 153, 198}, {  1,  92, 138},
                        {  1,  58,  91}, {  1,  32,  52}, {  1,  12,  18},
                    },
                    {  // band 4
                        { 14, 206, 232}, {  3, 162, 206}, {  1,  89, 134},
                        {  1,  52,  83}, {  1,  28,  46}, {  1,  11,  17},
                    },
                    {  // band 5
                        {  6, 225, 241}, {  1, 175, 210}, {  1,  81, 125},
                        {  1,  48,  78}, {  1,  28,  46}, {  1,  13,  21},
                    },
                },
            },
            {  // UV plane
                {  // Inter
                    {  // band 0
                        {124,  23,  93}, { 31,  24,  63}, {  6,  24,  46},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 23,  86, 126}, { 45,  90, 145}, { 31,  91, 133},
                        { 19,  80, 114}, {  7,  53,  72}, {  1,  20,  27},
                    },
                    {  // band 2
                        { 51, 186, 189}, { 48, 159, 182}, { 33, 128, 156},
                        { 15,  92, 124}, {  2,  62,  83}, {  1,  29,  43},
                    },
                    {  // band 3
                        { 36, 198, 211}, { 15, 156, 187}, {  3,  97, 137},
                        {  1,  61,  93}, {  1,  35,  57}, {  1,  15,  23},
                    },
                    {  // band 4
                        { 34, 219, 223}, {  9, 162, 193}, {  1,  91, 136},
                        {  1,  58,  92}, {  1,  35,  54}, {  1,  14,  23},
                    },
                    {  // band 5
                        { 19, 243, 243}, {  3, 191, 208}, {  1,  91, 137},
                        {  1,  56,  90}, {  1,  34,  55}, {  1,  16,  24},
                    },
                },
                {  // Inter
                    {  // band 0
                        {119,  20, 197}, { 19,  29, 156}, {  3,  30, 107},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 24, 192, 226}, {161, 193, 227}, { 97, 185, 222},
                        { 31, 158, 204}, { 16, 122, 165}, { 17,  84, 112},
                    },
                    {  // band 2
                        { 26, 202, 229}, { 11, 165, 210}, {  2, 103, 152},
                        {  1,  68, 104}, {  1,  42,  70}, {  1,  16,  36},
                    },
                    {  // band 3
                        { 24, 209, 237}, {  6, 169, 214}, {  1, 102, 154},
                        {  1,  65, 107}, {  1,  45,  68}, {  1,  17,  24},
                    },
                    {  // band 4
                        { 19, 219, 243}, {  4, 183, 226}, {  1, 115, 172},
                        {  1,  73, 119}, {  1,  43,  77}, {  1,  15,  37},
                    },
                    {  // band 5
                        { 11, 237, 241}, {  2, 190, 216}, {  1, 108, 146},
                        {  1,  59,  94}, {  1,  40,  67}, {  1,  30,  53},
                    },
                },
            },
        },
#endif  // CONFIG_TX64X64
    },
    {  // Q_Index 1
#if CONFIG_CB4X4
        {  // TX_SIZE 0
            {  // Y plane
                {  // Intra
                    {  // band 0
                        {174,  30, 159}, { 76,  38, 115}, { 15,  33,  65},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 60,  80, 153}, { 72,  75, 147}, { 36,  68, 129},
                        { 15,  59, 104}, {  4,  45,  74}, {  1,  28,  45},
                    },
                    {  // band 2
                        { 70, 122, 186}, { 55, 104, 175}, { 21,  83, 144},
                        {  8,  67, 112}, {  2,  51,  82}, {  1,  34,  57},
                    },
                    {  // band 3
                        { 97, 144, 207}, { 52, 109, 195}, { 16,  77, 153},
                        {  4,  58, 113}, {  1,  43,  77}, {  1,  27,  48},
                    },
                    {  // band 4
                        {128, 148, 229}, { 76, 104, 210}, { 18,  77, 159},
                        {  4,  65, 110}, {  1,  52,  82}, {  1,  31,  55},
                    },
                    {  // band 5
                        {165,  51, 238}, {128,  50, 230}, { 57,  49, 185},
                        { 28,  47, 130}, { 12,  44,  96}, {  3,  36,  60},
                    },
                },
                {  // Intra
                    {  // band 0
                        {169, 103, 203}, {117,  96, 176}, { 56,  81, 137},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 31, 150, 224}, { 49, 128, 212}, { 19,  92, 165},
                        {  6,  67, 116}, {  2,  43,  71}, {  1,  21,  36},
                    },
                    {  // band 2
                        { 58, 156, 230}, { 47, 130, 215}, {  7,  87, 158},
                        {  2,  63, 114}, {  1,  39,  71}, {  1,  18,  36},
                    },
                    {  // band 3
                        { 85, 176, 240}, { 43, 138, 226}, {  8,  93, 172},
                        {  2,  70, 127}, {  1,  46,  81}, {  1,  26,  47},
                    },
                    {  // band 4
                        {155, 144, 248}, { 93, 116, 235}, { 21,  83, 180},
                        {  4,  59, 119}, {  1,  43,  80}, {  1,  25,  50},
                    },
                    {  // band 5
                        {203,  61, 250}, {171,  57, 243}, { 71,  57, 199},
                        { 31,  49, 144}, { 13,  42,  96}, {  7,  30,  52},
                    },
                },
            },
            {  // UV plane
                {  // Inter
                    {  // band 0
                        {204,  44, 204}, {137,  57, 184}, { 72,  62, 152},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {145, 117, 236}, {151, 112, 231}, { 87,  95, 208},
                        { 31,  77, 165}, {  5,  49,  98}, {  1,  24,  39},
                    },
                    {  // band 2
                        {146, 152, 241}, {140, 132, 236}, { 41, 103, 209},
                        { 10,  86, 165}, {  2,  55, 106}, {  1,  25,  58},
                    },
                    {  // band 3
                        {154, 181, 249}, { 84, 143, 240}, { 23, 114, 210},
                        {  6, 102, 182}, {  2,  71, 137}, {  1,  35,  90},
                    },
                    {  // band 4
                        {184, 150, 251}, {115, 130, 244}, { 34, 105, 215},
                        { 15,  89, 173}, {  1,  51, 141}, {128, 128, 128},
                    },
                    {  // band 5
                        {211,  71, 253}, {193,  78, 249}, {106,  91, 232},
                        { 61,  87, 198}, { 85, 153, 254}, {128, 128, 128},
                    },
                },
                {  // Inter
                    {  // band 0
                        {232, 104, 242}, {165, 114, 227}, { 96, 120, 206},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {137, 178, 250}, {146, 153, 245}, { 74, 108, 205},
                        { 41,  81, 149}, { 24,  55, 104}, { 13,  36,  68},
                    },
                    {  // band 2
                        {147, 185, 252}, {127, 161, 246}, { 30, 104, 208},
                        { 11,  74, 154}, {  6,  54, 100}, {  2,  29,  63},
                    },
                    {  // band 3
                        {163, 191, 254}, {101, 161, 249}, { 22, 114, 215},
                        {  6,  89, 173}, {  1,  65, 120}, {  1,   1, 170},
                    },
                    {  // band 4
                        {197, 160, 254}, {142, 141, 251}, { 39, 102, 218},
                        { 10,  76, 158}, {  1,  56, 122}, {128, 128, 128},
                    },
                    {  // band 5
                        {224,  76, 254}, {215,  84, 253}, {107,  85, 232},
                        { 43,  71, 177}, {  1,   1, 254}, {128, 128, 128},
                    },
                },
            },
        },
#endif
        {  // TX_SIZE 0
            {  // Y plane
                {  // Intra
                    {  // band 0
                        {174,  30, 159}, { 76,  38, 115}, { 15,  33,  65},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 60,  80, 153}, { 72,  75, 147}, { 36,  68, 129},
                        { 15,  59, 104}, {  4,  45,  74}, {  1,  28,  45},
                    },
                    {  // band 2
                        { 70, 122, 186}, { 55, 104, 175}, { 21,  83, 144},
                        {  8,  67, 112}, {  2,  51,  82}, {  1,  34,  57},
                    },
                    {  // band 3
                        { 97, 144, 207}, { 52, 109, 195}, { 16,  77, 153},
                        {  4,  58, 113}, {  1,  43,  77}, {  1,  27,  48},
                    },
                    {  // band 4
                        {128, 148, 229}, { 76, 104, 210}, { 18,  77, 159},
                        {  4,  65, 110}, {  1,  52,  82}, {  1,  31,  55},
                    },
                    {  // band 5
                        {165,  51, 238}, {128,  50, 230}, { 57,  49, 185},
                        { 28,  47, 130}, { 12,  44,  96}, {  3,  36,  60},
                    },
                },
                {  // Intra
                    {  // band 0
                        {169, 103, 203}, {117,  96, 176}, { 56,  81, 137},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 31, 150, 224}, { 49, 128, 212}, { 19,  92, 165},
                        {  6,  67, 116}, {  2,  43,  71}, {  1,  21,  36},
                    },
                    {  // band 2
                        { 58, 156, 230}, { 47, 130, 215}, {  7,  87, 158},
                        {  2,  63, 114}, {  1,  39,  71}, {  1,  18,  36},
                    },
                    {  // band 3
                        { 85, 176, 240}, { 43, 138, 226}, {  8,  93, 172},
                        {  2,  70, 127}, {  1,  46,  81}, {  1,  26,  47},
                    },
                    {  // band 4
                        {155, 144, 248}, { 93, 116, 235}, { 21,  83, 180},
                        {  4,  59, 119}, {  1,  43,  80}, {  1,  25,  50},
                    },
                    {  // band 5
                        {203,  61, 250}, {171,  57, 243}, { 71,  57, 199},
                        { 31,  49, 144}, { 13,  42,  96}, {  7,  30,  52},
                    },
                },
            },
            {  // UV plane
                {  // Inter
                    {  // band 0
                        {204,  44, 204}, {137,  57, 184}, { 72,  62, 152},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {145, 117, 236}, {151, 112, 231}, { 87,  95, 208},
                        { 31,  77, 165}, {  5,  49,  98}, {  1,  24,  39},
                    },
                    {  // band 2
                        {146, 152, 241}, {140, 132, 236}, { 41, 103, 209},
                        { 10,  86, 165}, {  2,  55, 106}, {  1,  25,  58},
                    },
                    {  // band 3
                        {154, 181, 249}, { 84, 143, 240}, { 23, 114, 210},
                        {  6, 102, 182}, {  2,  71, 137}, {  1,  35,  90},
                    },
                    {  // band 4
                        {184, 150, 251}, {115, 130, 244}, { 34, 105, 215},
                        { 15,  89, 173}, {  1,  51, 141}, {128, 128, 128},
                    },
                    {  // band 5
                        {211,  71, 253}, {193,  78, 249}, {106,  91, 232},
                        { 61,  87, 198}, { 85, 153, 254}, {128, 128, 128},
                    },
                },
                {  // Inter
                    {  // band 0
                        {232, 104, 242}, {165, 114, 227}, { 96, 120, 206},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {137, 178, 250}, {146, 153, 245}, { 74, 108, 205},
                        { 41,  81, 149}, { 24,  55, 104}, { 13,  36,  68},
                    },
                    {  // band 2
                        {147, 185, 252}, {127, 161, 246}, { 30, 104, 208},
                        { 11,  74, 154}, {  6,  54, 100}, {  2,  29,  63},
                    },
                    {  // band 3
                        {163, 191, 254}, {101, 161, 249}, { 22, 114, 215},
                        {  6,  89, 173}, {  1,  65, 120}, {  1,   1, 170},
                    },
                    {  // band 4
                        {197, 160, 254}, {142, 141, 251}, { 39, 102, 218},
                        { 10,  76, 158}, {  1,  56, 122}, {128, 128, 128},
                    },
                    {  // band 5
                        {224,  76, 254}, {215,  84, 253}, {107,  85, 232},
                        { 43,  71, 177}, {  1,   1, 254}, {128, 128, 128},
                    },
                },
            },
        },
        {  // TX_SIZE 1
            {  // Y plane
                {  // Intra
                    {  // band 0
                        { 68,  37, 120}, { 21,  34,  82}, {  5,  26,  49},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 41,  89, 138}, { 56,  83, 132}, { 31,  73, 115},
                        { 16,  62,  92}, {  5,  45,  62}, {  1,  24,  32},
                    },
                    {  // band 2
                        { 48, 139, 165}, { 30, 114, 160}, { 13,  92, 132},
                        {  6,  72, 103}, {  3,  49,  72}, {  1,  26,  41},
                    },
                    {  // band 3
                        { 44, 162, 191}, { 20, 127, 175}, {  5,  90, 137},
                        {  1,  62, 100}, {  1,  38,  63}, {  1,  20,  32},
                    },
                    {  // band 4
                        { 51, 184, 213}, { 16, 137, 193}, {  2,  89, 143},
                        {  1,  60, 102}, {  1,  39,  66}, {  1,  23,  37},
                    },
                    {  // band 5
                        { 76, 200, 235}, { 27, 150, 216}, {  3,  99, 164},
                        {  1,  70, 119}, {  1,  45,  77}, {  1,  22,  38},
                    },
                },
                {  // Intra
                    {  // band 0
                        { 81, 112, 199}, { 49, 101, 164}, { 19,  80, 119},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 12, 181, 217}, { 48, 151, 212}, { 38, 118, 180},
                        { 22,  95, 140}, { 11,  67,  92}, { 13,  46,  44},
                    },
                    {  // band 2
                        { 29, 188, 226}, { 19, 147, 210}, {  5,  95, 154},
                        {  4,  68, 106}, {  3,  44,  60}, {  1,  24,  27},
                    },
                    {  // band 3
                        { 30, 195, 234}, { 15, 153, 216}, {  3,  95, 156},
                        {  2,  66, 108}, {  2,  44,  62}, {  1,  24,  29},
                    },
                    {  // band 4
                        { 36, 203, 243}, { 12, 162, 225}, {  2,  98, 163},
                        {  2,  67, 113}, {  2,  45,  68}, {  1,  24,  34},
                    },
                    {  // band 5
                        { 86, 207, 248}, { 35, 165, 236}, {  3, 107, 180},
                        {  1,  73, 128}, {  1,  45,  78}, {  1,  20,  34},
                    },
                },
            },
            {  // UV plane
                {  // Inter
                    {  // band 0
                        {188,  37, 205}, {118,  51, 172}, { 56,  57, 135},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {116, 135, 225}, {144, 123, 221}, { 72, 103, 197},
                        { 35,  77, 153}, { 15,  47,  82}, {  6,  25,  34},
                    },
                    {  // band 2
                        {128, 171, 233}, { 82, 142, 226}, { 31, 106, 191},
                        { 16,  82, 146}, {  9,  59,  98}, {  4,  33,  54},
                    },
                    {  // band 3
                        {126, 197, 241}, { 66, 155, 230}, { 18, 108, 190},
                        {  7,  82, 148}, {  3,  58,  98}, {  1,  25,  50},
                    },
                    {  // band 4
                        {117, 207, 244}, { 44, 163, 233}, {  9, 112, 191},
                        {  5,  84, 148}, {  3,  61,  87}, {  1,  28,  38},
                    },
                    {  // band 5
                        {112, 214, 249}, { 39, 174, 240}, {  6, 125, 205},
                        {  4,  96, 163}, {  5,  66, 100}, {  1, 128, 254},
                    },
                },
                {  // Inter
                    {  // band 0
                        {227,  70, 234}, {145,  91, 213}, { 61, 100, 173},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {108, 198, 243}, {171, 172, 240}, {118, 130, 210},
                        {104, 107, 165}, { 64,  85, 114}, { 55,  64,  60},
                    },
                    {  // band 2
                        {110, 208, 247}, { 64, 175, 237}, { 24, 112, 187},
                        { 24,  81, 133}, { 24,  63,  83}, { 21,  47,  53},
                    },
                    {  // band 3
                        { 91, 218, 249}, { 46, 188, 238}, {  8, 113, 184},
                        {  5,  83, 137}, {  6,  62,  95}, { 17,  44,  94},
                    },
                    {  // band 4
                        { 84, 216, 248}, { 30, 187, 237}, {  2, 117, 188},
                        {  1,  88, 141}, {  3,  63,  98}, {  1,   1,   1},
                    },
                    {  // band 5
                        {116, 218, 252}, { 47, 186, 242}, {  2, 132, 204},
                        {  1, 106, 175}, {  1,  88, 104}, {  1, 254, 128},
                    },
                },
            },
        },
        {  // TX_SIZE 2
            {  // Y plane
                {  // Intra
                    {  // band 0
                        { 35,  41, 129}, { 12,  30,  70}, {  2,  19,  32},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 30,  77, 116}, { 39,  70, 110}, { 20,  58,  96},
                        {  8,  47,  77}, {  2,  33,  52}, {  1,  17,  26},
                    },
                    {  // band 2
                        { 31, 123, 146}, { 18, 103, 140}, {  7,  81, 119},
                        {  2,  62,  95}, {  1,  44,  70}, {  1,  26,  42},
                    },
                    {  // band 3
                        { 21, 149, 170}, {  9, 114, 158}, {  2,  80, 126},
                        {  1,  57,  94}, {  1,  36,  61}, {  1,  18,  31},
                    },
                    {  // band 4
                        { 20, 178, 199}, {  6, 134, 183}, {  1,  87, 139},
                        {  1,  60, 100}, {  1,  37,  64}, {  1,  18,  31},
                    },
                    {  // band 5
                        { 36, 218, 233}, {  6, 160, 207}, {  1,  92, 147},
                        {  1,  59, 101}, {  1,  35,  62}, {  1,  18,  31},
                    },
                },
                {  // Intra
                    {  // band 0
                        { 17,  62, 211}, { 14,  62, 153}, {  5,  50,  84},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 11, 180, 205}, { 87, 160, 205}, { 53, 128, 184},
                        { 27, 106, 156}, { 13,  79, 115}, {  6,  46,  67},
                    },
                    {  // band 2
                        { 32, 194, 220}, { 20, 145, 202}, {  4,  96, 152},
                        {  1,  67, 111}, {  1,  42,  70}, {  1,  21,  37},
                    },
                    {  // band 3
                        { 30, 204, 228}, { 14, 152, 207}, {  1,  92, 149},
                        {  1,  61, 103}, {  1,  34,  59}, {  1,  16,  28},
                    },
                    {  // band 4
                        { 27, 213, 235}, {  7, 159, 210}, {  1,  88, 143},
                        {  1,  55,  94}, {  1,  31,  53}, {  1,  16,  27},
                    },
                    {  // band 5
                        { 28, 223, 243}, {  4, 173, 217}, {  1,  91, 146},
                        {  1,  58,  98}, {  1,  35,  60}, {  1,  19,  33},
                    },
                },
            },
            {  // UV plane
                {  // Inter
                    {  // band 0
                        {172,  37, 202}, { 83,  51, 156}, { 24,  53, 110},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 76, 134, 206}, {110, 124, 200}, { 47, 106, 180},
                        { 15,  82, 145}, {  3,  48,  83}, {  1,  19,  32},
                    },
                    {  // band 2
                        { 80, 176, 220}, { 49, 145, 212}, { 17, 112, 180},
                        {  7,  84, 140}, {  1,  53,  89}, {  1,  27,  43},
                    },
                    {  // band 3
                        { 74, 201, 232}, { 38, 158, 221}, {  8, 112, 179},
                        {  2,  79, 132}, {  1,  47,  82}, {  1,  26,  42},
                    },
                    {  // band 4
                        { 73, 215, 239}, { 28, 169, 227}, {  3, 112, 176},
                        {  1,  74, 126}, {  1,  48,  79}, {  1,  27,  44},
                    },
                    {  // band 5
                        { 71, 233, 244}, { 18, 180, 230}, {  1, 114, 180},
                        {  1,  80, 134}, {  1,  51,  85}, {  1,  26,  36},
                    },
                },
                {  // Inter
                    {  // band 0
                        {213,  34, 244}, {126,  57, 212}, { 46,  67, 151},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {120, 202, 245}, {198, 173, 241}, {119, 146, 224},
                        { 76, 126, 195}, { 44, 102, 159}, { 40,  76, 115},
                    },
                    {  // band 2
                        {120, 215, 248}, { 69, 171, 237}, { 23, 119, 194},
                        { 10,  86, 147}, {  2,  56,  94}, {  1,  25,  44},
                    },
                    {  // band 3
                        {102, 226, 250}, { 53, 183, 239}, {  9, 118, 188},
                        {  2,  78, 131}, {  1,  48,  89}, {  1,  17,  36},
                    },
                    {  // band 4
                        { 86, 235, 252}, { 34, 194, 240}, {  2, 109, 173},
                        {  1,  68, 118}, {  1,  44,  79}, {  1,   1,  38},
                    },
                    {  // band 5
                        { 59, 236, 243}, { 11, 189, 228}, {  1, 112, 187},
                        {  1,  88, 145}, {  1,  55,  92}, {  1,   1, 128},
                    },
                },
            },
        },
        {  // TX_SIZE 3
            {  // Y plane
                {  // Intra
                    {  // band 0
                        { 41,  40, 104}, { 12,  31,  64}, {  2,  16,  28},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 65,  58, 132}, { 50,  61, 130}, { 40,  57, 116},
                        { 22,  46,  87}, {  2,  28,  44}, {  1,  11,  17},
                    },
                    {  // band 2
                        { 55, 139, 135}, { 46, 122, 132}, { 21,  89, 110},
                        {  6,  60,  78}, {  1,  38,  54}, {  1,  17,  27},
                    },
                    {  // band 3
                        { 29, 167, 161}, { 10, 120, 141}, {  1,  69,  98},
                        {  1,  42,  66}, {  1,  28,  44}, {  1,  15,  24},
                    },
                    {  // band 4
                        { 19, 191, 180}, {  4, 125, 154}, {  1,  70, 107},
                        {  1,  48,  77}, {  1,  33,  53}, {  1,  17,  28},
                    },
                    {  // band 5
                        { 16, 238, 231}, {  2, 163, 198}, {  1,  85, 134},
                        {  1,  54,  90}, {  1,  34,  57}, {  1,  17,  29},
                    },
                },
                {  // Intra
                    {  // band 0
                        { 70,  15, 216}, { 40,  18, 164}, { 14,  17,  83},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 25, 150, 200}, {185, 154, 211}, {123, 137, 199},
                        { 67, 119, 177}, { 31,  96, 137}, { 18,  63,  86},
                    },
                    {  // band 2
                        { 57, 187, 223}, { 35, 148, 207}, {  7, 104, 159},
                        {  2,  72, 113}, {  1,  44,  71}, {  1,  20,  34},
                    },
                    {  // band 3
                        { 44, 203, 233}, { 18, 157, 212}, {  1,  98, 150},
                        {  1,  61, 102}, {  1,  38,  62}, {  1,  19,  31},
                    },
                    {  // band 4
                        { 41, 215, 238}, { 11, 166, 215}, {  1,  94, 146},
                        {  1,  60, 101}, {  1,  37,  63}, {  1,  17,  28},
                    },
                    {  // band 5
                        { 19, 236, 246}, {  3, 188, 223}, {  1,  95, 146},
                        {  1,  58,  95}, {  1,  34,  56}, {  1,  17,  27},
                    },
                },
            },
            {  // UV plane
                {  // Inter
                    {  // band 0
                        {146,  27, 156}, { 49,  32, 116}, { 10,  39,  77},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 47, 101, 172}, { 93, 100, 178}, { 58,  91, 165},
                        { 26,  75, 134}, {  4,  49,  82}, {  2,  22,  33},
                    },
                    {  // band 2
                        { 60, 158, 196}, { 44, 135, 186}, { 25, 106, 157},
                        {  8,  81, 124}, {  2,  56,  86}, {  1,  28,  45},
                    },
                    {  // band 3
                        { 44, 169, 212}, { 15, 138, 196}, {  2, 100, 157},
                        {  1,  74, 119}, {  1,  49,  76}, {  1,  20,  34},
                    },
                    {  // band 4
                        { 38, 199, 231}, { 11, 158, 214}, {  1, 111, 167},
                        {  1,  76, 122}, {  1,  44,  76}, {  1,  17,  39},
                    },
                    {  // band 5
                        { 40, 236, 246}, { 10, 187, 230}, {  1, 115, 175},
                        {  1,  74, 122}, {  1,  42,  71}, {  1,  14,  59},
                    },
                },
                {  // Inter
                    {  // band 0
                        {161,  26, 237}, { 65,  46, 209}, { 21,  46, 161},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 87, 229, 245}, {206, 214, 244}, {148, 186, 236},
                        { 89, 165, 221}, { 41, 132, 186}, { 37,  93, 141},
                    },
                    {  // band 2
                        { 93, 231, 246}, { 47, 181, 231}, {  8, 117, 188},
                        {  2,  84, 138}, {  1,  43,  87}, {  1,  27,  41},
                    },
                    {  // band 3
                        { 80, 239, 250}, { 28, 190, 236}, {  1, 119, 183},
                        {  1,  84, 135}, {  1,  81,  69}, {  1, 102,   1},
                    },
                    {  // band 4
                        { 67, 245, 252}, { 22, 206, 242}, {  1, 130, 195},
                        {  1,  77, 136}, {  1,  35,  88}, {128, 128, 128},
                    },
                    {  // band 5
                        { 43, 250, 228}, { 31, 185, 204}, {  6, 101, 183},
                        {  1,  92, 151}, {  1,  84, 137}, {128, 128, 128},
                    },
                },
            },
        },
#if CONFIG_TX64X64
        {  // TX_SIZE 4
            {  // Y plane
                {  // Intra
                    {  // band 0
                        { 41,  40, 104}, { 12,  31,  64}, {  2,  16,  28},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 65,  58, 132}, { 50,  61, 130}, { 40,  57, 116},
                        { 22,  46,  87}, {  2,  28,  44}, {  1,  11,  17},
                    },
                    {  // band 2
                        { 55, 139, 135}, { 46, 122, 132}, { 21,  89, 110},
                        {  6,  60,  78}, {  1,  38,  54}, {  1,  17,  27},
                    },
                    {  // band 3
                        { 29, 167, 161}, { 10, 120, 141}, {  1,  69,  98},
                        {  1,  42,  66}, {  1,  28,  44}, {  1,  15,  24},
                    },
                    {  // band 4
                        { 19, 191, 180}, {  4, 125, 154}, {  1,  70, 107},
                        {  1,  48,  77}, {  1,  33,  53}, {  1,  17,  28},
                    },
                    {  // band 5
                        { 16, 238, 231}, {  2, 163, 198}, {  1,  85, 134},
                        {  1,  54,  90}, {  1,  34,  57}, {  1,  17,  29},
                    },
                },
                {  // Intra
                    {  // band 0
                        { 70,  15, 216}, { 40,  18, 164}, { 14,  17,  83},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 25, 150, 200}, {185, 154, 211}, {123, 137, 199},
                        { 67, 119, 177}, { 31,  96, 137}, { 18,  63,  86},
                    },
                    {  // band 2
                        { 57, 187, 223}, { 35, 148, 207}, {  7, 104, 159},
                        {  2,  72, 113}, {  1,  44,  71}, {  1,  20,  34},
                    },
                    {  // band 3
                        { 44, 203, 233}, { 18, 157, 212}, {  1,  98, 150},
                        {  1,  61, 102}, {  1,  38,  62}, {  1,  19,  31},
                    },
                    {  // band 4
                        { 41, 215, 238}, { 11, 166, 215}, {  1,  94, 146},
                        {  1,  60, 101}, {  1,  37,  63}, {  1,  17,  28},
                    },
                    {  // band 5
                        { 19, 236, 246}, {  3, 188, 223}, {  1,  95, 146},
                        {  1,  58,  95}, {  1,  34,  56}, {  1,  17,  27},
                    },
                },
            },
            {  // UV plane
                {  // Inter
                    {  // band 0
                        {146,  27, 156}, { 49,  32, 116}, { 10,  39,  77},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 47, 101, 172}, { 93, 100, 178}, { 58,  91, 165},
                        { 26,  75, 134}, {  4,  49,  82}, {  2,  22,  33},
                    },
                    {  // band 2
                        { 60, 158, 196}, { 44, 135, 186}, { 25, 106, 157},
                        {  8,  81, 124}, {  2,  56,  86}, {  1,  28,  45},
                    },
                    {  // band 3
                        { 44, 169, 212}, { 15, 138, 196}, {  2, 100, 157},
                        {  1,  74, 119}, {  1,  49,  76}, {  1,  20,  34},
                    },
                    {  // band 4
                        { 38, 199, 231}, { 11, 158, 214}, {  1, 111, 167},
                        {  1,  76, 122}, {  1,  44,  76}, {  1,  17,  39},
                    },
                    {  // band 5
                        { 40, 236, 246}, { 10, 187, 230}, {  1, 115, 175},
                        {  1,  74, 122}, {  1,  42,  71}, {  1,  14,  59},
                    },
                },
                {  // Inter
                    {  // band 0
                        {161,  26, 237}, { 65,  46, 209}, { 21,  46, 161},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 87, 229, 245}, {206, 214, 244}, {148, 186, 236},
                        { 89, 165, 221}, { 41, 132, 186}, { 37,  93, 141},
                    },
                    {  // band 2
                        { 93, 231, 246}, { 47, 181, 231}, {  8, 117, 188},
                        {  2,  84, 138}, {  1,  43,  87}, {  1,  27,  41},
                    },
                    {  // band 3
                        { 80, 239, 250}, { 28, 190, 236}, {  1, 119, 183},
                        {  1,  84, 135}, {  1,  81,  69}, {  1, 102,   1},
                    },
                    {  // band 4
                        { 67, 245, 252}, { 22, 206, 242}, {  1, 130, 195},
                        {  1,  77, 136}, {  1,  35,  88}, {128, 128, 128},
                    },
                    {  // band 5
                        { 43, 250, 228}, { 31, 185, 204}, {  6, 101, 183},
                        {  1,  92, 151}, {  1,  84, 137}, {128, 128, 128},
                    },
                },
            },
        },
#endif  // CONFIG_TX64X64
    },
    {  // Q_Index 2
#if CONFIG_CB4X4
        {  // TX_SIZE 0
            {  // Y plane
                {  // Intra
                    {  // band 0
                        {181,  22, 175}, { 96,  37, 147}, { 35,  41, 105},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 80,  95, 197}, {111,  92, 193}, { 59,  87, 175},
                        { 29,  79, 150}, { 10,  65, 118}, {  2,  47,  82},
                    },
                    {  // band 2
                        { 90, 141, 216}, { 77, 120, 210}, { 23,  95, 184},
                        { 11,  81, 151}, {  6,  75, 130}, {  2,  58, 113},
                    },
                    {  // band 3
                        {122, 167, 231}, { 66, 119, 225}, { 26,  87, 189},
                        {  7,  76, 151}, {  2,  63, 125}, {  1,  59,  77},
                    },
                    {  // band 4
                        {162, 147, 244}, {110,  97, 236}, { 32,  88, 204},
                        { 11,  89, 174}, {  5,  78, 151}, {128, 128, 128},
                    },
                    {  // band 5
                        {205,  59, 251}, {176,  68, 248}, { 90,  71, 223},
                        { 49,  72, 188}, { 17,  74, 203}, {128, 128, 128},
                    },
                },
                {  // Intra
                    {  // band 0
                        {188,  70, 207}, {140,  73, 189}, { 85,  73, 163},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 59, 144, 239}, { 79, 126, 237}, { 31, 102, 202},
                        { 10,  81, 153}, {  3,  56, 102}, {  2,  33,  59},
                    },
                    {  // band 2
                        {100, 152, 243}, { 80, 129, 236}, { 14,  94, 194},
                        {  4,  72, 150}, {  1,  50, 103}, {  1,  35,  60},
                    },
                    {  // band 3
                        {130, 183, 247}, { 70, 139, 242}, { 19, 100, 203},
                        {  4,  83, 159}, {  1,  59, 119}, {  1,  44,  72},
                    },
                    {  // band 4
                        {197, 138, 252}, {135, 107, 247}, { 31,  86, 210},
                        {  7,  74, 160}, {  1,  53, 107}, {128, 128, 128},
                    },
                    {  // band 5
                        {229,  54, 254}, {200,  51, 251}, { 83,  61, 226},
                        { 33,  55, 177}, { 12,  74, 145}, {128, 128, 128},
                    },
                },
            },
            {  // UV plane
                {  // Inter
                    {  // band 0
                        {229,  20, 235}, {183,  37, 221}, {127,  47, 198},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {188, 115, 251}, {208, 110, 250}, {101,  99, 235},
                        { 38,  81, 197}, {  9,  56, 132}, {  9,  52,  63},
                    },
                    {  // band 2
                        {189, 150, 252}, {186, 137, 251}, { 54, 107, 236},
                        { 14,  90, 195}, {  1,  89, 104}, {128, 128, 128},
                    },
                    {  // band 3
                        {209, 180, 254}, {142, 145, 253}, { 51, 130, 236},
                        {  6, 128, 214}, {  1, 128, 254}, {128, 128, 128},
                    },
                    {  // band 4
                        {231, 140, 254}, {194, 128, 254}, { 75, 119, 233},
                        {128,  23, 230}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 5
                        {244,  59, 254}, {239,  81, 254}, {128,  85, 254},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                },
                {  // Inter
                    {  // band 0
                        {246,  55, 247}, {197,  64, 235}, {141,  74, 218},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {178, 163, 254}, {192, 138, 252}, { 85, 103, 231},
                        { 49,  81, 179}, { 32,  54, 133}, { 12,  26,  98},
                    },
                    {  // band 2
                        {189, 173, 254}, {179, 150, 253}, { 60,  94, 237},
                        { 34,  81, 198}, { 20,  53, 187}, {128, 128, 128},
                    },
                    {  // band 3
                        {202, 191, 254}, {157, 160, 254}, { 57, 117, 240},
                        { 28, 105, 211}, {  1, 128,   1}, {128, 128, 128},
                    },
                    {  // band 4
                        {231, 146, 254}, {208, 133, 254}, { 66,  78, 233},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 5
                        {246,  49, 254}, {246,  63, 254}, { 85, 142, 254},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                },
            },
        },
#endif
        {  // TX_SIZE 0
            {  // Y plane
                {  // Intra
                    {  // band 0
                        {181,  22, 175}, { 96,  37, 147}, { 35,  41, 105},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 80,  95, 197}, {111,  92, 193}, { 59,  87, 175},
                        { 29,  79, 150}, { 10,  65, 118}, {  2,  47,  82},
                    },
                    {  // band 2
                        { 90, 141, 216}, { 77, 120, 210}, { 23,  95, 184},
                        { 11,  81, 151}, {  6,  75, 130}, {  2,  58, 113},
                    },
                    {  // band 3
                        {122, 167, 231}, { 66, 119, 225}, { 26,  87, 189},
                        {  7,  76, 151}, {  2,  63, 125}, {  1,  59,  77},
                    },
                    {  // band 4
                        {162, 147, 244}, {110,  97, 236}, { 32,  88, 204},
                        { 11,  89, 174}, {  5,  78, 151}, {128, 128, 128},
                    },
                    {  // band 5
                        {205,  59, 251}, {176,  68, 248}, { 90,  71, 223},
                        { 49,  72, 188}, { 17,  74, 203}, {128, 128, 128},
                    },
                },
                {  // Intra
                    {  // band 0
                        {188,  70, 207}, {140,  73, 189}, { 85,  73, 163},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 59, 144, 239}, { 79, 126, 237}, { 31, 102, 202},
                        { 10,  81, 153}, {  3,  56, 102}, {  2,  33,  59},
                    },
                    {  // band 2
                        {100, 152, 243}, { 80, 129, 236}, { 14,  94, 194},
                        {  4,  72, 150}, {  1,  50, 103}, {  1,  35,  60},
                    },
                    {  // band 3
                        {130, 183, 247}, { 70, 139, 242}, { 19, 100, 203},
                        {  4,  83, 159}, {  1,  59, 119}, {  1,  44,  72},
                    },
                    {  // band 4
                        {197, 138, 252}, {135, 107, 247}, { 31,  86, 210},
                        {  7,  74, 160}, {  1,  53, 107}, {128, 128, 128},
                    },
                    {  // band 5
                        {229,  54, 254}, {200,  51, 251}, { 83,  61, 226},
                        { 33,  55, 177}, { 12,  74, 145}, {128, 128, 128},
                    },
                },
            },
            {  // UV plane
                {  // Inter
                    {  // band 0
                        {229,  20, 235}, {183,  37, 221}, {127,  47, 198},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {188, 115, 251}, {208, 110, 250}, {101,  99, 235},
                        { 38,  81, 197}, {  9,  56, 132}, {  9,  52,  63},
                    },
                    {  // band 2
                        {189, 150, 252}, {186, 137, 251}, { 54, 107, 236},
                        { 14,  90, 195}, {  1,  89, 104}, {128, 128, 128},
                    },
                    {  // band 3
                        {209, 180, 254}, {142, 145, 253}, { 51, 130, 236},
                        {  6, 128, 214}, {  1, 128, 254}, {128, 128, 128},
                    },
                    {  // band 4
                        {231, 140, 254}, {194, 128, 254}, { 75, 119, 233},
                        {128,  23, 230}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 5
                        {244,  59, 254}, {239,  81, 254}, {128,  85, 254},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                },
                {  // Inter
                    {  // band 0
                        {246,  55, 247}, {197,  64, 235}, {141,  74, 218},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {178, 163, 254}, {192, 138, 252}, { 85, 103, 231},
                        { 49,  81, 179}, { 32,  54, 133}, { 12,  26,  98},
                    },
                    {  // band 2
                        {189, 173, 254}, {179, 150, 253}, { 60,  94, 237},
                        { 34,  81, 198}, { 20,  53, 187}, {128, 128, 128},
                    },
                    {  // band 3
                        {202, 191, 254}, {157, 160, 254}, { 57, 117, 240},
                        { 28, 105, 211}, {  1, 128,   1}, {128, 128, 128},
                    },
                    {  // band 4
                        {231, 146, 254}, {208, 133, 254}, { 66,  78, 233},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 5
                        {246,  49, 254}, {246,  63, 254}, { 85, 142, 254},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                },
            },
        },
        {  // TX_SIZE 1
            {  // Y plane
                {  // Intra
                    {  // band 0
                        { 45,  28, 124}, { 23,  35, 107}, { 10,  34,  78},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 53,  99, 177}, { 82,  96, 174}, { 46,  89, 158},
                        { 21,  76, 133}, {  6,  56,  94}, {  1,  33,  54},
                    },
                    {  // band 2
                        { 68, 147, 201}, { 42, 124, 195}, { 17,  98, 166},
                        {  7,  75, 131}, {  2,  53,  93}, {  1,  33,  59},
                    },
                    {  // band 3
                        { 65, 176, 217}, { 30, 137, 206}, {  6,  97, 167},
                        {  2,  70, 128}, {  1,  47,  88}, {  1,  29,  46},
                    },
                    {  // band 4
                        { 69, 195, 232}, { 24, 146, 218}, {  4, 100, 175},
                        {  2,  72, 134}, {  1,  51,  93}, {  1,  29,  52},
                    },
                    {  // band 5
                        { 96, 212, 246}, { 39, 158, 234}, {  6, 109, 192},
                        {  2,  77, 144}, {  1,  50,  95}, {  1,  20,  45},
                    },
                },
                {  // Intra
                    {  // band 0
                        { 71,  80, 213}, { 53,  73, 181}, { 25,  66, 141},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 35, 168, 231}, { 91, 150, 229}, { 49, 122, 202},
                        { 22,  97, 162}, { 10,  68, 108}, {  9,  48,  57},
                    },
                    {  // band 2
                        { 56, 178, 236}, { 32, 148, 225}, {  9,  99, 176},
                        {  4,  69, 127}, {  2,  44,  78}, {  1,  25,  41},
                    },
                    {  // band 3
                        { 57, 191, 242}, { 27, 155, 230}, {  5, 102, 180},
                        {  2,  71, 133}, {  1,  44,  78}, {  1,  27,  41},
                    },
                    {  // band 4
                        { 67, 201, 247}, { 24, 162, 237}, {  3, 106, 188},
                        {  3,  74, 137}, {  1,  46,  85}, {  1,  34,  48},
                    },
                    {  // band 5
                        {111, 210, 251}, { 47, 166, 244}, {  3, 113, 199},
                        {  2,  77, 146}, {  1,  48,  93}, {  1,  38,  22},
                    },
                },
            },
            {  // UV plane
                {  // Inter
                    {  // band 0
                        {206,  21, 221}, {150,  36, 195}, { 94,  44, 164},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {147, 128, 239}, {194, 122, 238}, { 95, 104, 220},
                        { 39,  81, 183}, { 13,  53, 111}, {  3,  24,  49},
                    },
                    {  // band 2
                        {164, 163, 244}, {106, 142, 239}, { 50, 112, 215},
                        { 26,  90, 177}, { 12,  67, 130}, {  1,   1,  64},
                    },
                    {  // band 3
                        {155, 193, 249}, { 88, 158, 244}, { 26, 124, 220},
                        { 10,  98, 173}, {  1,  77, 126}, {128, 128, 128},
                    },
                    {  // band 4
                        {141, 205, 252}, { 64, 174, 248}, { 17, 124, 221},
                        { 12,  92, 176}, {  1,  29, 148}, {128, 128, 128},
                    },
                    {  // band 5
                        {150, 217, 254}, { 74, 191, 252}, { 30, 144, 215},
                        {  1, 106, 137}, {128,   1, 128}, {128, 128, 128},
                    },
                },
                {  // Inter
                    {  // band 0
                        {241,  37, 242}, {175,  48, 223}, { 99,  53, 189},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {153, 183, 248}, {212, 156, 247}, {134, 124, 221},
                        { 88, 103, 184}, { 59,  86, 132}, { 29,  61,  67},
                    },
                    {  // band 2
                        {162, 199, 250}, {106, 167, 247}, { 56, 110, 207},
                        { 32,  85, 165}, { 16,  71, 130}, {  1,  93, 254},
                    },
                    {  // band 3
                        {143, 213, 252}, { 86, 187, 250}, { 23, 124, 220},
                        {  7,  95, 176}, {  1, 109, 102}, {128, 128, 128},
                    },
                    {  // band 4
                        {130, 219, 254}, { 70, 201, 253}, { 15, 128, 215},
                        {  1, 101, 201}, {  1,  64, 170}, {128, 128, 128},
                    },
                    {  // band 5
                        {155, 219, 254}, {105, 207, 254}, { 28, 155, 229},
                        {  1, 153, 191}, {128, 128, 128}, {128, 128, 128},
                    },
                },
            },
        },
        {  // TX_SIZE 2
            {  // Y plane
                {  // Intra
                    {  // band 0
                        { 18,  26, 117}, { 10,  29,  82}, {  3,  25,  52},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 35,  88, 152}, { 62,  85, 150}, { 36,  77, 137},
                        { 16,  66, 116}, {  4,  47,  81}, {  1,  26,  44},
                    },
                    {  // band 2
                        { 55, 141, 182}, { 32, 119, 177}, { 12,  93, 154},
                        {  4,  71, 123}, {  1,  51,  89}, {  1,  32,  56},
                    },
                    {  // band 3
                        { 46, 171, 202}, { 21, 130, 191}, {  5,  91, 154},
                        {  1,  64, 115}, {  1,  42,  77}, {  1,  25,  41},
                    },
                    {  // band 4
                        { 43, 195, 219}, { 12, 142, 203}, {  1,  91, 156},
                        {  1,  63, 115}, {  1,  41,  77}, {  1,  22,  43},
                    },
                    {  // band 5
                        { 42, 221, 238}, {  8, 162, 219}, {  1,  98, 167},
                        {  1,  67, 123}, {  1,  43,  83}, {  1,  25,  38},
                    },
                },
                {  // Intra
                    {  // band 0
                        { 16,  51, 216}, { 20,  48, 168}, {  9,  44, 109},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 34, 164, 226}, {124, 148, 226}, { 72, 127, 207},
                        { 36, 107, 175}, { 15,  81, 129}, {  6,  51,  79},
                    },
                    {  // band 2
                        { 61, 182, 234}, { 35, 148, 220}, {  9, 101, 178},
                        {  4,  71, 134}, {  1,  46,  90}, {  1,  24,  51},
                    },
                    {  // band 3
                        { 54, 198, 239}, { 25, 156, 224}, {  3,  98, 173},
                        {  1,  66, 124}, {  1,  41,  78}, {  1,  15,  37},
                    },
                    {  // band 4
                        { 48, 209, 242}, { 12, 162, 226}, {  1,  96, 169},
                        {  1,  63, 119}, {  1,  40,  78}, {  1,  18,  45},
                    },
                    {  // band 5
                        { 44, 223, 247}, {  6, 173, 232}, {  1, 105, 178},
                        {  1,  71, 131}, {  1,  44,  84}, {  1,  13,  46},
                    },
                },
            },
            {  // UV plane
                {  // Inter
                    {  // band 0
                        {188,  26, 214}, {121,  42, 181}, { 66,  49, 149},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {136, 128, 233}, {172, 124, 230}, { 80, 106, 211},
                        { 27,  81, 174}, {  6,  49,  98}, {  8,  28,  49},
                    },
                    {  // band 2
                        {145, 166, 239}, { 92, 141, 229}, { 28, 108, 196},
                        {  8,  87, 154}, {  1,  58, 105}, {  1,  27,  59},
                    },
                    {  // band 3
                        {131, 193, 242}, { 66, 151, 231}, { 13, 112, 192},
                        {  2,  81, 152}, {  1,  66, 121}, {  1,  23,  64},
                    },
                    {  // band 4
                        {112, 211, 246}, { 41, 164, 235}, {  5, 117, 202},
                        {  1,  83, 162}, {  1,  64, 111}, {128, 128, 128},
                    },
                    {  // band 5
                        { 96, 230, 250}, { 28, 185, 243}, {  2, 132, 204},
                        {  1,  91, 166}, {  1,  85,  46}, {128, 128, 128},
                    },
                },
                {  // Inter
                    {  // band 0
                        {238,  23, 242}, {157,  29, 215}, { 73,  27, 162},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {165, 173, 250}, {222, 151, 247}, {152, 134, 235},
                        {114, 120, 210}, { 86, 109, 176}, { 53,  88, 145},
                    },
                    {  // band 2
                        {164, 194, 249}, {100, 158, 241}, { 35, 111, 212},
                        { 17,  85, 167}, {  1,  52, 112}, {  1,  73,   1},
                    },
                    {  // band 3
                        {151, 215, 252}, { 83, 172, 245}, { 16, 122, 208},
                        {  6, 101, 165}, {  1,  74, 113}, {  1,   1,   1},
                    },
                    {  // band 4
                        {138, 230, 253}, { 65, 184, 248}, {  8, 128, 212},
                        {  1, 111, 182}, {128,   1,   1}, {128, 128, 128},
                    },
                    {  // band 5
                        {123, 240, 253}, { 36, 201, 250}, {  3, 127, 211},
                        {  1,  68, 204}, {128,   1,   1}, {128, 128, 128},
                    },
                },
            },
        },
        {  // TX_SIZE 3
            {  // Y plane
                {  // Intra
                    {  // band 0
                        { 51,  21, 156}, { 30,  23,  86}, {  4,  18,  37},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 38,  77, 129}, { 79,  76, 129}, { 40,  66, 117},
                        { 12,  54,  95}, {  1,  36,  60}, {  1,  17,  29},
                    },
                    {  // band 2
                        { 44, 133, 149}, { 24, 107, 143}, {  8,  78, 121},
                        {  3,  59,  97}, {  1,  42,  71}, {  1,  22,  37},
                    },
                    {  // band 3
                        { 29, 160, 171}, {  9, 114, 158}, {  1,  76, 125},
                        {  1,  54,  93}, {  1,  36,  63}, {  1,  20,  35},
                    },
                    {  // band 4
                        { 22, 188, 205}, {  6, 132, 186}, {  1,  87, 144},
                        {  1,  62, 107}, {  1,  41,  72}, {  1,  23,  41},
                    },
                    {  // band 5
                        { 25, 233, 236}, {  5, 165, 214}, {  1,  96, 158},
                        {  1,  63, 112}, {  1,  40,  73}, {  1,  23,  40},
                    },
                },
                {  // Intra
                    {  // band 0
                        { 48,  20, 231}, { 37,  21, 179}, { 15,  18, 109},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 41, 154, 216}, {196, 142, 221}, {131, 125, 207},
                        { 84, 111, 181}, { 45,  91, 142}, { 27,  62,  89},
                    },
                    {  // band 2
                        { 72, 181, 230}, { 41, 147, 215}, { 10, 102, 173},
                        {  3,  73, 132}, {  1,  47,  89}, {  1,  23,  50},
                    },
                    {  // band 3
                        { 60, 201, 236}, { 23, 157, 219}, {  2,  99, 167},
                        {  1,  69, 124}, {  1,  43,  80}, {  1,  22,  39},
                    },
                    {  // band 4
                        { 53, 214, 242}, { 15, 165, 224}, {  1, 101, 173},
                        {  1,  70, 131}, {  1,  44,  83}, {  1,  23,  49},
                    },
                    {  // band 5
                        { 39, 239, 248}, {  7, 186, 233}, {  1, 108, 174},
                        {  1,  70, 123}, {  1,  43,  77}, {  1,  16,  42},
                    },
                },
            },
            {  // UV plane
                {  // Inter
                    {  // band 0
                        {161,  26, 204}, { 77,  40, 160}, { 26,  50, 117},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 80, 140, 218}, {136, 133, 215}, { 63, 117, 197},
                        { 20,  93, 170}, {  7,  55, 102}, { 13,  32,  52},
                    },
                    {  // band 2
                        { 86, 173, 231}, { 46, 150, 220}, { 18, 118, 190},
                        {  8,  90, 150}, {  2,  60,  95}, {  1,  39,  41},
                    },
                    {  // band 3
                        { 80, 183, 242}, { 37, 160, 231}, {  6, 120, 182},
                        {  1,  86, 137}, {  1,  46,  78}, {  1,  15,  24},
                    },
                    {  // band 4
                        { 88, 215, 247}, { 42, 179, 235}, {  4, 116, 182},
                        {  2,  80, 133}, {  1,  46,  85}, {  1,  64,  43},
                    },
                    {  // band 5
                        {100, 236, 250}, { 31, 186, 234}, {  1, 114, 181},
                        {  1,  85, 135}, {  1,  78,  64}, {128, 128, 128},
                    },
                },
                {  // Inter
                    {  // band 0
                        {213,  13, 245}, {106,  16, 211}, { 32,  11, 156},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {140, 214, 247}, {241, 186, 243}, {177, 172, 235},
                        {128, 156, 219}, {106, 130, 191}, { 99, 105, 152},
                    },
                    {  // band 2
                        {125, 218, 248}, { 75, 167, 239}, { 29, 111, 212},
                        {  6,  66, 152}, {  1,  42,  96}, {  1,  85, 128},
                    },
                    {  // band 3
                        {120, 232, 252}, { 60, 189, 247}, {  8, 141, 200},
                        {  1,  89, 134}, {  1,  32, 128}, {128, 128, 128},
                    },
                    {  // band 4
                        {111, 238, 253}, { 56, 198, 245}, {  1, 123, 208},
                        {  1,  93, 176}, {  1,   1,  73}, {128, 128, 128},
                    },
                    {  // band 5
                        { 98, 251, 249}, { 56, 189, 244}, { 17, 113, 220},
                        {  1, 109, 179}, {128, 128, 128}, {128, 128, 128},
                    },
                },
            },
        },
#if CONFIG_TX64X64
        {  // TX_SIZE 4
            {  // Y plane
                {  // Intra
                    {  // band 0
                        { 51,  21, 156}, { 30,  23,  86}, {  4,  18,  37},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 38,  77, 129}, { 79,  76, 129}, { 40,  66, 117},
                        { 12,  54,  95}, {  1,  36,  60}, {  1,  17,  29},
                    },
                    {  // band 2
                        { 44, 133, 149}, { 24, 107, 143}, {  8,  78, 121},
                        {  3,  59,  97}, {  1,  42,  71}, {  1,  22,  37},
                    },
                    {  // band 3
                        { 29, 160, 171}, {  9, 114, 158}, {  1,  76, 125},
                        {  1,  54,  93}, {  1,  36,  63}, {  1,  20,  35},
                    },
                    {  // band 4
                        { 22, 188, 205}, {  6, 132, 186}, {  1,  87, 144},
                        {  1,  62, 107}, {  1,  41,  72}, {  1,  23,  41},
                    },
                    {  // band 5
                        { 25, 233, 236}, {  5, 165, 214}, {  1,  96, 158},
                        {  1,  63, 112}, {  1,  40,  73}, {  1,  23,  40},
                    },
                },
                {  // Intra
                    {  // band 0
                        { 48,  20, 231}, { 37,  21, 179}, { 15,  18, 109},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 41, 154, 216}, {196, 142, 221}, {131, 125, 207},
                        { 84, 111, 181}, { 45,  91, 142}, { 27,  62,  89},
                    },
                    {  // band 2
                        { 72, 181, 230}, { 41, 147, 215}, { 10, 102, 173},
                        {  3,  73, 132}, {  1,  47,  89}, {  1,  23,  50},
                    },
                    {  // band 3
                        { 60, 201, 236}, { 23, 157, 219}, {  2,  99, 167},
                        {  1,  69, 124}, {  1,  43,  80}, {  1,  22,  39},
                    },
                    {  // band 4
                        { 53, 214, 242}, { 15, 165, 224}, {  1, 101, 173},
                        {  1,  70, 131}, {  1,  44,  83}, {  1,  23,  49},
                    },
                    {  // band 5
                        { 39, 239, 248}, {  7, 186, 233}, {  1, 108, 174},
                        {  1,  70, 123}, {  1,  43,  77}, {  1,  16,  42},
                    },
                },
            },
            {  // UV plane
                {  // Inter
                    {  // band 0
                        {161,  26, 204}, { 77,  40, 160}, { 26,  50, 117},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 80, 140, 218}, {136, 133, 215}, { 63, 117, 197},
                        { 20,  93, 170}, {  7,  55, 102}, { 13,  32,  52},
                    },
                    {  // band 2
                        { 86, 173, 231}, { 46, 150, 220}, { 18, 118, 190},
                        {  8,  90, 150}, {  2,  60,  95}, {  1,  39,  41},
                    },
                    {  // band 3
                        { 80, 183, 242}, { 37, 160, 231}, {  6, 120, 182},
                        {  1,  86, 137}, {  1,  46,  78}, {  1,  15,  24},
                    },
                    {  // band 4
                        { 88, 215, 247}, { 42, 179, 235}, {  4, 116, 182},
                        {  2,  80, 133}, {  1,  46,  85}, {  1,  64,  43},
                    },
                    {  // band 5
                        {100, 236, 250}, { 31, 186, 234}, {  1, 114, 181},
                        {  1,  85, 135}, {  1,  78,  64}, {128, 128, 128},
                    },
                },
                {  // Inter
                    {  // band 0
                        {213,  13, 245}, {106,  16, 211}, { 32,  11, 156},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {140, 214, 247}, {241, 186, 243}, {177, 172, 235},
                        {128, 156, 219}, {106, 130, 191}, { 99, 105, 152},
                    },
                    {  // band 2
                        {125, 218, 248}, { 75, 167, 239}, { 29, 111, 212},
                        {  6,  66, 152}, {  1,  42,  96}, {  1,  85, 128},
                    },
                    {  // band 3
                        {120, 232, 252}, { 60, 189, 247}, {  8, 141, 200},
                        {  1,  89, 134}, {  1,  32, 128}, {128, 128, 128},
                    },
                    {  // band 4
                        {111, 238, 253}, { 56, 198, 245}, {  1, 123, 208},
                        {  1,  93, 176}, {  1,   1,  73}, {128, 128, 128},
                    },
                    {  // band 5
                        { 98, 251, 249}, { 56, 189, 244}, { 17, 113, 220},
                        {  1, 109, 179}, {128, 128, 128}, {128, 128, 128},
                    },
                },
            },
        },
#endif  // CONFIG_TX64X64
    },
    {  // Q_Index 3
#if CONFIG_CB4X4
        {  // TX_SIZE 0
            {  // Y plane
                {  // Intra
                    {  // band 0
                        {186,  16, 200}, {122,  31, 187}, { 78,  40, 161},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {146, 119, 245}, {182, 115, 244}, {130, 113, 238},
                        { 88, 110, 225}, { 47, 103, 208}, {  5, 102, 188},
                    },
                    {  // band 2
                        {164, 157, 248}, {155, 141, 250}, { 71, 116, 243},
                        { 88, 129, 233}, { 50,  99, 228}, { 26, 148, 191},
                    },
                    {  // band 3
                        {200, 158, 253}, {177, 118, 252}, { 99, 113, 245},
                        { 77, 120, 210}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 4
                        {231, 104, 254}, {209,  82, 254}, {143, 112, 252},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 5
                        {250,  36, 254}, {243,  55, 254}, {223, 170, 254},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                },
                {  // Intra
                    {  // band 0
                        {207,  37, 226}, {164,  46, 218}, {122,  58, 201},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {149, 154, 253}, {170, 137, 253}, { 94, 123, 247},
                        { 42, 113, 222}, { 16,  97, 174}, { 49,  98, 159},
                    },
                    {  // band 2
                        {177, 162, 253}, {165, 142, 252}, { 51, 108, 243},
                        { 18, 108, 213}, {  1,  98, 254}, {128, 128, 128},
                    },
                    {  // band 3
                        {211, 152, 254}, {184, 116, 254}, { 70, 110, 244},
                        {  8, 108, 237}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 4
                        {236,  89, 254}, {210,  67, 254}, {112, 111, 248},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 5
                        {246,  26, 254}, {233,  35, 254}, {128,   1, 254},
                        {254, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                },
            },
            {  // UV plane
                {  // Inter
                    {  // band 0
                        {247,   2, 247}, {226,   8, 242}, {191,  14, 235},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {231,  94, 254}, {248,  91, 254}, {186,  89, 252},
                        {128,  92, 244}, { 79, 112, 254}, {128, 128, 128},
                    },
                    {  // band 2
                        {228, 145, 253}, {240, 130, 254}, {223, 105, 254},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 3
                        {245, 153, 253}, {240, 120, 254}, {128, 128, 128},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 4
                        {254, 128, 254}, {204, 128, 254}, {128, 128, 128},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 5
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                },
                {  // Inter
                    {  // band 0
                        {253,   7, 249}, {224,   9, 244}, {182,  13, 231},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {234, 109, 254}, {242, 104, 254}, {160,  98, 254},
                        {123,  85, 243}, { 82,  43, 217}, {128, 128, 128},
                    },
                    {  // band 2
                        {243, 137, 254}, {240, 118, 254}, {136,  53, 254},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 3
                        {251, 173, 254}, {229, 129, 250}, {128, 128, 128},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 4
                        {254, 119, 254}, {254, 128, 128}, {128, 128, 128},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 5
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                },
            },
        },
#endif
        {  // TX_SIZE 0
            {  // Y plane
                {  // Intra
                    {  // band 0
                        {186,  16, 200}, {122,  31, 187}, { 78,  40, 161},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {146, 119, 245}, {182, 115, 244}, {130, 113, 238},
                        { 88, 110, 225}, { 47, 103, 208}, {  5, 102, 188},
                    },
                    {  // band 2
                        {164, 157, 248}, {155, 141, 250}, { 71, 116, 243},
                        { 88, 129, 233}, { 50,  99, 228}, { 26, 148, 191},
                    },
                    {  // band 3
                        {200, 158, 253}, {177, 118, 252}, { 99, 113, 245},
                        { 77, 120, 210}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 4
                        {231, 104, 254}, {209,  82, 254}, {143, 112, 252},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 5
                        {250,  36, 254}, {243,  55, 254}, {223, 170, 254},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                },
                {  // Intra
                    {  // band 0
                        {207,  37, 226}, {164,  46, 218}, {122,  58, 201},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {149, 154, 253}, {170, 137, 253}, { 94, 123, 247},
                        { 42, 113, 222}, { 16,  97, 174}, { 49,  98, 159},
                    },
                    {  // band 2
                        {177, 162, 253}, {165, 142, 252}, { 51, 108, 243},
                        { 18, 108, 213}, {  1,  98, 254}, {128, 128, 128},
                    },
                    {  // band 3
                        {211, 152, 254}, {184, 116, 254}, { 70, 110, 244},
                        {  8, 108, 237}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 4
                        {236,  89, 254}, {210,  67, 254}, {112, 111, 248},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 5
                        {246,  26, 254}, {233,  35, 254}, {128,   1, 254},
                        {254, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                },
            },
            {  // UV plane
                {  // Inter
                    {  // band 0
                        {247,   2, 247}, {226,   8, 242}, {191,  14, 235},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {231,  94, 254}, {248,  91, 254}, {186,  89, 252},
                        {128,  92, 244}, { 79, 112, 254}, {128, 128, 128},
                    },
                    {  // band 2
                        {228, 145, 253}, {240, 130, 254}, {223, 105, 254},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 3
                        {245, 153, 253}, {240, 120, 254}, {128, 128, 128},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 4
                        {254, 128, 254}, {204, 128, 254}, {128, 128, 128},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 5
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                },
                {  // Inter
                    {  // band 0
                        {253,   7, 249}, {224,   9, 244}, {182,  13, 231},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {234, 109, 254}, {242, 104, 254}, {160,  98, 254},
                        {123,  85, 243}, { 82,  43, 217}, {128, 128, 128},
                    },
                    {  // band 2
                        {243, 137, 254}, {240, 118, 254}, {136,  53, 254},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 3
                        {251, 173, 254}, {229, 129, 250}, {128, 128, 128},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 4
                        {254, 119, 254}, {254, 128, 128}, {128, 128, 128},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 5
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                },
            },
        },
        {  // TX_SIZE 1
            {  // Y plane
                {  // Intra
                    {  // band 0
                        { 49,  26, 159}, { 36,  34, 150}, { 26,  38, 124},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 99, 122, 226}, {143, 119, 225}, { 90, 113, 213},
                        { 46, 102, 193}, { 14,  84, 157}, {  3,  59, 107},
                    },
                    {  // band 2
                        {109, 164, 237}, { 74, 142, 233}, { 29, 112, 216},
                        { 14,  92, 184}, { 10,  80, 156}, {  1,  52, 137},
                    },
                    {  // band 3
                        {110, 191, 245}, { 59, 156, 240}, { 18, 121, 220},
                        {  8,  97, 184}, {  3,  84, 150}, {128, 128, 128},
                    },
                    {  // band 4
                        {115, 203, 250}, { 59, 167, 246}, { 16, 130, 226},
                        {  7,  97, 192}, {  1,  71,  99}, {128, 128, 128},
                    },
                    {  // band 5
                        {149, 218, 253}, { 93, 171, 251}, { 28, 125, 233},
                        { 28,  99, 192}, {128,  85,  85}, {128, 128, 128},
                    },
                },
                {  // Intra
                    {  // band 0
                        { 97,  45, 229}, { 79,  52, 205}, { 46,  58, 171},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 99, 180, 249}, {156, 165, 249}, { 73, 141, 237},
                        { 31, 116, 208}, { 13,  81, 153}, {  5,  42,  86},
                    },
                    {  // band 2
                        {113, 188, 251}, { 68, 161, 244}, { 16, 108, 216},
                        {  6,  81, 168}, {  2,  65, 118}, {128,   1,   1},
                    },
                    {  // band 3
                        {117, 201, 252}, { 62, 171, 248}, { 12, 119, 221},
                        {  5,  90, 182}, {  4,  66, 116}, {128, 128, 128},
                    },
                    {  // band 4
                        {128, 207, 253}, { 70, 176, 251}, { 11, 126, 228},
                        {  6,  89, 189}, {  1,  44, 148}, {128, 128, 128},
                    },
                    {  // band 5
                        {162, 218, 254}, {107, 170, 253}, { 22, 131, 238},
                        {  1,  77, 182}, {  1, 254, 128}, {128, 128, 128},
                    },
                },
            },
            {  // UV plane
                {  // Inter
                    {  // band 0
                        {235,   5, 238}, {194,  14, 223}, {152,  22, 205},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {200, 121, 251}, {241, 115, 252}, {167, 108, 248},
                        { 93,  93, 233}, { 36,  66, 189}, {128, 128, 128},
                    },
                    {  // band 2
                        {220, 151, 253}, {176, 135, 252}, { 95, 124, 254},
                        { 64, 105, 217}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 3
                        {225, 189, 254}, {175, 155, 254}, {102, 119, 254},
                        {  1,   1,   1}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 4
                        {218, 195, 254}, {125, 157, 253}, {128, 128, 254},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 5
                        {221, 197, 254}, { 85, 210, 254}, {128, 128, 128},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                },
                {  // Inter
                    {  // band 0
                        {250,   9, 246}, {204,  13, 234}, {144,  18, 211},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {213, 157, 253}, {243, 138, 253}, {170, 117, 250},
                        {109,  91, 233}, { 66,  77, 163}, { 64,  85, 254},
                    },
                    {  // band 2
                        {221, 169, 254}, {182, 141, 253}, {112, 120, 239},
                        { 85, 165, 254}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 3
                        {226, 192, 254}, {189, 174, 251}, {153, 128, 254},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 4
                        {232, 192, 254}, {195, 187, 247}, {  1, 191, 254},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 5
                        {247, 185, 254}, {254,  93, 254}, {128, 128, 128},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                },
            },
        },
        {  // TX_SIZE 2
            {  // Y plane
                {  // Intra
                    {  // band 0
                        { 14,  30, 136}, { 15,  33, 120}, { 10,  33,  90},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 92, 109, 209}, {113, 108, 207}, { 77, 102, 193},
                        { 39,  91, 171}, { 11,  70, 129}, {  2,  44,  77},
                    },
                    {  // band 2
                        { 99, 158, 223}, { 66, 135, 217}, { 23, 109, 194},
                        {  9,  85, 160}, {  3,  66, 124}, {  1,  51, 100},
                    },
                    {  // band 3
                        { 89, 189, 234}, { 46, 149, 225}, { 10, 110, 194},
                        {  2,  83, 156}, {  1,  57, 113}, {  1,  47,  73},
                    },
                    {  // band 4
                        { 78, 206, 242}, { 28, 161, 232}, {  3, 114, 200},
                        {  1,  86, 161}, {  1,  62, 118}, {  1,   1,   1},
                    },
                    {  // band 5
                        { 72, 227, 250}, { 20, 182, 242}, {  3, 126, 210},
                        {  2,  91, 166}, {  1,  64, 126}, {128, 128, 128},
                    },
                },
                {  // Intra
                    {  // band 0
                        { 23,  42, 227}, { 41,  43, 195}, { 25,  45, 146},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {100, 172, 245}, {165, 158, 246}, { 88, 137, 234},
                        { 44, 116, 203}, { 18,  85, 149}, {  7,  56,  92},
                    },
                    {  // band 2
                        {117, 188, 247}, { 70, 155, 239}, { 18, 105, 204},
                        {  7,  78, 158}, {  2,  50, 111}, {  1,  38,  77},
                    },
                    {  // band 3
                        {104, 207, 250}, { 54, 166, 241}, {  6, 110, 199},
                        {  1,  78, 155}, {  1,  45, 100}, {  1,   1,   1},
                    },
                    {  // band 4
                        { 87, 216, 251}, { 30, 177, 243}, {  1, 114, 203},
                        {  1,  85, 157}, {  1,  53, 108}, {128, 128, 128},
                    },
                    {  // band 5
                        { 80, 230, 253}, { 23, 193, 248}, {  1, 127, 215},
                        {  1,  94, 170}, {  1,  71,  59}, {128, 128, 128},
                    },
                },
            },
            {  // UV plane
                {  // Inter
                    {  // band 0
                        {222,   9, 234}, {161,  20, 210}, {113,  30, 185},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {195, 120, 248}, {231, 124, 247}, {148, 116, 238},
                        { 64,  98, 207}, { 20,  70, 147}, { 87,  68, 100},
                    },
                    {  // band 2
                        {186, 161, 250}, {124, 148, 245}, { 44, 123, 230},
                        { 23, 107, 205}, {  1,  80, 131}, {128, 128, 128},
                    },
                    {  // band 3
                        {172, 196, 252}, {110, 160, 248}, { 37, 134, 235},
                        { 23, 125, 200}, {128, 254, 128}, {128, 128, 128},
                    },
                    {  // band 4
                        {173, 209, 253}, {103, 175, 250}, {  1, 120, 240},
                        {  1, 146, 254}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 5
                        {184, 235, 254}, { 81, 186, 251}, {128, 109, 254},
                        {128, 254, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                },
                {  // Inter
                    {  // band 0
                        {248,   8, 243}, {185,  11, 225}, {108,  11, 189},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {208, 158, 254}, {244, 147, 252}, {195, 132, 248},
                        {161, 122, 224}, {129, 114, 188}, { 59, 119, 159},
                    },
                    {  // band 2
                        {202, 182, 253}, {143, 161, 251}, { 73, 115, 247},
                        {146, 175, 204}, {128,   1, 254}, {128, 128, 128},
                    },
                    {  // band 3
                        {202, 204, 254}, {131, 174, 251}, { 18, 153, 207},
                        {128, 254, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 4
                        {192, 221, 254}, {114, 190, 254}, {128, 170, 254},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 5
                        {166, 236, 254}, {119, 200, 254}, {128, 128, 128},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                },
            },
        },
        {  // TX_SIZE 3
            {  // Y plane
                {  // Intra
                    {  // band 0
                        { 30,  32, 144}, { 21,  35,  96}, {  4,  27,  55},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 35, 107, 172}, { 61, 104, 170}, { 33,  94, 160},
                        { 13,  80, 139}, {  2,  55,  97}, {  1,  28,  49},
                    },
                    {  // band 2
                        { 51, 153, 195}, { 29, 129, 189}, {  9,  99, 163},
                        {  3,  75, 129}, {  1,  49,  88}, {  1,  29,  50},
                    },
                    {  // band 3
                        { 53, 164, 210}, { 21, 134, 201}, {  3,  97, 164},
                        {  1,  69, 124}, {  1,  45,  82}, {  1,  31,  58},
                    },
                    {  // band 4
                        { 47, 205, 234}, { 18, 158, 220}, {  2, 109, 177},
                        {  1,  78, 137}, {  1,  53, 101}, {  1,  34,  70},
                    },
                    {  // band 5
                        { 55, 233, 245}, { 16, 179, 233}, {  1, 116, 191},
                        {  1,  79, 145}, {  1,  53, 101}, {  1,  37,  58},
                    },
                },
                {  // Intra
                    {  // band 0
                        { 36,  33, 227}, { 39,  28, 190}, { 18,  27, 134},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 76, 156, 235}, {184, 147, 235}, {114, 130, 220},
                        { 72, 112, 191}, { 42,  87, 144}, { 21,  65,  93},
                    },
                    {  // band 2
                        { 96, 179, 240}, { 51, 149, 228}, { 12, 105, 191},
                        {  6,  74, 148}, {  1,  47, 100}, {  1,  29,  53},
                    },
                    {  // band 3
                        { 88, 191, 242}, { 35, 154, 231}, {  3, 106, 187},
                        {  1,  74, 140}, {  1,  41,  84}, {  1,  25,  38},
                    },
                    {  // band 4
                        { 77, 212, 249}, { 28, 171, 239}, {  2, 117, 199},
                        {  1,  79, 151}, {  1,  45,  99}, {  1,   1,   1},
                    },
                    {  // band 5
                        { 77, 236, 252}, { 27, 190, 246}, {  2, 120, 203},
                        {  1,  78, 147}, {  1,  42,  72}, {128, 128, 128},
                    },
                },
            },
            {  // UV plane
                {  // Inter
                    {  // band 0
                        {185,  11, 227}, {113,  30, 182}, { 57,  44, 144},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {151, 139, 244}, {212, 139, 241}, {124, 126, 231},
                        { 59, 104, 213}, { 26,  73, 158}, { 20,  45,  95},
                    },
                    {  // band 2
                        {155, 163, 247}, {108, 152, 239}, { 39, 124, 214},
                        {  7, 109, 162}, { 29,  57, 128}, {128, 128, 128},
                    },
                    {  // band 3
                        {158, 176, 250}, { 89, 164, 243}, { 11, 114, 196},
                        {  1,  96, 141}, {  1,  81, 118}, {128,   1,   1},
                    },
                    {  // band 4
                        {148, 212, 251}, { 59, 174, 240}, {  2, 130, 203},
                        {  1,  70, 168}, {  1,  51, 106}, {128, 128, 128},
                    },
                    {  // band 5
                        {104, 237, 252}, { 39, 190, 246}, {  1, 154, 220},
                        {128, 102,   1}, {128, 128, 128}, {128, 128, 128},
                    },
                },
                {  // Inter
                    {  // band 0
                        {236,   6, 242}, {111,   6, 206}, { 36,   5, 161},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {193, 193, 252}, {248, 182, 251}, {218, 150, 246},
                        {182, 134, 244}, {151, 137, 227}, { 45, 102, 195},
                    },
                    {  // band 2
                        {188, 202, 251}, {125, 165, 249}, { 64,  75, 218},
                        {  1, 128, 254}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 3
                        {178, 225, 254}, {107, 188, 231}, { 21, 135, 233},
                        {128,   1, 254}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 4
                        {164, 227, 253}, { 55, 193, 251}, {  1, 111, 225},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 5
                        {151, 243, 254}, { 50, 203, 254}, {128, 179, 254},
                        {128,   1, 254}, {128, 128, 128}, {128, 128, 128},
                    },
                },
            },
        },
#if CONFIG_TX64X64
        {  // TX_SIZE 4
            {  // Y plane
                {  // Intra
                    {  // band 0
                        { 30,  32, 144}, { 21,  35,  96}, {  4,  27,  55},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 35, 107, 172}, { 61, 104, 170}, { 33,  94, 160},
                        { 13,  80, 139}, {  2,  55,  97}, {  1,  28,  49},
                    },
                    {  // band 2
                        { 51, 153, 195}, { 29, 129, 189}, {  9,  99, 163},
                        {  3,  75, 129}, {  1,  49,  88}, {  1,  29,  50},
                    },
                    {  // band 3
                        { 53, 164, 210}, { 21, 134, 201}, {  3,  97, 164},
                        {  1,  69, 124}, {  1,  45,  82}, {  1,  31,  58},
                    },
                    {  // band 4
                        { 47, 205, 234}, { 18, 158, 220}, {  2, 109, 177},
                        {  1,  78, 137}, {  1,  53, 101}, {  1,  34,  70},
                    },
                    {  // band 5
                        { 55, 233, 245}, { 16, 179, 233}, {  1, 116, 191},
                        {  1,  79, 145}, {  1,  53, 101}, {  1,  37,  58},
                    },
                },
                {  // Intra
                    {  // band 0
                        { 36,  33, 227}, { 39,  28, 190}, { 18,  27, 134},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 76, 156, 235}, {184, 147, 235}, {114, 130, 220},
                        { 72, 112, 191}, { 42,  87, 144}, { 21,  65,  93},
                    },
                    {  // band 2
                        { 96, 179, 240}, { 51, 149, 228}, { 12, 105, 191},
                        {  6,  74, 148}, {  1,  47, 100}, {  1,  29,  53},
                    },
                    {  // band 3
                        { 88, 191, 242}, { 35, 154, 231}, {  3, 106, 187},
                        {  1,  74, 140}, {  1,  41,  84}, {  1,  25,  38},
                    },
                    {  // band 4
                        { 77, 212, 249}, { 28, 171, 239}, {  2, 117, 199},
                        {  1,  79, 151}, {  1,  45,  99}, {  1,   1,   1},
                    },
                    {  // band 5
                        { 77, 236, 252}, { 27, 190, 246}, {  2, 120, 203},
                        {  1,  78, 147}, {  1,  42,  72}, {128, 128, 128},
                    },
                },
            },
            {  // UV plane
                {  // Inter
                    {  // band 0
                        {185,  11, 227}, {113,  30, 182}, { 57,  44, 144},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {151, 139, 244}, {212, 139, 241}, {124, 126, 231},
                        { 59, 104, 213}, { 26,  73, 158}, { 20,  45,  95},
                    },
                    {  // band 2
                        {155, 163, 247}, {108, 152, 239}, { 39, 124, 214},
                        {  7, 109, 162}, { 29,  57, 128}, {128, 128, 128},
                    },
                    {  // band 3
                        {158, 176, 250}, { 89, 164, 243}, { 11, 114, 196},
                        {  1,  96, 141}, {  1,  81, 118}, {128,   1,   1},
                    },
                    {  // band 4
                        {148, 212, 251}, { 59, 174, 240}, {  2, 130, 203},
                        {  1,  70, 168}, {  1,  51, 106}, {128, 128, 128},
                    },
                    {  // band 5
                        {104, 237, 252}, { 39, 190, 246}, {  1, 154, 220},
                        {128, 102,   1}, {128, 128, 128}, {128, 128, 128},
                    },
                },
                {  // Inter
                    {  // band 0
                        {236,   6, 242}, {111,   6, 206}, { 36,   5, 161},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {193, 193, 252}, {248, 182, 251}, {218, 150, 246},
                        {182, 134, 244}, {151, 137, 227}, { 45, 102, 195},
                    },
                    {  // band 2
                        {188, 202, 251}, {125, 165, 249}, { 64,  75, 218},
                        {  1, 128, 254}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 3
                        {178, 225, 254}, {107, 188, 231}, { 21, 135, 233},
                        {128,   1, 254}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 4
                        {164, 227, 253}, { 55, 193, 251}, {  1, 111, 225},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 5
                        {151, 243, 254}, { 50, 203, 254}, {128, 179, 254},
                        {128,   1, 254}, {128, 128, 128}, {128, 128, 128},
                    },
                },
            },
        },
#endif  // CONFIG_TX64X64
    },
};
#else
#if CONFIG_NEW_TOKENSET
static const av1_coeff_probs_model default_coef_probs_4x4[PLANE_TYPES] = {
  { // Y plane
    { // Intra
      { // Band 0
        {97, 27, 144}, {81, 38, 128}, {51, 42, 99}
      },
      { // Band 1
        {74, 113, 204}, {68, 101, 199}, {50, 87, 173},
        {31, 76, 133}, {13, 55, 86}, {3, 30, 39}
      },
      { // Band 2
        {83, 156, 222}, {74, 127, 215}, {46, 101, 179},
        {30, 80, 129}, {14, 57, 81}, {3, 27, 37}
      },
      { // Band 3
        {105, 164, 233}, {84, 128, 224}, {49, 92, 175},
        {28, 60, 114}, {12, 34, 53}, {20, 59, 98}
      },
      { // Band 4
        {131, 159, 243}, {98, 123, 228}, {40, 78, 151},
        {19, 46, 97}, {13, 47, 19}, {19, 16, 19}
      },
      { // Band 5
        {192, 71, 241}, {174, 70, 226}, {125, 46, 153},
        {108, 49, 116}, {82, 24, 46}, {60, 14, 30}
      }
    },
    { // Inter
      { // Band 0
        {111, 66, 218}, {107, 87, 211}, {93, 99, 207}
      },
      { // Band 1
        {107, 166, 250}, {107, 143, 247}, {73, 119, 221},
        {43, 91, 166}, {17, 74, 102}, {3, 70, 53}
      },
      { // Band 2
        {126, 177, 251}, {109, 148, 246}, {64, 99, 204},
        {42, 68, 140}, {28, 52, 84}, {20, 34, 1}
      },
      { // Band 3
        {143, 178, 252}, {114, 144, 245}, {46, 92, 188},
        {45, 65, 104}, {40, 44, 76}, {1, 1, 1}
      },
      { // Band 4
        {163, 159, 251}, {120, 131, 243}, {47, 81, 182},
        {32, 39, 128}, {33, 44, 56}, {1, 17, 34}
      },
      { // Band 5
        {209, 94, 251}, {190, 81, 241}, {139, 45, 147},
        {123, 35, 73}, {118, 1, 118}, {3, 16, 42}
      }
    }
  },
  { // UV plane
    { // Intra
      { // Band 0
        {189, 37, 229}, {145, 68, 205}, {99, 74, 171}
      },
      { // Band 1
        {153, 139, 242}, {135, 125, 235}, {84, 100, 200},
        {49, 75, 162}, {9, 21, 84}, {3, 31, 69}
      },
      { // Band 2
        {165, 165, 244}, {128, 144, 240}, {68, 94, 204},
        {39, 72, 132}, {22, 44, 93}, {26, 73, 26}
      },
      { // Band 3
        {181, 174, 246}, {142, 132, 241}, {81, 96, 212},
        {41, 70, 166}, {9, 48, 92}, {1, 19, 38}
      },
      { // Band 4
        {197, 159, 251}, {168, 121, 245}, {107, 75, 218},
        {70, 43, 158}, {1, 128, 1}, {1, 18, 37}
      },
      { // Band 5
        {231, 79, 255}, {211, 74, 249}, {157, 104, 210},
        {128, 102, 213}, {12, 34, 96}, {2, 20, 47}
      }
    },
    { // Inter
      { // Band 0
        {220, 53, 252}, {191, 80, 248}, {154, 100, 245}
      },
      { // Band 1
        {205, 153, 255}, {182, 147, 254}, {110, 131, 231},
        {68, 114, 161}, {50, 114, 140}, {1, 33, 57}
      },
      { // Band 2
        {213, 171, 255}, {184, 163, 254}, {116, 104, 235},
        {79, 71, 207}, {1, 41, 79}, {1, 20, 39}
      },
      { // Band 3
        {223, 158, 255}, {203, 137, 255}, {111, 142, 244},
        {2, 255, 133}, {1, 44, 85}, {1, 22, 47}
      },
      { // Band 4
        {232, 148, 255}, {222, 123, 255}, {255, 128, 255},
        {3, 61, 124}, {1, 41, 84}, {1, 21, 52}
      },
      { // Band 5
        {248, 92, 255}, {248, 96, 255}, {69, 58, 184},
        {31, 44, 137}, {14, 38, 105}, {8, 23, 61}
      }
    }
  }
};
static const av1_coeff_probs_model default_coef_probs_8x8[PLANE_TYPES] = {
  { // Y plane
    { // Intra
      { // Band 0
        {112, 31, 159}, {72, 37, 119}, {22, 35, 68}
      },
      { // Band 1
        {42, 109, 174}, {45, 99, 172}, {32, 84, 149},
        {18, 69, 119}, {6, 46, 76}, {1, 19, 31}
      },
      { // Band 2
        {40, 154, 202}, {35, 126, 191}, {19, 98, 160},
        {10, 75, 122}, {5, 53, 82}, {1, 23, 39}
      },
      { // Band 3
        {39, 176, 215}, {28, 135, 200}, {11, 93, 156},
        {5, 63, 109}, {1, 36, 64}, {1, 14, 26}
      },
      { // Band 4
        {41, 191, 230}, {25, 147, 212}, {9, 97, 160},
        {3, 65, 109}, {1, 33, 58}, {1, 14, 20}
      },
      { // Band 5
        {68, 203, 242}, {40, 159, 220}, {12, 97, 153},
        {5, 58, 97}, {1, 29, 55}, {1, 11, 18}
      }
    },
    { // Inter
      { // Band 0
        {99, 67, 221}, {86, 80, 204}, {60, 87, 184}
      },
      { // Band 1
        {73, 169, 246}, {79, 158, 242}, {50, 135, 220},
        {30, 113, 181}, {18, 76, 126}, {5, 54, 85}
      },
      { // Band 2
        {90, 184, 250}, {78, 162, 243}, {47, 118, 214},
        {35, 85, 171}, {32, 53, 115}, {20, 28, 76}
      },
      { // Band 3
        {109, 197, 252}, {89, 172, 247}, {52, 119, 217},
        {37, 80, 161}, {23, 44, 100}, {1, 18, 34}
      },
      { // Band 4
        {132, 202, 254}, {110, 175, 251}, {63, 128, 228},
        {37, 86, 168}, {64, 91, 102}, {1, 17, 34}
      },
      { // Band 5
        {126, 204, 253}, {100, 174, 250}, {50, 148, 237},
        {25, 90, 133}, {1, 64, 85}, {3, 16, 42}
      }
    }
  },
  { // UV plane
    { // Intra
      { // Band 0
        {195, 35, 235}, {137, 63, 201}, {62, 70, 145}
      },
      { // Band 1
        {110, 158, 233}, {102, 143, 227}, {60, 120, 199},
        {30, 85, 156}, {9, 50, 90}, {1, 16, 33}
      },
      { // Band 2
        {102, 185, 233}, {71, 152, 224}, {29, 111, 187},
        {18, 74, 138}, {4, 56, 87}, {1, 18, 46}
      },
      { // Band 3
        {101, 205, 239}, {66, 161, 229}, {23, 109, 183},
        {9, 85, 135}, {5, 71, 142}, {1, 1, 102}
      },
      { // Band 4
        {109, 216, 243}, {69, 168, 233}, {23, 119, 191},
        {8, 137, 115}, {1, 54, 98}, {1, 1, 255}
      },
      { // Band 5
        {139, 224, 249}, {98, 176, 238}, {55, 129, 187},
        {25, 101, 131}, {26, 59, 154}, {2, 20, 47}
      }
    },
    { // Inter
      { // Band 0
        {220, 72, 254}, {176, 108, 251}, {114, 132, 247}
      },
      { // Band 1
        {161, 185, 255}, {141, 185, 254}, {131, 180, 249},
        {111, 164, 186}, {50, 98, 142}, {1, 128, 1}
      },
      { // Band 2
        {171, 195, 255}, {133, 184, 254}, {68, 140, 231},
        {102, 96, 205}, {1, 1, 128}, {1, 20, 39}
      },
      { // Band 3
        {180, 206, 255}, {148, 191, 254}, {83, 157, 241},
        {128, 171, 128}, {1, 44, 85}, {1, 22, 47}
      },
      { // Band 4
        {194, 214, 255}, {159, 188, 255}, {122, 148, 250},
        {3, 255, 124}, {1, 41, 84}, {1, 21, 52}
      },
      { // Band 5
        {231, 217, 255}, {209, 149, 255}, {205, 145, 205},
        {31, 44, 137}, {14, 38, 105}, {8, 23, 61}
      }
    }
  }
};
static const av1_coeff_probs_model default_coef_probs_16x16[PLANE_TYPES] = {
  { // Y plane
    { // Intra
      { // Band 0
        {91, 31, 117}, {49, 31, 89}, {14, 25, 48}
      },
      { // Band 1
        {31, 97, 151}, {33, 89, 148}, {28, 76, 133},
        {17, 60, 106}, {7, 42, 72}, {1, 19, 32}
      },
      { // Band 2
        {28, 152, 182}, {28, 120, 174}, {15, 93, 146},
        {9, 72, 116}, {5, 47, 82}, {1, 21, 37}
      },
      { // Band 3
        {29, 174, 203}, {23, 127, 187}, {9, 89, 145},
        {2, 56, 100}, {1, 31, 56}, {1, 12, 25}
      },
      { // Band 4
        {28, 193, 220}, {17, 141, 197}, {4, 87, 142},
        {1, 54, 95}, {1, 31, 56}, {1, 12, 26}
      },
      { // Band 5
        {29, 221, 240}, {11, 167, 215}, {2, 93, 149},
        {1, 58, 100}, {1, 35, 61}, {1, 16, 28}
      }
    },
    { // Inter
      { // Band 0
        {108, 52, 214}, {84, 60, 186}, {45, 69, 161}
      },
      { // Band 1
        {43, 164, 236}, {57, 161, 233}, {38, 146, 214},
        {24, 120, 182}, {15, 80, 126}, {5, 28, 66}
      },
      { // Band 2
        {58, 187, 242}, {47, 163, 234}, {28, 118, 204},
        {26, 82, 165}, {21, 54, 112}, {4, 28, 55}
      },
      { // Band 3
        {65, 201, 248}, {51, 170, 239}, {22, 117, 204},
        {11, 81, 159}, {10, 43, 102}, {1, 1, 1}
      },
      { // Band 4
        {80, 206, 252}, {57, 179, 245}, {25, 129, 214},
        {16, 97, 170}, {6, 60, 130}, {1, 128, 1}
      },
      { // Band 5
        {97, 217, 253}, {68, 186, 250}, {26, 138, 216},
        {20, 105, 166}, {11, 78, 111}, {3, 16, 42}
      }
    }
  },
  { // UV plane
    { // Intra
      { // Band 0
        {181, 37, 233}, {121, 55, 192}, {46, 52, 124}
      },
      { // Band 1
        {108, 157, 221}, {98, 140, 215}, {59, 124, 187},
        {34, 92, 158}, {9, 68, 112}, {1, 41, 70}
      },
      { // Band 2
        {80, 188, 223}, {46, 153, 204}, {25, 91, 173},
        {11, 73, 131}, {5, 43, 82}, {1, 17, 91}
      },
      { // Band 3
        {63, 209, 228}, {31, 157, 206}, {8, 104, 167},
        {3, 63, 122}, {1, 44, 87}, {1, 43, 51}
      },
      { // Band 4
        {52, 220, 234}, {22, 165, 216}, {4, 104, 163},
        {2, 62, 129}, {1, 33, 50}, {1, 26, 28}
      },
      { // Band 5
        {58, 238, 242}, {24, 183, 224}, {4, 109, 172},
        {2, 87, 141}, {1, 52, 79}, {1, 51, 64}
      }
    },
    { // Inter
      { // Band 0
        {224, 52, 250}, {188, 81, 239}, {138, 114, 228}
      },
      { // Band 1
        {131, 206, 255}, {128, 193, 254}, {119, 173, 247},
        {106, 127, 187}, {50, 100, 124}, {1, 96, 1}
      },
      { // Band 2
        {123, 214, 254}, {86, 194, 254}, {64, 119, 221},
        {43, 51, 128}, {1, 32, 110}, {1, 20, 39}
      },
      { // Band 3
        {115, 223, 255}, {78, 200, 254}, {75, 164, 203},
        {128, 85, 255}, {1, 44, 85}, {1, 22, 47}
      },
      { // Band 4
        {132, 226, 255}, {88, 207, 254}, {20, 140, 225},
        {3, 61, 124}, {1, 41, 84}, {1, 21, 52}
      },
      { // Band 5
        {180, 236, 255}, {138, 223, 254}, {73, 166, 238},
        {31, 255, 137}, {14, 38, 105}, {8, 23, 61}
      }
    }
  }
};
static const av1_coeff_probs_model default_coef_probs_32x32[PLANE_TYPES] = {
  { // Y plane
    { // Intra
      { // Band 0
        {163, 26, 188}, {78, 29, 105}, {22, 22, 48}
      },
      { // Band 1
        {72, 93, 168}, {74, 91, 170}, {62, 72, 151},
        {37, 55, 112}, {10, 33, 63}, {1, 14, 23}
      },
      { // Band 2
        {41, 163, 182}, {36, 136, 177}, {20, 102, 153},
        {10, 76, 114}, {5, 45, 71}, {1, 17, 27}
      },
      { // Band 3
        {43, 202, 213}, {28, 142, 193}, {10, 90, 141},
        {2, 51, 93}, {1, 24, 48}, {1, 10, 19}
      },
      { // Band 4
        {46, 216, 220}, {26, 150, 199}, {7, 87, 136},
        {2, 49, 86}, {1, 28, 47}, {1, 12, 24}
      },
      { // Band 5
        {19, 241, 237}, {5, 172, 200}, {1, 82, 126},
        {1, 47, 79}, {1, 29, 47}, {1, 14, 25}
      }
    },
    { // Inter
      { // Band 0
        {185, 20, 226}, {151, 26, 187}, {109, 34, 144}
      },
      { // Band 1
        {56, 151, 227}, {76, 165, 232}, {62, 161, 222},
        {47, 139, 201}, {29, 106, 150}, {14, 61, 98}
      },
      { // Band 2
        {57, 200, 237}, {43, 164, 227}, {22, 106, 190},
        {14, 68, 140}, {10, 48, 90}, {1, 15, 40}
      },
      { // Band 3
        {46, 209, 238}, {28, 165, 225}, {7, 107, 180},
        {2, 69, 125}, {2, 36, 94}, {1, 1, 1}
      },
      { // Band 4
        {55, 225, 248}, {28, 181, 237}, {7, 117, 198},
        {6, 77, 144}, {3, 60, 90}, {1, 1, 1}
      },
      { // Band 5
        {63, 243, 251}, {27, 193, 242}, {4, 124, 200},
        {1, 58, 153}, {1, 59, 124}, {3, 16, 42}
      }
    }
  },
  { // UV plane
    { // Intra
      { // Band 0
        {208, 28, 218}, {183, 32, 188}, {169, 21, 189}
      },
      { // Band 1
        {205, 124, 247}, {190, 96, 240}, {233, 89, 233},
        {177, 44, 212}, {59, 58, 59}, {32, 33, 38}
      },
      { // Band 2
        {194, 195, 250}, {179, 190, 226}, {32, 174, 128},
        {32, 85, 128}, {12, 64, 122}, {1, 85, 90}
      },
      { // Band 3
        {149, 232, 249}, {95, 159, 227}, {28, 91, 171},
        {28, 102, 114}, {1, 1, 73}, {1, 19, 38}
      },
      { // Band 4
        {154, 239, 246}, {138, 151, 235}, {1, 123, 138},
        {128, 183, 255}, {1, 128, 1}, {1, 18, 37}
      },
      { // Band 5
        {157, 255, 253}, {75, 171, 241}, {43, 102, 171},
        {30, 44, 136}, {12, 34, 96}, {2, 20, 47}
      }
    },
    { // Inter
      { // Band 0
        {249, 13, 248}, {238, 14, 220}, {225, 16, 174}
      },
      { // Band 1
        {190, 189, 254}, {169, 134, 253}, {124, 179, 248},
        {138, 131, 223}, {64, 133, 192}, {1, 85, 128}
      },
      { // Band 2
        {139, 212, 254}, {126, 177, 255}, {93, 39, 186},
        {1, 1, 171}, {1, 41, 79}, {1, 20, 39}
      },
      { // Band 3
        {153, 216, 255}, {165, 204, 255}, {1, 1, 255},
        {2, 73, 133}, {1, 1, 1}, {1, 22, 47}
      },
      { // Band 4
        {147, 226, 254}, {119, 196, 255}, {1, 128, 255},
        {1, 1, 171}, {1, 1, 1}, {1, 21, 52}
      },
      { // Band 5
        {168, 240, 255}, {95, 179, 255}, {1, 171, 1},
        {31, 44, 137}, {14, 38, 105}, {8, 23, 61}
      }
    }
  }
};
#else  // CONFIG_NEW_TOKENSET
static const av1_coeff_probs_model default_coef_probs_4x4[PLANE_TYPES] = {
  {  // Y plane
    {  // Intra
      {  // Band 0
        { 195,  29, 183 }, {  84,  49, 136 }, {   8,  42,  71 }
      }, {  // Band 1
        {  31, 107, 169 }, {  35,  99, 159 }, {  17,  82, 140 },
        {   8,  66, 114 }, {   2,  44,  76 }, {   1,  19,  32 }
      }, {  // Band 2
        {  40, 132, 201 }, {  29, 114, 187 }, {  13,  91, 157 },
        {   7,  75, 127 }, {   3,  58,  95 }, {   1,  28,  47 }
      }, {  // Band 3
        {  69, 142, 221 }, {  42, 122, 201 }, {  15,  91, 159 },
        {   6,  67, 121 }, {   1,  42,  77 }, {   1,  17,  31 }
      }, {  // Band 4
        { 102, 148, 228 }, {  67, 117, 204 }, {  17,  82, 154 },
        {   6,  59, 114 }, {   2,  39,  75 }, {   1,  15,  29 }
      }, {  // Band 5
        { 156,  57, 233 }, { 119,  57, 212 }, {  58,  48, 163 },
        {  29,  40, 124 }, {  12,  30,  81 }, {   3,  12,  31 }
      }
    }, {  // Inter
      {  // Band 0
        { 191, 107, 226 }, { 124, 117, 204 }, {  25,  99, 155 }
      }, {  // Band 1
        {  29, 148, 210 }, {  37, 126, 194 }, {   8,  93, 157 },
        {   2,  68, 118 }, {   1,  39,  69 }, {   1,  17,  33 }
      }, {  // Band 2
        {  41, 151, 213 }, {  27, 123, 193 }, {   3,  82, 144 },
        {   1,  58, 105 }, {   1,  32,  60 }, {   1,  13,  26 }
      }, {  // Band 3
        {  59, 159, 220 }, {  23, 126, 198 }, {   4,  88, 151 },
        {   1,  66, 114 }, {   1,  38,  71 }, {   1,  18,  34 }
      }, {  // Band 4
        { 114, 136, 232 }, {  51, 114, 207 }, {  11,  83, 155 },
        {   3,  56, 105 }, {   1,  33,  65 }, {   1,  17,  34 }
      }, {  // Band 5
        { 149,  65, 234 }, { 121,  57, 215 }, {  61,  49, 166 },
        {  28,  36, 114 }, {  12,  25,  76 }, {   3,  16,  42 }
      }
    }
  }, {  // UV plane
    {  // Intra
      {  // Band 0
        { 214,  49, 220 }, { 132,  63, 188 }, {  42,  65, 137 }
      }, {  // Band 1
        {  85, 137, 221 }, { 104, 131, 216 }, {  49, 111, 192 },
        {  21,  87, 155 }, {   2,  49,  87 }, {   1,  16,  28 }
      }, {  // Band 2
        {  89, 163, 230 }, {  90, 137, 220 }, {  29, 100, 183 },
        {  10,  70, 135 }, {   2,  42,  81 }, {   1,  17,  33 }
      }, {  // Band 3
        { 108, 167, 237 }, {  55, 133, 222 }, {  15,  97, 179 },
        {   4,  72, 135 }, {   1,  45,  85 }, {   1,  19,  38 }
      }, {  // Band 4
        { 124, 146, 240 }, {  66, 124, 224 }, {  17,  88, 175 },
        {   4,  58, 122 }, {   1,  36,  75 }, {   1,  18,  37 }
      }, {  //  Band 5
        { 141,  79, 241 }, { 126,  70, 227 }, {  66,  58, 182 },
        {  30,  44, 136 }, {  12,  34,  96 }, {   2,  20,  47 }
      }
    }, {  // Inter
      {  // Band 0
        { 229,  99, 249 }, { 143, 111, 235 }, {  46, 109, 192 }
      }, {  // Band 1
        {  82, 158, 236 }, {  94, 146, 224 }, {  25, 117, 191 },
        {   9,  87, 149 }, {   3,  56,  99 }, {   1,  33,  57 }
      }, {  // Band 2
        {  83, 167, 237 }, {  68, 145, 222 }, {  10, 103, 177 },
        {   2,  72, 131 }, {   1,  41,  79 }, {   1,  20,  39 }
      }, {  // Band 3
        {  99, 167, 239 }, {  47, 141, 224 }, {  10, 104, 178 },
        {   2,  73, 133 }, {   1,  44,  85 }, {   1,  22,  47 }
      }, {  // Band 4
        { 127, 145, 243 }, {  71, 129, 228 }, {  17,  93, 177 },
        {   3,  61, 124 }, {   1,  41,  84 }, {   1,  21,  52 }
      }, {  // Band 5
        { 157,  78, 244 }, { 140,  72, 231 }, {  69,  58, 184 },
        {  31,  44, 137 }, {  14,  38, 105 }, {   8,  23,  61 }
      }
    }
  }
};

static const av1_coeff_probs_model default_coef_probs_8x8[PLANE_TYPES] = {
  {  // Y plane
    {  // Intra
      {  // Band 0
        { 125,  34, 187 }, {  52,  41, 133 }, {   6,  31,  56 }
      }, {  // Band 1
        {  37, 109, 153 }, {  51, 102, 147 }, {  23,  87, 128 },
        {   8,  67, 101 }, {   1,  41,  63 }, {   1,  19,  29 }
      }, {  // Band 2
        {  31, 154, 185 }, {  17, 127, 175 }, {   6,  96, 145 },
        {   2,  73, 114 }, {   1,  51,  82 }, {   1,  28,  45 }
      }, {  // Band 3
        {  23, 163, 200 }, {  10, 131, 185 }, {   2,  93, 148 },
        {   1,  67, 111 }, {   1,  41,  69 }, {   1,  14,  24 }
      }, {  // Band 4
        {  29, 176, 217 }, {  12, 145, 201 }, {   3, 101, 156 },
        {   1,  69, 111 }, {   1,  39,  63 }, {   1,  14,  23 }
      }, {  // Band 5
        {  57, 192, 233 }, {  25, 154, 215 }, {   6, 109, 167 },
        {   3,  78, 118 }, {   1,  48,  69 }, {   1,  21,  29 }
      }
    }, {  // Inter
      {  // Band 0
        { 202, 105, 245 }, { 108, 106, 216 }, {  18,  90, 144 }
      }, {  // Band 1
        {  33, 172, 219 }, {  64, 149, 206 }, {  14, 117, 177 },
        {   5,  90, 141 }, {   2,  61,  95 }, {   1,  37,  57 }
      }, {  // Band 2
        {  33, 179, 220 }, {  11, 140, 198 }, {   1,  89, 148 },
        {   1,  60, 104 }, {   1,  33,  57 }, {   1,  12,  21 }
      }, {  // Band 3
        {  30, 181, 221 }, {   8, 141, 198 }, {   1,  87, 145 },
        {   1,  58, 100 }, {   1,  31,  55 }, {   1,  12,  20 }
      }, {  // Band 4
        {  32, 186, 224 }, {   7, 142, 198 }, {   1,  86, 143 },
        {   1,  58, 100 }, {   1,  31,  55 }, {   1,  12,  22 }
      }, {  // Band 5
        {  57, 192, 227 }, {  20, 143, 204 }, {   3,  96, 154 },
        {   1,  68, 112 }, {   1,  42,  69 }, {   1,  19,  32 }
      }
    }
  }, {  // UV plane
    {  // Intra
      {  // Band 0
        { 212,  35, 215 }, { 113,  47, 169 }, {  29,  48, 105 }
      }, {  // Band 1
        {  74, 129, 203 }, { 106, 120, 203 }, {  49, 107, 178 },
        {  19,  84, 144 }, {   4,  50,  84 }, {   1,  15,  25 }
      }, {  // Band 2
        {  71, 172, 217 }, {  44, 141, 209 }, {  15, 102, 173 },
        {   6,  76, 133 }, {   2,  51,  89 }, {   1,  24,  42 }
      }, {  // Band 3
        {  64, 185, 231 }, {  31, 148, 216 }, {   8, 103, 175 },
        {   3,  74, 131 }, {   1,  46,  81 }, {   1,  18,  30 }
      }, {  // Band 4
        {  65, 196, 235 }, {  25, 157, 221 }, {   5, 105, 174 },
        {   1,  67, 120 }, {   1,  38,  69 }, {   1,  15,  30 }
      }, {  // Band 5
        {  65, 204, 238 }, {  30, 156, 224 }, {   7, 107, 177 },
        {   2,  70, 124 }, {   1,  42,  73 }, {   1,  18,  34 }
      }
    }, {  // Inter
      {  // Band 0
        { 225,  86, 251 }, { 144, 104, 235 }, {  42,  99, 181 }
      }, {  // Band 1
        {  85, 175, 239 }, { 112, 165, 229 }, {  29, 136, 200 },
        {  12, 103, 162 }, {   6,  77, 123 }, {   2,  53,  84 }
      }, {  // Band 2
        {  75, 183, 239 }, {  30, 155, 221 }, {   3, 106, 171 },
        {   1,  74, 128 }, {   1,  44,  76 }, {   1,  17,  28 }
      }, {  // Band 3
        {  73, 185, 240 }, {  27, 159, 222 }, {   2, 107, 172 },
        {   1,  75, 127 }, {   1,  42,  73 }, {   1,  17,  29 }
      }, {  // Band 4
        {  62, 190, 238 }, {  21, 159, 222 }, {   2, 107, 172 },
        {   1,  72, 122 }, {   1,  40,  71 }, {   1,  18,  32 }
      }, {  // Band 5
        {  61, 199, 240 }, {  27, 161, 226 }, {   4, 113, 180 },
        {   1,  76, 129 }, {   1,  46,  80 }, {   1,  23,  41 }
      }
    }
  }
};

static const av1_coeff_probs_model default_coef_probs_16x16[PLANE_TYPES] = {
  {  // Y plane
    {  // Intra
      {  // Band 0
        {   7,  27, 153 }, {   5,  30,  95 }, {   1,  16,  30 }
      }, {  // Band 1
        {  50,  75, 127 }, {  57,  75, 124 }, {  27,  67, 108 },
        {  10,  54,  86 }, {   1,  33,  52 }, {   1,  12,  18 }
      }, {  // Band 2
        {  43, 125, 151 }, {  26, 108, 148 }, {   7,  83, 122 },
        {   2,  59,  89 }, {   1,  38,  60 }, {   1,  17,  27 }
      }, {  // Band 3
        {  23, 144, 163 }, {  13, 112, 154 }, {   2,  75, 117 },
        {   1,  50,  81 }, {   1,  31,  51 }, {   1,  14,  23 }
      }, {  // Band 4
        {  18, 162, 185 }, {   6, 123, 171 }, {   1,  78, 125 },
        {   1,  51,  86 }, {   1,  31,  54 }, {   1,  14,  23 }
      }, {  // Band 5
        {  15, 199, 227 }, {   3, 150, 204 }, {   1,  91, 146 },
        {   1,  55,  95 }, {   1,  30,  53 }, {   1,  11,  20 }
      }
    }, {  // Inter
      {  // Band 0
        {  19,  55, 240 }, {  19,  59, 196 }, {   3,  52, 105 }
      }, {  // Band 1
        {  41, 166, 207 }, { 104, 153, 199 }, {  31, 123, 181 },
        {  14, 101, 152 }, {   5,  72, 106 }, {   1,  36,  52 }
      }, {  // Band 2
        {  35, 176, 211 }, {  12, 131, 190 }, {   2,  88, 144 },
        {   1,  60, 101 }, {   1,  36,  60 }, {   1,  16,  28 }
      }, {  // Band 3
        {  28, 183, 213 }, {   8, 134, 191 }, {   1,  86, 142 },
        {   1,  56,  96 }, {   1,  30,  53 }, {   1,  12,  20 }
      }, {  // Band 4
        {  20, 190, 215 }, {   4, 135, 192 }, {   1,  84, 139 },
        {   1,  53,  91 }, {   1,  28,  49 }, {   1,  11,  20 }
      }, {  // Band 5
        {  13, 196, 216 }, {   2, 137, 192 }, {   1,  86, 143 },
        {   1,  57,  99 }, {   1,  32,  56 }, {   1,  13,  24 }
      }
    }
  }, {  // UV plane
    {  // Intra
      {  // Band 0
        { 211,  29, 217 }, {  96,  47, 156 }, {  22,  43,  87 }
      }, {  // Band 1
        {  78, 120, 193 }, { 111, 116, 186 }, {  46, 102, 164 },
        {  15,  80, 128 }, {   2,  49,  76 }, {   1,  18,  28 }
      }, {  // Band 2
        {  71, 161, 203 }, {  42, 132, 192 }, {  10,  98, 150 },
        {   3,  69, 109 }, {   1,  44,  70 }, {   1,  18,  29 }
      }, {  // Band 3
        {  57, 186, 211 }, {  30, 140, 196 }, {   4,  93, 146 },
        {   1,  62, 102 }, {   1,  38,  65 }, {   1,  16,  27 }
      }, {  // Band 4
        {  47, 199, 217 }, {  14, 145, 196 }, {   1,  88, 142 },
        {   1,  57,  98 }, {   1,  36,  62 }, {   1,  15,  26 }
      }, {  // Band 5
        {  26, 219, 229 }, {   5, 155, 207 }, {   1,  94, 151 },
        {   1,  60, 104 }, {   1,  36,  62 }, {   1,  16,  28 }
      }
    }, {  // Inter
      {  // Band 0
        { 233,  29, 248 }, { 146,  47, 220 }, {  43,  52, 140 }
      }, {  // Band 1
        { 100, 163, 232 }, { 179, 161, 222 }, {  63, 142, 204 },
        {  37, 113, 174 }, {  26,  89, 137 }, {  18,  68,  97 }
      }, {  // Band 2
        {  85, 181, 230 }, {  32, 146, 209 }, {   7, 100, 164 },
        {   3,  71, 121 }, {   1,  45,  77 }, {   1,  18,  30 }
      }, {  // Band 3
        {  65, 187, 230 }, {  20, 148, 207 }, {   2,  97, 159 },
        {   1,  68, 116 }, {   1,  40,  70 }, {   1,  14,  29 }
      }, {  // Band 4
        {  40, 194, 227 }, {   8, 147, 204 }, {   1,  94, 155 },
        {   1,  65, 112 }, {   1,  39,  66 }, {   1,  14,  26 }
      }, {  // Band 5
        {  16, 208, 228 }, {   3, 151, 207 }, {   1,  98, 160 },
        {   1,  67, 117 }, {   1,  41,  74 }, {   1,  17,  31 }
      }
    }
  }
};

static const av1_coeff_probs_model default_coef_probs_32x32[PLANE_TYPES] = {
  {  // Y plane
    {  // Intra
      {  // Band 0
        {  17,  38, 140 }, {   7,  34,  80 }, {   1,  17,  29 }
      }, {  // Band 1
        {  37,  75, 128 }, {  41,  76, 128 }, {  26,  66, 116 },
        {  12,  52,  94 }, {   2,  32,  55 }, {   1,  10,  16 }
      }, {  // Band 2
        {  50, 127, 154 }, {  37, 109, 152 }, {  16,  82, 121 },
        {   5,  59,  85 }, {   1,  35,  54 }, {   1,  13,  20 }
      }, {  // Band 3
        {  40, 142, 167 }, {  17, 110, 157 }, {   2,  71, 112 },
        {   1,  44,  72 }, {   1,  27,  45 }, {   1,  11,  17 }
      }, {  // Band 4
        {  30, 175, 188 }, {   9, 124, 169 }, {   1,  74, 116 },
        {   1,  48,  78 }, {   1,  30,  49 }, {   1,  11,  18 }
      }, {  // Band 5
        {  10, 222, 223 }, {   2, 150, 194 }, {   1,  83, 128 },
        {   1,  48,  79 }, {   1,  27,  45 }, {   1,  11,  17 }
      }
    }, {  // Inter
      {  // Band 0
        {  36,  41, 235 }, {  29,  36, 193 }, {  10,  27, 111 }
      }, {  // Band 1
        {  85, 165, 222 }, { 177, 162, 215 }, { 110, 135, 195 },
        {  57, 113, 168 }, {  23,  83, 120 }, {  10,  49,  61 }
      }, {  // Band 2
        {  85, 190, 223 }, {  36, 139, 200 }, {   5,  90, 146 },
        {   1,  60, 103 }, {   1,  38,  65 }, {   1,  18,  30 }
      }, {  // Band 3
        {  72, 202, 223 }, {  23, 141, 199 }, {   2,  86, 140 },
        {   1,  56,  97 }, {   1,  36,  61 }, {   1,  16,  27 }
      }, {  // Band 4
        {  55, 218, 225 }, {  13, 145, 200 }, {   1,  86, 141 },
        {   1,  57,  99 }, {   1,  35,  61 }, {   1,  13,  22 }
      }, {  // Band 5
        {  15, 235, 212 }, {   1, 132, 184 }, {   1,  84, 139 },
        {   1,  57,  97 }, {   1,  34,  56 }, {   1,  14,  23 }
      }
    }
  }, {  // UV plane
    {  // Intra
      {  // Band 0
        { 181,  21, 201 }, {  61,  37, 123 }, {  10,  38,  71 }
      }, {  // Band 1
        {  47, 106, 172 }, {  95, 104, 173 }, {  42,  93, 159 },
        {  18,  77, 131 }, {   4,  50,  81 }, {   1,  17,  23 }
      }, {  // Band 2
        {  62, 147, 199 }, {  44, 130, 189 }, {  28, 102, 154 },
        {  18,  75, 115 }, {   2,  44,  65 }, {   1,  12,  19 }
      }, {  // Band 3
        {  55, 153, 210 }, {  24, 130, 194 }, {   3,  93, 146 },
        {   1,  61,  97 }, {   1,  31,  50 }, {   1,  10,  16 }
      }, {  // Band 4
        {  49, 186, 223 }, {  17, 148, 204 }, {   1,  96, 142 },
        {   1,  53,  83 }, {   1,  26,  44 }, {   1,  11,  17 }
      }, {  // Band 5
        {  13, 217, 212 }, {   2, 136, 180 }, {   1,  78, 124 },
        {   1,  50,  83 }, {   1,  29,  49 }, {   1,  14,  23 }
      }
    }, {  // Inter
      {  // Band 0
        { 197,  13, 247 }, {  82,  17, 222 }, {  25,  17, 162 }
      }, {  // Band 1
        { 126, 186, 247 }, { 234, 191, 243 }, { 176, 177, 234 },
        { 104, 158, 220 }, {  66, 128, 186 }, {  55,  90, 137 }
      }, {  // Band 2
        { 111, 197, 242 }, {  46, 158, 219 }, {   9, 104, 171 },
        {   2,  65, 125 }, {   1,  44,  80 }, {   1,  17,  91 }
      }, {  // Band 3
        { 104, 208, 245 }, {  39, 168, 224 }, {   3, 109, 162 },
        {   1,  79, 124 }, {   1,  50, 102 }, {   1,  43, 102 }
      }, {  // Band 4
        {  84, 220, 246 }, {  31, 177, 231 }, {   2, 115, 180 },
        {   1,  79, 134 }, {   1,  55,  77 }, {   1,  60,  79 }
      }, {  // Band 5
        {  43, 243, 240 }, {   8, 180, 217 }, {   1, 115, 166 },
        {   1,  84, 121 }, {   1,  51,  67 }, {   1,  16,   6 }
      }
    }
  }
};
#endif  // CONFIG_NEW_TOKENSET

#if CONFIG_TX64X64
// FIXME. Optimize for EC_MULTISYMBOL
static const av1_coeff_probs_model default_coef_probs_64x64[PLANE_TYPES] = {
  {  // Y plane
    {  // Intra
      {  // Band 0
        {  17,  38, 140 }, {   7,  34,  80 }, {   1,  17,  29 }
      }, {  // Band 1
        {  37,  75, 128 }, {  41,  76, 128 }, {  26,  66, 116 },
        {  12,  52,  94 }, {   2,  32,  55 }, {   1,  10,  16 }
      }, {  // Band 2
        {  50, 127, 154 }, {  37, 109, 152 }, {  16,  82, 121 },
        {   5,  59,  85 }, {   1,  35,  54 }, {   1,  13,  20 }
      }, {  // Band 3
        {  40, 142, 167 }, {  17, 110, 157 }, {   2,  71, 112 },
        {   1,  44,  72 }, {   1,  27,  45 }, {   1,  11,  17 }
      }, {  // Band 4
        {  30, 175, 188 }, {   9, 124, 169 }, {   1,  74, 116 },
        {   1,  48,  78 }, {   1,  30,  49 }, {   1,  11,  18 }
      }, {  // Band 5
        {  10, 222, 223 }, {   2, 150, 194 }, {   1,  83, 128 },
        {   1,  48,  79 }, {   1,  27,  45 }, {   1,  11,  17 }
      }
    }, {  // Inter
      {  // Band 0
        {  36,  41, 235 }, {  29,  36, 193 }, {  10,  27, 111 }
      }, {  // Band 1
        {  85, 165, 222 }, { 177, 162, 215 }, { 110, 135, 195 },
        {  57, 113, 168 }, {  23,  83, 120 }, {  10,  49,  61 }
      }, {  // Band 2
        {  85, 190, 223 }, {  36, 139, 200 }, {   5,  90, 146 },
        {   1,  60, 103 }, {   1,  38,  65 }, {   1,  18,  30 }
      }, {  // Band 3
        {  72, 202, 223 }, {  23, 141, 199 }, {   2,  86, 140 },
        {   1,  56,  97 }, {   1,  36,  61 }, {   1,  16,  27 }
      }, {  // Band 4
        {  55, 218, 225 }, {  13, 145, 200 }, {   1,  86, 141 },
        {   1,  57,  99 }, {   1,  35,  61 }, {   1,  13,  22 }
      }, {  // Band 5
        {  15, 235, 212 }, {   1, 132, 184 }, {   1,  84, 139 },
        {   1,  57,  97 }, {   1,  34,  56 }, {   1,  14,  23 }
      }
    }
  }, {  // UV plane
    {  // Intra
      {  // Band 0
        { 181,  21, 201 }, {  61,  37, 123 }, {  10,  38,  71 }
      }, {  // Band 1
        {  47, 106, 172 }, {  95, 104, 173 }, {  42,  93, 159 },
        {  18,  77, 131 }, {   4,  50,  81 }, {   1,  17,  23 }
      }, {  // Band 2
        {  62, 147, 199 }, {  44, 130, 189 }, {  28, 102, 154 },
        {  18,  75, 115 }, {   2,  44,  65 }, {   1,  12,  19 }
      }, {  // Band 3
        {  55, 153, 210 }, {  24, 130, 194 }, {   3,  93, 146 },
        {   1,  61,  97 }, {   1,  31,  50 }, {   1,  10,  16 }
      }, {  // Band 4
        {  49, 186, 223 }, {  17, 148, 204 }, {   1,  96, 142 },
        {   1,  53,  83 }, {   1,  26,  44 }, {   1,  11,  17 }
      }, {  // Band 5
        {  13, 217, 212 }, {   2, 136, 180 }, {   1,  78, 124 },
        {   1,  50,  83 }, {   1,  29,  49 }, {   1,  14,  23 }
      }
    }, {  // Inter
      {  // Band 0
        { 197,  13, 247 }, {  82,  17, 222 }, {  25,  17, 162 }
      }, {  // Band 1
        { 126, 186, 247 }, { 234, 191, 243 }, { 176, 177, 234 },
        { 104, 158, 220 }, {  66, 128, 186 }, {  55,  90, 137 }
      }, {  // Band 2
        { 111, 197, 242 }, {  46, 158, 219 }, {   9, 104, 171 },
        {   2,  65, 125 }, {   1,  44,  80 }, {   1,  17,  91 }
      }, {  // Band 3
        { 104, 208, 245 }, {  39, 168, 224 }, {   3, 109, 162 },
        {   1,  79, 124 }, {   1,  50, 102 }, {   1,  43, 102 }
      }, {  // Band 4
        {  84, 220, 246 }, {  31, 177, 231 }, {   2, 115, 180 },
        {   1,  79, 134 }, {   1,  55,  77 }, {   1,  60,  79 }
      }, {  // Band 5
        {  43, 243, 240 }, {   8, 180, 217 }, {   1, 115, 166 },
        {   1,  84, 121 }, {   1,  51,  67 }, {   1,  16,   6 }
      }
    }
  }
};
#endif  // CONFIG_TX64X64
#endif  // CONFIG_Q_ADAPT_PROBS
#if CONFIG_NEW_TOKENSET
static const aom_prob av1_default_blockzero_probs[TX_SIZES][PLANE_TYPES]
                                           [REF_TYPES][BLOCKZ_CONTEXTS] = {
  { // TX_4x4
    { // Y plane
      { 195, 84, 8, },  // Intra
      { 191, 124, 25, },  // Inter
    },
    { // UV plane
      { 214, 132, 42, },  // Intra
      { 229, 143, 46, },  // Inter
    },
  },
  { // TX_8x8
    { // Y plane
      { 125, 52, 6, },  // Intra
      { 202, 108, 18, },  // Inter
    },
    { // UV plane
      { 212, 113, 29, },  // Intra
      { 225, 144, 42, },  // Inter
    },
  },
  { // TX_16x16
    { // Y plane
      { 7, 5, 1, },  // Intra
      { 19, 19, 3, },  // Inter
    },
    { // UV plane
      { 211, 96, 22, },  // Intra
      { 233, 146, 43, },  // Inter
    },
  },
  { // TX_32x32
    { //  Y plane
      { 17, 7, 1, },  // Intra
      { 36, 29, 10, },  // Inter
    },
    { // UV plane
      { 181, 61, 10, },  // Intra
      { 197, 82, 25, },  // Inter
    },
  },
#if CONFIG_TX64X64
  { // TX_64x64 FIXME: currently the same as 32x32
    { //  Y plane
      { 17, 7, 1, },  // Intra
      { 36, 29, 10, },  // Inter
    },
    { // UV plane
      { 181, 61, 10, },  // Intra
      { 197, 82, 25, },  // Inter
    },
  },
#endif
};

static const coeff_cdf_model default_coef_head_cdf_4x4[PLANE_TYPES] = {
  {     // Y plane
    {   // Intra
      { // Band 0
        { 25024, 25863, 27361, 29796, 30374, 32768 },
        { 10816, 14127, 17116, 23516, 24999, 32768 },
        { 1088, 6358, 8428, 16648, 18276, 32768 } },
      { // Band 1
        {14529, 18769, 29100, 29634, 32768 },
        {12993, 17117, 28404, 28988, 32768 },
        {11201, 14084, 25818, 26504, 32768 },
        {9793, 11267, 21775, 22451, 32768 },
        {7105, 7562, 15777, 16225, 32768 },
        {3905, 3966, 8359, 8526, 32768 } },
      { // Band 2
        {20033, 23643, 31102, 31374, 32768 },
        {16321, 20350, 30167, 30546, 32768 },
        {12993, 15512, 26859, 27396, 32768 },
        {10305, 11659, 21669, 22330, 32768 },
        {7361, 7819, 15450, 15940, 32768 },
        {3521, 3580, 7805, 7976, 32768 } },
      { // Band 3
        {21057, 25460, 31740, 31952, 32768 },
        {16449, 21173, 30761, 31092, 32768 },
        {11841, 14615, 26188, 26824, 32768 },
        {7745, 8991, 18937, 19707, 32768 },
        {4417, 4706, 10342, 10890, 32768 },
        {7617, 8392, 17295, 17915, 32768 } },
      { // Band 4
        {20417, 26452, 32166, 32321, 32768 },
        {15809, 21634, 30947, 31298, 32768 },
        {10049, 12176, 23495, 24229, 32768 },
        {5953, 6731, 16166, 16798, 32768 },
        {6081, 6188, 8114, 8764, 32768 },
        {2113, 2291, 4448, 5527, 32768 } },
      { // Band 5
        {9153, 25905, 31431, 31934, 32768 },
        {9025, 23345, 30033, 30965, 32768 },
        {5953, 13835, 22032, 24664, 32768 },
        {6337, 11435, 18366, 21418, 32768 },
        {3137, 4871, 8519, 12426, 32768 },
        {1857, 2727, 5540, 8757, 32768 } } },
    {   // Intra
      { // Band 0
        { 24512, 26673, 28962, 31929, 32126, 32768 },
        { 15936, 21711, 25569, 30899, 31305, 32768 },
        { 3264, 14756, 20107, 29407, 30032, 32768 } },
      { // Band 1
        {21313, 26020, 32523, 32575, 32768 },
        {18369, 24215, 32291, 32391, 32768 },
        {15297, 19637, 30414, 30752, 32768 },
        {11713, 14040, 25408, 26033, 32768 },
        {9537, 10173, 18839, 19315, 32768 },
        {9025, 9093, 13987, 14115, 32768 } },
      { // Band 2
        {22721, 27599, 32592, 32636, 32768 },
        {19009, 24676, 32258, 32367, 32768 },
        {12737, 16769, 28739, 29247, 32768 },
        {8769, 10956, 21941, 22840, 32768 },
        {6721, 7678, 15319, 16290, 32768 },
        {4417, 4430, 4583, 5712, 32768 } },
      { // Band 3
        {22849, 28333, 32633, 32671, 32768 },
        {18497, 24619, 32184, 32315, 32768 },
        {11841, 14640, 27251, 27752, 32768 },
        {8385, 10154, 18339, 19621, 32768 },
        {5697, 6977, 13787, 15289, 32768 },
        {193, 194, 384, 479, 32768 } },
      { // Band 4
        {20417, 28167, 32552, 32621, 32768 },
        {16833, 23968, 31991, 32174, 32768 },
        {10433, 13387, 26356, 26951, 32768 },
        {5057, 6823, 18967, 19843, 32768 },
        {5697, 6479, 11672, 13052, 32768 },
        {2241, 2265, 6355, 6432, 32768 } },
      { // Band 5
        {12097, 28717, 32406, 32555, 32768 },
        {10433, 26113, 31504, 31975, 32768 },
        {5825, 14284, 21349, 24461, 32768 },
        {4545, 8454, 12648, 17501, 32768 },
        {193, 7173, 15272, 19322, 32768 },
        {2113, 2183, 7202, 7377, 32768 } } } },
  {     // UV plane
    {   // Inter
      { // Band 0
        { 27456, 28244, 31289, 32358, 32534, 32768 },
        { 16960, 21207, 26511, 30539, 31190, 32768 },
        { 5440, 13412, 18469, 26423, 27669, 32768 } },
      { // Band 1
        {17857, 26327, 31983, 32219, 32768 },
        {16065, 24198, 31431, 31785, 32768 },
        {12865, 18011, 28454, 29166, 32768 },
        {9665, 12501, 24331, 25147, 32768 },
        {2753, 3121, 12661, 13034, 32768 },
        {4033, 4140, 11834, 11977, 32768 } },
      { // Band 2
        {21185, 28338, 32249, 32417, 32768 },
        {18497, 25227, 31905, 32122, 32768 },
        {12097, 16516, 28610, 29166, 32768 },
        {9281, 11157, 21438, 22312, 32768 },
        {5697, 6566, 15585, 16340, 32768 },
        {9409, 9659, 11827, 12911, 32768 } },
      { // Band 3
        {22337, 29459, 32382, 32519, 32768 },
        {16961, 25262, 31874, 32123, 32768 },
        {12353, 17748, 29300, 29852, 32768 },
        {9025, 11528, 24468, 25141, 32768 },
        {6209, 6565, 15806, 16121, 32768 },
        {2497, 2524, 7050, 7125, 32768 } },
      { // Band 4
        {20417, 29779, 32552, 32636, 32768 },
        {15553, 26420, 32063, 32295, 32768 },
        {9665, 17946, 29385, 30096, 32768 },
        {5569, 10207, 22410, 23836, 32768 },
        {16449, 16450, 16545, 16593, 32768 },
        {2369, 2395, 6822, 6898, 32768 } },
      { // Band 5
        {10177, 30567, 32725, 32745, 32768 },
        {9537, 28243, 32179, 32423, 32768 },
        {13377, 23187, 29322, 30382, 32768 },
        {13121, 21346, 29507, 30326, 32768 },
        {4417, 4939, 15104, 15535, 32768 },
        {2625, 2680, 8218, 8338, 32768 } } },
    {   // Inter
      { // Band 0
        { 29376, 30098, 32421, 32766, 32767, 32768 },
        { 18368, 22916, 30116, 32541, 32650, 32768 },
        { 5952, 16505, 25955, 32163, 32365, 32768 } },
      { // Band 1
        {19649, 30160, 32743, 32753, 32768 },
        {18881, 28724, 32688, 32717, 32768 },
        {16833, 23053, 31244, 31573, 32768 },
        {14657, 17714, 26083, 26978, 32768 },
        {14657, 16618, 24597, 25403, 32768 },
        {4289, 4326, 10686, 10751, 32768 } },
      { // Band 2
        {21953, 30956, 32748, 32757, 32768 },
        {20929, 29412, 32700, 32725, 32768 },
        {13377, 21495, 31216, 31569, 32768 },
        {9153, 15097, 28295, 28990, 32768 },
        {5313, 5363, 13839, 13894, 32768 },
        {2625, 2652, 7276, 7351, 32768 } },
      { // Band 3
        {20289, 31164, 32745, 32755, 32768 },
        {17601, 29635, 32739, 32751, 32768 },
        {18241, 24284, 32116, 32258, 32768 },
        {32705, 32706, 32739, 32740, 32768 },
        {5697, 5750, 14739, 14792, 32768 },
        {2881, 2913, 8427, 8498, 32768 } },
      { // Band 4
        {19009, 31481, 32742, 32754, 32768 },
        {15809, 30521, 32736, 32750, 32768 },
        {16449, 32705, 32737, 32753, 32768 },
        {7873, 8039, 19981, 20068, 32768 },
        {5313, 5366, 14376, 14430, 32768 },
        {2753, 2789, 8909, 8979, 32768 } },
      { // Band 5
        {11841, 32116, 32728, 32748, 32768 },
        {12353, 32132, 32729, 32748, 32768 },
        {7489, 12435, 25708, 26666, 32768 },
        {5697, 7486, 20238, 21009, 32768 },
        {4929, 5579, 16402, 16866, 32768 },
        {3009, 3246, 10158, 10533, 32768 } } } }
};
static const coeff_cdf_model default_coef_head_cdf_8x8[PLANE_TYPES] = {
  {     // Y plane
    {   // Intra
      { // Band 0
        { 16064, 18127, 22153, 27289, 28507, 32768 },
        { 6720, 10545, 13491, 20948, 22631, 32768 },
        { 832, 5270, 5918, 12645, 13532, 32768 } },
      { // Band 1
        {14017, 16139, 26799, 27295, 32768 },
        {12737, 15136, 26235, 26816, 32768 },
        {10817, 12445, 23637, 24217, 32768 },
        {8897, 9702, 20040, 20500, 32768 },
        {5953, 6156, 13966, 14205, 32768 },
        {2497, 2519, 6222, 6300, 32768 } },
      { // Band 2
        {19777, 21403, 30054, 30269, 32768 },
        {16193, 17913, 28593, 28883, 32768 },
        {12609, 13572, 25248, 25534, 32768 },
        {9665, 10118, 20721, 20968, 32768 },
        {6849, 7028, 15202, 15391, 32768 },
        {3009, 3036, 7601, 7675, 32768 } },
      { // Band 3
        {22593, 23915, 31159, 31283, 32768 },
        {17345, 18690, 29425, 29611, 32768 },
        {11969, 12540, 24685, 24867, 32768 },
        {8129, 8355, 18668, 18819, 32768 },
        {4673, 4714, 11752, 11814, 32768 },
        {1857, 1876, 5057, 5138, 32768 } },
      { // Band 4
        {24513, 25718, 31947, 32014, 32768 },
        {18881, 20029, 30409, 30527, 32768 },
        {12481, 12953, 25201, 25341, 32768 },
        {8385, 8528, 18815, 18910, 32768 },
        {4289, 4327, 10797, 10861, 32768 },
        {1857, 1872, 4332, 4415, 32768 } },
      { // Band 5
        {26049, 27752, 32415, 32462, 32768 },
        {20417, 22100, 31056, 31192, 32768 },
        {12481, 13075, 24646, 24844, 32768 },
        {7489, 7696, 17117, 17285, 32768 },
        {3777, 3814, 10062, 10129, 32768 },
        {1473, 1486, 3735, 3820, 32768 } } },
    {   // Intra
      { // Band 0
        { 25920, 27743, 29455, 32147, 32280, 32768 },
        { 13888, 19845, 23350, 30219, 30660, 32768 },
        { 2368, 12781, 16196, 27232, 27894, 32768 } },
      { // Band 1
        {21697, 24758, 32358, 32417, 32768 },
        {20289, 23960, 32111, 32213, 32768 },
        {17345, 19966, 30630, 30841, 32768 },
        {14529, 16070, 27461, 27777, 32768 },
        {9793, 10613, 21146, 21566, 32768 },
        {6977, 7162, 15591, 15776, 32768 } },
      { // Band 2
        {23617, 26783, 32572, 32607, 32768 },
        {20801, 24292, 32185, 32275, 32768 },
        {15169, 17905, 29916, 30181, 32768 },
        {10945, 12972, 25565, 26064, 32768 },
        {6849, 8334, 18543, 19446, 32768 },
        {3649, 4346, 12351, 13169, 32768 } },
      { // Band 3
        {25281, 28440, 32667, 32689, 32768 },
        {22081, 25694, 32414, 32476, 32768 },
        {15297, 18341, 30141, 30410, 32768 },
        {10305, 12381, 24477, 25084, 32768 },
        {5697, 6673, 16325, 17080, 32768 },
        {2369, 2393, 6466, 6543, 32768 } },
      { // Band 4
        {25921, 29445, 32729, 32739, 32768 },
        {22465, 26834, 32588, 32627, 32768 },
        {16449, 20062, 31016, 31233, 32768 },
        {11073, 13165, 25353, 25896, 32768 },
        {11713, 13837, 20144, 21734, 32768 },
        {2241, 2265, 6355, 6432, 32768 } },
      { // Band 5
        {26177, 29403, 32705, 32721, 32768 },
        {22337, 26344, 32545, 32589, 32768 },
        {19009, 21527, 31775, 31873, 32768 },
        {11585, 12685, 22632, 23137, 32768 },
        {8257, 8305, 16444, 16492, 32768 },
        {2113, 2183, 7202, 7377, 32768 } } } },
  {     // UV plane
    {   // Inter
      { // Band 0
        { 27200, 27981, 31389, 32444, 32592, 32768 },
        { 14528, 19068, 24887, 29901, 30688, 32768 },
        { 3776, 11778, 14700, 23745, 24854, 32768 } },
      { // Band 1
        {20289, 25202, 31672, 31909, 32768 },
        {18369, 23493, 31166, 31487, 32768 },
        {15425, 18619, 28941, 29393, 32768 },
        {10945, 12535, 24287, 24792, 32768 },
        {6465, 6810, 15764, 16080, 32768 },
        {2113, 2137, 6125, 6203, 32768 } },
      { // Band 2
        {23745, 27041, 31976, 32135, 32768 },
        {19521, 22766, 31139, 31367, 32768 },
        {14273, 15834, 27820, 28105, 32768 },
        {9537, 10445, 22106, 22491, 32768 },
        {7233, 7386, 15961, 16109, 32768 },
        {2369, 2401, 7891, 7964, 32768 } },
      { // Band 3
        {26305, 28703, 32352, 32435, 32768 },
        {20673, 23490, 31517, 31680, 32768 },
        {14017, 15251, 27458, 27702, 32768 },
        {10945, 11374, 22496, 22687, 32768 },
        {9153, 9435, 22299, 22411, 32768 },
        {193, 269, 13236, 13293, 32768 } },
      { // Band 4
        {27713, 29770, 32522, 32575, 32768 },
        {21569, 24342, 31785, 31919, 32768 },
        {15297, 16497, 28367, 28569, 32768 },
        {17601, 17828, 24444, 24582, 32768 },
        {6977, 7035, 16901, 16947, 32768 },
        {193, 384, 32706, 32707, 32768 } },
      { // Band 5
        {28737, 30879, 32667, 32695, 32768 },
        {22593, 26241, 32073, 32207, 32768 },
        {16577, 19148, 28436, 28906, 32768 },
        {12993, 14005, 23151, 23630, 32768 },
        {7617, 9188, 22797, 23313, 32768 },
        {2625, 2680, 8218, 8338, 32768 } } },
    {   // Inter
      { // Band 0
        { 28864, 29988, 32423, 32766, 32767, 32768 },
        { 18496, 24572, 30167, 32687, 32737, 32768 },
        { 5440, 19618, 25332, 32393, 32491, 32768 } },
      { // Band 1
        {23745, 29427, 32751, 32757, 32768 },
        {23745, 28704, 32716, 32731, 32768 },
        {23105, 27943, 32524, 32587, 32768 },
        {21057, 24773, 29589, 30282, 32768 },
        {12609, 14823, 23831, 24713, 32768 },
        {16449, 16450, 16545, 16593, 32768 } },
      { // Band 2
        {25025, 30203, 32754, 32759, 32768 },
        {23617, 28361, 32715, 32729, 32768 },
        {17985, 21562, 31354, 31543, 32768 },
        {12353, 18915, 28742, 29548, 32768 },
        {193, 289, 16545, 16593, 32768 },
        {2625, 2652, 7276, 7351, 32768 } },
      { // Band 3
        {26433, 30892, 32757, 32761, 32768 },
        {24513, 29274, 32721, 32735, 32768 },
        {20161, 24040, 32055, 32171, 32768 },
        {21953, 24678, 27382, 28734, 32768 },
        {5697, 5750, 14739, 14792, 32768 },
        {2881, 2913, 8427, 8498, 32768 } },
      { // Band 4
        {27457, 31485, 32759, 32763, 32768 },
        {24129, 29502, 32752, 32757, 32768 },
        {19009, 25452, 32473, 32544, 32768 },
        {32705, 32706, 32737, 32738, 32768 },
        {5313, 5366, 14376, 14430, 32768 },
        {2753, 2789, 8909, 8979, 32768 } },
      { // Band 5
        {27841, 32288, 32759, 32764, 32768 },
        {19137, 30271, 32742, 32753, 32768 },
        {18625, 27739, 29979, 31099, 32768 },
        {5697, 7486, 20238, 21009, 32768 },
        {4929, 5579, 16402, 16866, 32768 },
        {3009, 3246, 10158, 10533, 32768 } } } }
};
static const coeff_cdf_model default_coef_head_cdf_16x16[PLANE_TYPES] = {
  {     // Y plane
    {   // Intra
      { // Band 0
        { 960, 4882, 9467, 17710, 20412, 32768 },
        { 704, 4657, 6561, 14507, 16279, 32768 },
        { 192, 3443, 3759, 9011, 9685, 32768 } },
      { // Band 1
        {12481, 13958, 24487, 24997, 32768 },
        {11457, 13075, 23820, 24406, 32768 },
        {9793, 11127, 21775, 22387, 32768 },
        {7745, 8457, 18155, 18655, 32768 },
        {5441, 5668, 13180, 13467, 32768 },
        {2497, 2520, 6340, 6417, 32768 } },
      { // Band 2
        {19521, 20572, 28965, 29177, 32768 },
        {15425, 16741, 27247, 27554, 32768 },
        {11969, 12690, 23872, 24141, 32768 },
        {9281, 9678, 19970, 20207, 32768 },
        {6081, 6266, 14682, 14876, 32768 },
        {2753, 2779, 7150, 7225, 32768 } },
      { // Band 3
        {22337, 23293, 30630, 30753, 32768 },
        {16321, 17427, 28368, 28570, 32768 },
        {11457, 11907, 23570, 23741, 32768 },
        {7233, 7331, 17258, 17334, 32768 },
        {4033, 4070, 10375, 10441, 32768 },
        {1601, 1619, 4706, 4788, 32768 } },
      { // Band 4
        {24769, 25536, 31660, 31722, 32768 },
        {18113, 18886, 29420, 29534, 32768 },
        {11201, 11412, 23207, 23291, 32768 },
        {6977, 7033, 16599, 16646, 32768 },
        {4033, 4070, 10375, 10441, 32768 },
        {1601, 1620, 4827, 4909, 32768 } },
      { // Band 5
        {28353, 28831, 32502, 32517, 32768 },
        {21441, 21869, 30977, 31017, 32768 },
        {11969, 12088, 24116, 24158, 32768 },
        {7489, 7547, 17413, 17458, 32768 },
        {4545, 4585, 11325, 11388, 32768 },
        {2113, 2133, 5526, 5606, 32768 } } },
    {   // Intra
      { // Band 0
        { 2496, 8717, 17280, 28922, 29751, 32768 },
        { 2496, 9665, 15235, 26542, 27580, 32768 },
        { 448, 9240, 11886, 24124, 24898, 32768 } },
      { // Band 1
        {21057, 22896, 31877, 31953, 32768 },
        {20673, 23151, 31706, 31825, 32768 },
        {18753, 20519, 30497, 30668, 32768 },
        {15425, 16608, 27789, 28027, 32768 },
        {10305, 10977, 21405, 21749, 32768 },
        {3649, 3812, 11213, 11445, 32768 } },
      { // Band 2
        {24001, 25899, 32307, 32360, 32768 },
        {20929, 22941, 31775, 31867, 32768 },
        {15169, 16734, 29228, 29425, 32768 },
        {10561, 12047, 24918, 25324, 32768 },
        {6977, 7929, 18311, 18918, 32768 },
        {3649, 3760, 9962, 10162, 32768 } },
      { // Band 3
        {25793, 27526, 32565, 32591, 32768 },
        {21825, 23885, 32064, 32135, 32768 },
        {15041, 16286, 29203, 29360, 32768 },
        {10433, 11058, 24349, 24538, 32768 },
        {5569, 6016, 16460, 16794, 32768 },
        {193, 194, 384, 479, 32768 } },
      { // Band 4
        {26433, 28398, 32682, 32696, 32768 },
        {22977, 25086, 32367, 32412, 32768 },
        {16577, 17928, 30144, 30275, 32768 },
        {12481, 13352, 25993, 26211, 32768 },
        {7745, 8069, 20501, 20657, 32768 },
        {16449, 16450, 16545, 16593, 32768 } },
      { // Band 5
        {27841, 29700, 32721, 32730, 32768 },
        {23873, 26202, 32578, 32604, 32768 },
        {17729, 19046, 30448, 30568, 32768 },
        {13505, 14508, 26034, 26304, 32768 },
        {10049, 10494, 19945, 20233, 32768 },
        {2113, 2183, 7202, 7377, 32768 } } } },
  {     // UV plane
    {   // Inter
      { // Band 0
        { 27072, 27916, 31095, 32400, 32553, 32768 },
        { 12352, 16792, 22516, 28853, 29797, 32768 },
        { 2880, 9023, 11126, 20602, 21713, 32768 } },
      { // Band 1
        {20161, 24785, 31070, 31430, 32768 },
        {17985, 22773, 30430, 30880, 32768 },
        {15937, 18802, 28265, 28788, 32768 },
        {11841, 13587, 24798, 25335, 32768 },
        {8769, 9160, 19316, 19566, 32768 },
        {5313, 5357, 12874, 12932, 32768 } },
      { // Band 2
        {24129, 26501, 31672, 31844, 32768 },
        {19649, 21553, 30130, 30370, 32768 },
        {11713, 13134, 25983, 26321, 32768 },
        {9409, 9948, 21408, 21663, 32768 },
        {5569, 5757, 14335, 14533, 32768 },
        {2241, 2305, 13152, 13209, 32768 } },
      { // Band 3
        {26817, 28135, 32130, 32209, 32768 },
        {20161, 21412, 30331, 30481, 32768 },
        {13377, 13798, 26065, 26176, 32768 },
        {8129, 8290, 19920, 20008, 32768 },
        {5697, 5751, 14950, 15002, 32768 },
        {5569, 5601, 11041, 11105, 32768 } },
      { // Band 4
        {28225, 29079, 32387, 32426, 32768 },
        {21185, 22046, 30982, 31061, 32768 },
        {13377, 13595, 25762, 25824, 32768 },
        {8001, 8123, 20530, 20590, 32768 },
        {4289, 4322, 9907, 9974, 32768 },
        {3393, 3412, 6663, 6739, 32768 } },
      { // Band 5
        {30529, 31014, 32651, 32664, 32768 },
        {23489, 24268, 31627, 31682, 32768 },
        {14017, 14239, 26653, 26707, 32768 },
        {11201, 11317, 23122, 23169, 32768 },
        {6721, 6768, 14810, 14863, 32768 },
        {6593, 6632, 13188, 13245, 32768 } } },
    {   // Inter
      { // Band 0
        { 29888, 30492, 32500, 32766, 32767, 32768 },
        { 18752, 23235, 29846, 32214, 32442, 32768 },
        { 5568, 17762, 25039, 31213, 31651, 32768 } },
      { // Band 1
        {26433, 29681, 32757, 32760, 32768 },
        {24769, 28761, 32722, 32734, 32768 },
        {22209, 26975, 32418, 32500, 32768 },
        {16321, 21333, 28368, 29283, 32768 },
        {12865, 14775, 22545, 23553, 32768 },
        {12353, 12354, 12473, 12532, 32768 } },
      { // Band 2
        {27457, 30005, 32738, 32745, 32768 },
        {24897, 27541, 32723, 32731, 32768 },
        {15297, 19106, 30414, 30711, 32768 },
        {6593, 8826, 19732, 20840, 32768 },
        {4161, 4233, 16509, 16557, 32768 },
        {2625, 2652, 7276, 7351, 32768 } },
      { // Band 3
        {28609, 30482, 32761, 32763, 32768 },
        {25665, 27830, 32727, 32733, 32768 },
        {21057, 23803, 30367, 30721, 32768 },
        {10945, 21878, 32726, 32737, 32768 },
        {5697, 5750, 14739, 14792, 32768 },
        {2881, 2913, 8427, 8498, 32768 } },
      { // Band 4
        {28993, 30944, 32762, 32764, 32768 },
        {26561, 28695, 32733, 32739, 32768 },
        {17985, 19028, 31008, 31079, 32768 },
        {7873, 8039, 19981, 20068, 32768 },
        {5313, 5366, 14376, 14430, 32768 },
        {2753, 2789, 8909, 8979, 32768 } },
      { // Band 5
        {30273, 32029, 32764, 32766, 32768 },
        {28609, 30847, 32745, 32751, 32768 },
        {21313, 24377, 31986, 32098, 32768 },
        {32705, 32709, 32739, 32741, 32768 },
        {4929, 5579, 16402, 16866, 32768 },
        {3009, 3246, 10158, 10533, 32768 } } } }
};
static const coeff_cdf_model default_coef_head_cdf_32x32[PLANE_TYPES] = {
  {     // Y plane
    {   // Intra
      { // Band 0
        { 2240, 5407, 18304, 25601, 27911, 32768 },
        { 960, 4633, 8197, 16254, 18796, 32768 },
        { 192, 3061, 3557, 8701, 9762, 32768 } },
      { // Band 1
        {11969, 15846, 25660, 26667, 32768 },
        {11713, 15794, 25737, 26760, 32768 },
        {9281, 12675, 23181, 24351, 32768 },
        {7105, 8757, 18383, 19437, 32768 },
        {4289, 4579, 11353, 11792, 32768 },
        {1857, 1874, 4695, 4777, 32768 } },
      { // Band 2
        {20929, 22297, 29370, 29646, 32768 },
        {17473, 18985, 28079, 28413, 32768 },
        {13121, 14064, 24902, 25217, 32768 },
        {9793, 10214, 20069, 20329, 32768 },
        {5825, 5987, 13350, 13559, 32768 },
        {2241, 2260, 5520, 5600, 32768 } },
      { // Band 3
        {25921, 26891, 31632, 31729, 32768 },
        {18241, 19463, 29222, 29419, 32768 },
        {11585, 12065, 23294, 23488, 32768 },
        {6593, 6686, 16153, 16234, 32768 },
        {3137, 3170, 8751, 8821, 32768 },
        {1345, 1359, 3739, 3824, 32768 } },
      { // Band 4
        {27713, 28504, 32068, 32132, 32768 },
        {19265, 20354, 29789, 29943, 32768 },
        {11201, 11538, 22701, 22848, 32768 },
        {6337, 6424, 15268, 15353, 32768 },
        {3649, 3681, 9052, 9121, 32768 },
        {1601, 1618, 4584, 4667, 32768 } },
      { // Band 5
        {30913, 31044, 32635, 32640, 32768 },
        {22081, 22261, 30452, 30477, 32768 },
        {10561, 10625, 21535, 21568, 32768 },
        {6081, 6130, 14369, 14423, 32768 },
        {3777, 3809, 9156, 9225, 32768 },
        {1857, 1875, 4936, 5018, 32768 } } },
    {   // Intra
      { // Band 0
        { 4672, 6927, 23534, 29846, 30928, 32768 },
        { 3776, 6784, 18075, 25863, 27926, 32768 },
        { 1344, 5588, 12166, 20966, 23504, 32768 } },
      { // Band 1
        {19393, 22016, 31280, 31444, 32768 },
        {21185, 24329, 31706, 31865, 32768 },
        {20673, 23240, 31186, 31379, 32768 },
        {17857, 20035, 29594, 29889, 32768 },
        {13633, 14929, 24883, 25337, 32768 },
        {7873, 8416, 17452, 17886, 32768 } },
      { // Band 2
        {25665, 27145, 32256, 32314, 32768 },
        {21057, 22826, 31465, 31576, 32768 },
        {13633, 14885, 27873, 28088, 32768 },
        {8769, 9515, 21941, 22248, 32768 },
        {6209, 6594, 15598, 15950, 32768 },
        {1985, 2014, 6855, 6931, 32768 } },
      { // Band 3
        {26817, 27824, 32362, 32399, 32768 },
        {21185, 22321, 31389, 31466, 32768 },
        {13761, 14154, 27163, 27245, 32768 },
        {8897, 9011, 20600, 20659, 32768 },
        {4673, 4774, 15044, 15131, 32768 },
        {193, 194, 384, 479, 32768 } },
      { // Band 4
        {28865, 29687, 32655, 32667, 32768 },
        {23233, 24218, 32080, 32118, 32768 },
        {15041, 15444, 28787, 28845, 32768 },
        {9921, 10248, 22818, 22944, 32768 },
        {7745, 7866, 16591, 16702, 32768 },
        {193, 194, 384, 479, 32768 } },
      { // Band 5
        {31169, 31559, 32741, 32744, 32768 },
        {24769, 25583, 32347, 32370, 32768 },
        {15937, 16169, 29120, 29152, 32768 },
        {7489, 7578, 22647, 22677, 32768 },
        {7617, 7689, 19849, 19887, 32768 },
        {2113, 2183, 7202, 7377, 32768 } } } },
  {     // UV plane
    {   // Inter
      { // Band 0
        { 23232, 24301, 30231, 31582, 32091, 32768 },
        { 7872, 11041, 22542, 27086, 29145, 32768 },
        { 1344, 3989, 18125, 25340, 27820, 32768 } },
      { // Band 1
        {15937, 29000, 32210, 32434, 32768 },
        {12353, 26626, 31533, 31993, 32768 },
        {11457, 29187, 30896, 31750, 32768 },
        {5697, 21278, 28169, 29764, 32768 },
        {7489, 8855, 13365, 15620, 32768 },
        {4289, 4833, 8572, 10108, 32768 } },
      { // Band 2
        {25025, 30783, 32603, 32666, 32768 },
        {24385, 29586, 31803, 32142, 32768 },
        {22337, 23002, 27573, 27903, 32768 },
        {10945, 12336, 21900, 22590, 32768 },
        {8257, 8830, 19986, 20298, 32768 },
        {10945, 10990, 18660, 18701, 32768 } },
      { // Band 3
        {29761, 31473, 32693, 32715, 32768 },
        {20417, 24512, 31394, 31650, 32768 },
        {11713, 13283, 25819, 26206, 32768 },
        {13121, 14099, 21909, 22514, 32768 },
        {193, 248, 9546, 9614, 32768 },
        {2497, 2524, 7050, 7125, 32768 } },
      { // Band 4
        {30657, 31885, 32691, 32715, 32768 },
        {19393, 26050, 31698, 31988, 32768 },
        {15809, 15863, 24985, 25008, 32768 },
        {23489, 28138, 32751, 32756, 32768 },
        {16449, 16450, 16545, 16593, 32768 },
        {2369, 2395, 6822, 6898, 32768 } },
      { // Band 5
        {32705, 32744, 32766, 32767, 32768 },
        {21953, 24962, 32156, 32246, 32768 },
        {13121, 15358, 26284, 26835, 32768 },
        {5697, 7417, 20132, 20885, 32768 },
        {4417, 4939, 15104, 15535, 32768 },
        {2625, 2680, 8218, 8338, 32768 } } },
    {   // Inter
      { // Band 0
        { 25280, 25678, 32446, 32622, 32724, 32768 },
        { 10560, 11822, 28682, 29919, 31276, 32768 },
        { 3264, 5170, 21779, 24026, 27905, 32768 } },
      { // Band 1
        {24257, 30554, 32719, 32738, 32768 },
        {17217, 27413, 32617, 32667, 32768 },
        {22977, 27600, 32482, 32552, 32768 },
        {16833, 24360, 30746, 31293, 32768 },
        {17089, 20060, 28880, 29370, 32768 },
        {10945, 11009, 21900, 21932, 32768 } },
      { // Band 2
        {27201, 30217, 32736, 32745, 32768 },
        {22721, 27676, 32749, 32754, 32768 },
        {5057, 12431, 25246, 26620, 32768 },
        {193, 321, 22016, 22048, 32768 },
        {5313, 5363, 13839, 13894, 32768 },
        {2625, 2652, 7276, 7351, 32768 } },
      { // Band 3
        {27713, 30739, 32759, 32762, 32768 },
        {26177, 30430, 32756, 32760, 32768 },
        {193, 384, 32706, 32707, 32768 },
        {9409, 9528, 21591, 21646, 32768 },
        {193, 194, 384, 479, 32768 },
        {2881, 2913, 8427, 8498, 32768 } },
      { // Band 4
        {28993, 31156, 32747, 32753, 32768 },
        {25153, 28701, 32754, 32758, 32768 },
        {16449, 16544, 32737, 32738, 32768 },
        {193, 321, 22016, 22048, 32768 },
        {193, 194, 384, 479, 32768 },
        {2753, 2789, 8909, 8979, 32768 } },
      { // Band 5
        {30785, 32088, 32765, 32766, 32768 },
        {22977, 26623, 32750, 32754, 32768 },
        {21953, 21954, 22017, 22049, 32768 },
        {5697, 7486, 20238, 21009, 32768 },
        {4929, 5579, 16402, 16866, 32768 },
        {3009, 3246, 10158, 10533, 32768 } } } }
};
#endif  // CONFIG_NEW_TOKENSET

/* clang-format on */

static void extend_to_full_distribution(aom_prob *probs, aom_prob p) {
  assert(p != 0);
  memcpy(probs, av1_pareto8_full[p - 1], MODEL_NODES * sizeof(aom_prob));
}

void av1_model_to_full_probs(const aom_prob *model, aom_prob *full) {
  if (full != model)
    memcpy(full, model, sizeof(aom_prob) * UNCONSTRAINED_NODES);
  extend_to_full_distribution(&full[UNCONSTRAINED_NODES], model[PIVOT_NODE]);
}

#if CONFIG_NEW_TOKENSET

static void build_tail_cdfs(aom_cdf_prob cdf_tail[ENTROPY_TOKENS + 1],
                            aom_cdf_prob cdf_head[ENTROPY_TOKENS + 1]) {
  int probNZ, prob1, prob_idx, i;
  int phead[6], sum, p;
  for (i = 0; i < 6; ++i) {
    phead[i] = cdf_head[i] - (i == 0 ? 0 : cdf_head[i - 1]);
  }
  // Do the tail
  probNZ = 32768 - phead[1 + ZERO_TOKEN] - phead[0];
  prob1 = phead[1 + ONE_TOKEN_EOB] + phead[1 + ONE_TOKEN_NEOB];
  prob_idx =
      AOMMIN(COEFF_PROB_MODELS - 1, AOMMAX(0, ((256 * prob1) / probNZ) - 1));

  sum = 0;
  for (i = 0; i < ENTROPY_TOKENS - 3; ++i) {
    p = av1_pareto8_tail_probs[prob_idx][i];
    cdf_tail[i] = sum += p;
  }
}

static void build_head_cdfs(const aom_prob *pdf_model,
                            const aom_prob *blockz_model,
                            aom_cdf_prob cdf_head[ENTROPY_TOKENS + 1]) {
  int i, p, p1, p2, phead[6], prob_NZ, prob_EOB_1, prob_EOB_2p, prob_NEOB_1,
      prob_NEOB_2p;
  int prob8_blocknz;
  // We have the first coefficient position and so an extended CDF
  const int is_dc = blockz_model != NULL;
  const int last_head_val = HEAD_TOKENS - 1 + is_dc;

  assert(pdf_model[2] != 0);

  /* FIXME: maintain true CDF counts. */

  /* Values are 0=BLOCK_ZERO 1=ZERO_TOKEN, 2=ONE_TOKEN_EOB
     3=ONE_TOKEN_NEOB, 4=TWO_TOKEN_PLUS_EOB, 5=TWO_TOKEN_PLUS_NEOB
     */
  // Block zero probability
  if (is_dc) {
    phead[0] =
        ((*blockz_model) << (CDF_PROB_BITS - 8)) + (1 << (CDF_PROB_BITS - 9));
    phead[0] = AOMMIN(CDF_PROB_TOP - (HEAD_TOKENS + 1), AOMMAX(1, phead[0]));
  }

  // Will scale the remaining probabilities by the probability of the block
  // being non-zero
  prob8_blocknz = is_dc ? (256 - *blockz_model) : 256;

  // Probability of zero
  phead[is_dc + ZERO_TOKEN] =
      (pdf_model[1] << (CDF_PROB_BITS - 8)) + (1 << (CDF_PROB_BITS - 9));

  // Will scale the non-zero values
  prob_NZ = CDF_PROB_TOP - phead[is_dc + ZERO_TOKEN];

  // Will scale the EOBs by the probability of and EOB_TOKEN ..
  prob_EOB_1 =
      (pdf_model[0] << (CDF_PROB_BITS - 8)) + (1 << (CDF_PROB_BITS - 9));
  // .. use a lower probability of EOB for larger values
  prob_EOB_2p = prob_EOB_1 / 2;

  prob_NEOB_1 = CDF_PROB_TOP - prob_EOB_1;
  prob_NEOB_2p = CDF_PROB_TOP - prob_EOB_2p;
  if (prob_NZ == 0 || prob_NZ == CDF_PROB_TOP) abort();
  if (prob_EOB_1 == 0 || prob_EOB_1 == CDF_PROB_TOP) abort();
  if (prob_EOB_2p == 0 || prob_EOB_2p == CDF_PROB_TOP) abort();

  // ONE_CONTEXT_NODE prob
  p = (pdf_model[2] << (CDF_PROB_BITS - 8)) + (1 << (CDF_PROB_BITS - 9));
  // Scale by the non-zero factor to get the probability of token = 1
  p1 = ROUND_POWER_OF_TWO(prob_NZ * p, 15);

  // Scale by the EOB factors
  phead[is_dc + ONE_TOKEN_EOB] = ROUND_POWER_OF_TWO(p1 * prob_EOB_1, 15);
  phead[is_dc + ONE_TOKEN_NEOB] = ROUND_POWER_OF_TWO(p1 * prob_NEOB_1, 15);

  // Probability token is 2 or more
  p2 = CDF_PROB_TOP - p1 - phead[is_dc + ZERO_TOKEN];

  phead[is_dc + TWO_TOKEN_PLUS_EOB] = ROUND_POWER_OF_TWO(p2 * prob_EOB_2p, 15);
  phead[is_dc + TWO_TOKEN_PLUS_NEOB] =
      ROUND_POWER_OF_TWO(p2 * prob_NEOB_2p, 15);

  // Now use block non-zerp prob to scale the values
  for (i = is_dc; i < last_head_val; ++i) {
    phead[i] = (prob8_blocknz * phead[i] + 128) >> 8;
  }

  for (i = 0; i < last_head_val; ++i) {
    p = AOMMAX(1, AOMMIN(CDF_PROB_TOP - (last_head_val - i) - cdf_head[i - 1],
                         phead[i]));
    cdf_head[i] = (i > 0 ? cdf_head[i - 1] : 0) + p;
  }
  cdf_head[last_head_val] = CDF_PROB_TOP;
}

static void av1_default_coef_cdfs(FRAME_CONTEXT *fc) {
  int i, j, k, l;
  for (i = 0; i < PLANE_TYPES; ++i)
    for (j = 0; j < REF_TYPES; ++j)
      for (k = 0; k < COEF_BANDS; ++k)
        for (l = 0; l < BAND_COEFF_CONTEXTS(k); ++l) {
#if CONFIG_CB4X4
          av1_copy(fc->coef_head_cdfs[TX_2X2][i][j][k][l],
                   default_coef_head_cdf_4x4[i][j][k][l]);
#endif
          av1_copy(fc->coef_head_cdfs[TX_4X4][i][j][k][l],
                   default_coef_head_cdf_4x4[i][j][k][l]);
          av1_copy(fc->coef_head_cdfs[TX_8X8][i][j][k][l],
                   default_coef_head_cdf_8x8[i][j][k][l]);
          av1_copy(fc->coef_head_cdfs[TX_16X16][i][j][k][l],
                   default_coef_head_cdf_16x16[i][j][k][l]);
          av1_copy(fc->coef_head_cdfs[TX_32X32][i][j][k][l],
                   default_coef_head_cdf_32x32[i][j][k][l]);
#if CONFIG_TX64X64
          av1_copy(fc->coef_head_cdfs[TX_64X64][i][j][k][l],
                   default_coef_head_cdf_64x64[i][j][k][l]);
#endif
        }
}

void av1_coef_head_cdfs(FRAME_CONTEXT *fc) {
  TX_SIZE t;
  int i, j, k, l;
  for (t = 0; t < TX_SIZES; ++t)
    for (i = 0; i < PLANE_TYPES; ++i)
      for (j = 0; j < REF_TYPES; ++j)
        for (k = 0; k < COEF_BANDS; ++k)
          for (l = 0; l < BAND_COEFF_CONTEXTS(k); ++l) {
            build_head_cdfs(fc->coef_probs[t][i][j][k][l],
                            k == 0 ? &fc->blockzero_probs[t][i][j][l] : NULL,
                            fc->coef_head_cdfs[t][i][j][k][l]);
          }
}

#elif CONFIG_EC_MULTISYMBOL
static void build_token_cdfs(const aom_prob *pdf_model,
                             aom_cdf_prob cdf[ENTROPY_TOKENS + 1]) {
  int i, sum = 0;
  assert(pdf_model[2] != 0);
  for (i = 0; i < ENTROPY_TOKENS - 2; ++i) {
    cdf[i] = sum += av1_pareto8_token_probs[pdf_model[2] - 1][i];
  }
}
#endif  // CONFIG_NEW_TOKENSET

#if CONFIG_EC_MULTISYMBOL
void av1_coef_pareto_cdfs(FRAME_CONTEXT *fc) {
  /* Build the tail based on a Pareto distribution */
  TX_SIZE t;
  int i, j, k, l;
  for (t = 0; t < TX_SIZES; ++t)
    for (i = 0; i < PLANE_TYPES; ++i)
      for (j = 0; j < REF_TYPES; ++j)
        for (k = 0; k < COEF_BANDS; ++k)
          for (l = 0; l < BAND_COEFF_CONTEXTS(k); ++l)
#if CONFIG_NEW_TOKENSET
            build_tail_cdfs(fc->coef_tail_cdfs[t][i][j][k][l],
                            fc->coef_head_cdfs[t][i][j][k][l]);
#else
            build_token_cdfs(fc->coef_probs[t][i][j][k][l],
                             fc->coef_cdfs[t][i][j][k][l]);
#endif
}
#endif

void av1_default_coef_probs(AV1_COMMON *cm) {
#if CONFIG_Q_ADAPT_PROBS
  const int index = AOMMIN(
      ROUND_POWER_OF_TWO(cm->base_qindex, 8 - QCTX_BIN_BITS), QCTX_BINS - 1);
  av1_copy(cm->fc->coef_probs, default_qctx_coef_probs[index]);
#else
#if CONFIG_CB4X4
  av1_copy(cm->fc->coef_probs[TX_2X2], default_coef_probs_4x4);
#endif
  av1_copy(cm->fc->coef_probs[TX_4X4], default_coef_probs_4x4);
  av1_copy(cm->fc->coef_probs[TX_8X8], default_coef_probs_8x8);
  av1_copy(cm->fc->coef_probs[TX_16X16], default_coef_probs_16x16);
  av1_copy(cm->fc->coef_probs[TX_32X32], default_coef_probs_32x32);
#if CONFIG_TX64X64
  av1_copy(cm->fc->coef_probs[TX_64X64], default_coef_probs_64x64);
#endif  // CONFIG_TX64X64
#endif  // CONFIG_Q_ADAPT_PROBS
#if CONFIG_NEW_TOKENSET
  av1_copy(cm->fc->blockzero_probs, av1_default_blockzero_probs);
#endif
#if CONFIG_NEW_TOKENSET
  /* Load the head tokens */
  av1_default_coef_cdfs(cm->fc);
#endif
#if CONFIG_EC_MULTISYMBOL
  av1_coef_pareto_cdfs(cm->fc);
#endif  // CONFIG_EC_MULTISYMBOL
}

#if !CONFIG_LV_MAP
static void adapt_coef_probs(AV1_COMMON *cm, TX_SIZE tx_size,
                             unsigned int count_sat,
                             unsigned int update_factor) {
  const FRAME_CONTEXT *pre_fc = &cm->frame_contexts[cm->frame_context_idx];
  av1_coeff_probs_model *const probs = cm->fc->coef_probs[tx_size];
#if CONFIG_SUBFRAME_PROB_UPDATE
  const av1_coeff_probs_model *const pre_probs =
      cm->partial_prob_update
          ? (const av1_coeff_probs_model *)cm->starting_coef_probs[tx_size]
          : pre_fc->coef_probs[tx_size];
#else
  const av1_coeff_probs_model *const pre_probs = pre_fc->coef_probs[tx_size];
#endif  // CONFIG_SUBFRAME_PROB_UPDATE
  const av1_coeff_count_model *const counts =
      (const av1_coeff_count_model *)cm->counts.coef[tx_size];
  const unsigned int(*eob_counts)[REF_TYPES][COEF_BANDS][COEFF_CONTEXTS] =
      (const unsigned int(*)[REF_TYPES][COEF_BANDS][COEFF_CONTEXTS])
          cm->counts.eob_branch[tx_size];
#if CONFIG_NEW_TOKENSET
  const av1_blockz_probs_model *const pre_blockz_probs =
      pre_fc->blockzero_probs[tx_size];
  av1_blockz_probs_model *const blockz_probs = cm->fc->blockzero_probs[tx_size];
  const av1_blockz_count_model *const blockz_counts =
      (const av1_blockz_count_model *)&cm->counts.blockz_count[tx_size][0];
#endif
  int i, j, k, l, m;
#if CONFIG_RECT_TX
  assert(!is_rect_tx(tx_size));
#endif  // CONFIG_RECT_TX

  for (i = 0; i < PLANE_TYPES; ++i)
    for (j = 0; j < REF_TYPES; ++j)
      for (k = 0; k < COEF_BANDS; ++k)
        for (l = 0; l < BAND_COEFF_CONTEXTS(k); ++l) {
          const int n0 = counts[i][j][k][l][ZERO_TOKEN];
          const int n1 = counts[i][j][k][l][ONE_TOKEN];
          const int n2 = counts[i][j][k][l][TWO_TOKEN];
          const int neob = counts[i][j][k][l][EOB_MODEL_TOKEN];
          const unsigned int branch_ct[UNCONSTRAINED_NODES][2] = {
            { neob, eob_counts[i][j][k][l] - neob }, { n0, n1 + n2 }, { n1, n2 }
          };
          for (m = 0; m < UNCONSTRAINED_NODES; ++m)
            probs[i][j][k][l][m] =
                av1_merge_probs(pre_probs[i][j][k][l][m], branch_ct[m],
                                count_sat, update_factor);
        }

#if CONFIG_NEW_TOKENSET
  for (i = 0; i < PLANE_TYPES; ++i) {
    for (j = 0; j < REF_TYPES; ++j) {
      for (k = 0; k < BLOCKZ_CONTEXTS; ++k) {
        const int n0 = blockz_counts[i][j][k][0];
        const int n1 = blockz_counts[i][j][k][1];
        const unsigned int branch_ct[2] = { n0, n1 };
        blockz_probs[i][j][k] = av1_merge_probs(
            pre_blockz_probs[i][j][k], branch_ct, count_sat, update_factor);
      }
    }
  }
#endif
}
#endif  // !CONFIG_LV_MAP

void av1_adapt_coef_probs(AV1_COMMON *cm) {
  unsigned int count_sat, update_factor;

  if (!frame_is_intra_only(cm) && cm->last_frame_type == KEY_FRAME) {
    update_factor = COEF_MAX_UPDATE_FACTOR_AFTER_KEY; /* adapt quickly */
    count_sat = COEF_COUNT_SAT_AFTER_KEY;
  } else {
    update_factor = COEF_MAX_UPDATE_FACTOR;
    count_sat = COEF_COUNT_SAT;
  }
#if CONFIG_SUBFRAME_PROB_UPDATE
  if (cm->partial_prob_update == 1) update_factor = COEF_MAX_UPDATE_FACTOR;
#endif  // CONFIG_SUBFRAME_PROB_UPDATE

#if CONFIG_LV_MAP
  av1_adapt_txb_probs(cm, count_sat, update_factor);
#else
  TX_SIZE tx_size;
  for (tx_size = 0; tx_size < TX_SIZES; tx_size++)
    adapt_coef_probs(cm, tx_size, count_sat, update_factor);
#endif
}

#if CONFIG_SUBFRAME_PROB_UPDATE
void av1_partial_adapt_probs(AV1_COMMON *cm, int mi_row, int mi_col) {
  (void)mi_row;
  (void)mi_col;

  if (cm->refresh_frame_context == REFRESH_FRAME_CONTEXT_BACKWARD) {
    cm->partial_prob_update = 1;
    av1_adapt_coef_probs(cm);
  }
}
#endif  // CONFIG_SUBFRAME_PROB_UPDATE

#if CONFIG_EC_ADAPT
static void av1_average_cdf(aom_cdf_prob *cdf_ptr[], aom_cdf_prob *fc_cdf_ptr,
                            int cdf_size, const int num_tiles) {
  int i, j;
  aom_cdf_prob last_val = 0;
  for (i = 0; i < cdf_size; ++i) {
    // Zero symbol counts for the next frame
    if (last_val == CDF_PROB_TOP) {
      fc_cdf_ptr[i] = 0;
    } else {
      int sum = 0;
      for (j = 0; j < num_tiles; ++j) sum += cdf_ptr[j][i];
      fc_cdf_ptr[i] = sum / num_tiles;
    }
    last_val = fc_cdf_ptr[i];
  }
}

#define AVERAGE_TILE_CDFS(cname)                            \
  for (i = 0; i < num_tiles; ++i)                           \
    cdf_ptr[i] = (aom_cdf_prob *)&ec_ctxs[i]->cname;        \
  fc_cdf_ptr = (aom_cdf_prob *)&fc->cname;                  \
  cdf_size = (int)sizeof(fc->cname) / sizeof(aom_cdf_prob); \
  av1_average_cdf(cdf_ptr, fc_cdf_ptr, cdf_size, num_tiles);

void av1_average_tile_coef_cdfs(FRAME_CONTEXT *fc, FRAME_CONTEXT *ec_ctxs[],
                                const int num_tiles) {
  int i, cdf_size;

  aom_cdf_prob *cdf_ptr[MAX_TILE_ROWS * MAX_TILE_COLS];
  aom_cdf_prob *fc_cdf_ptr;

#if CONFIG_NEW_TOKENSET
  AVERAGE_TILE_CDFS(coef_head_cdfs)
  AVERAGE_TILE_CDFS(coef_tail_cdfs)
#else
  AVERAGE_TILE_CDFS(coef_cdfs)
#endif
}

void av1_average_tile_mv_cdfs(FRAME_CONTEXT *fc, FRAME_CONTEXT *ec_ctxs[],
                              const int num_tiles) {
  int i, k, cdf_size;

  aom_cdf_prob *cdf_ptr[MAX_TILE_ROWS * MAX_TILE_COLS];
  aom_cdf_prob *fc_cdf_ptr;

#if CONFIG_REF_MV
  int j;
  for (j = 0; j < NMV_CONTEXTS; ++j) {
    AVERAGE_TILE_CDFS(nmvc[j].joint_cdf)

    for (k = 0; k < 2; ++k) {
      AVERAGE_TILE_CDFS(nmvc[j].comps[k].class_cdf);
      AVERAGE_TILE_CDFS(nmvc[j].comps[k].class0_fp_cdf);
      AVERAGE_TILE_CDFS(nmvc[j].comps[k].fp_cdf);
    }
  }
#else
  AVERAGE_TILE_CDFS(nmvc.joint_cdf)

  for (k = 0; k < 2; ++k) {
    AVERAGE_TILE_CDFS(nmvc.comps[k].class_cdf)
    AVERAGE_TILE_CDFS(nmvc.comps[k].class0_fp_cdf)
    AVERAGE_TILE_CDFS(nmvc.comps[k].fp_cdf)
  }
#endif
}

void av1_average_tile_intra_cdfs(FRAME_CONTEXT *fc, FRAME_CONTEXT *ec_ctxs[],
                                 const int num_tiles) {
  int i, cdf_size;

  aom_cdf_prob *cdf_ptr[MAX_TILE_ROWS * MAX_TILE_COLS];
  aom_cdf_prob *fc_cdf_ptr;

  AVERAGE_TILE_CDFS(tx_size_cdf);

#if CONFIG_VAR_TX
// FIXME: txfm_partition probs
#endif

  // FIXME: skip probs

  AVERAGE_TILE_CDFS(intra_ext_tx_cdf)
  AVERAGE_TILE_CDFS(inter_ext_tx_cdf);

  AVERAGE_TILE_CDFS(seg.tree_cdf)
  AVERAGE_TILE_CDFS(uv_mode_cdf)

  AVERAGE_TILE_CDFS(partition_cdf)

#if CONFIG_DELTA_Q
  AVERAGE_TILE_CDFS(delta_q_cdf)
#endif
#if CONFIG_EXT_INTRA && CONFIG_INTRA_INTERP
  AVERAGE_TILE_CDFS(intra_filter_cdf)
#endif  // CONFIG_EXT_INTRA && CONFIG_INTRA_INTERP
#if CONFIG_FILTER_INTRA
#endif  // CONFIG_FILTER_INTRA
}

void av1_average_tile_inter_cdfs(AV1_COMMON *cm, FRAME_CONTEXT *fc,
                                 FRAME_CONTEXT *ec_ctxs[],
                                 const int num_tiles) {
  int i, cdf_size;

  aom_cdf_prob *cdf_ptr[MAX_TILE_ROWS * MAX_TILE_COLS];
  aom_cdf_prob *fc_cdf_ptr;

// FIXME: comp_inter_cdf not defined

// FIXME: comp_ref_cdf and comp_bwd_ref not defined

// FIXME: single_ref_cdf not defined

#if CONFIG_REF_MV
// FIXME: cdfs not defined for newmv_mode, zeromv_mode, drl_mode, new2mv_mode
#else
  AVERAGE_TILE_CDFS(inter_mode_cdf)
#endif

  // FIXME: cdfs not defined for motion_mode_prob, obmc_prob

  // FIXME: cdfs not defined for super_tx

  // FIXME: CONFIG_EXT_INTER cdfs not defined for inter_compound_mode,
  // interintra_mode etc

  AVERAGE_TILE_CDFS(y_mode_cdf)

  if (cm->interp_filter == SWITCHABLE) {
    AVERAGE_TILE_CDFS(switchable_interp_cdf)
  }
}
#endif
